﻿/*
 * This file is part of AV Orchid.
 * Copyright (c) 2014, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * Пример: APlot2D
 * Автор: Кувшинов Д.Р. https://evetro.wordpress.com/
 * Используемые библиотеки: FreeGlut, OpenGL.
 *
 * Однопараметрические графики.
 *
 * Управление "камерой": стрелки и WASD, +/- Z/X для увеличения/уменьшения масштаба.
 * Колёсико мыши изменяет масштаб (как +/-).
 * HOME -- вернуть камеру в положение по умолчанию.
 * ESC -- выход из программы.
 */
#include <avo/CoordinateGrid2.hpp>
#include <avo/TextDrawer.hpp>
#include <avo/Camera2.hpp>
#include <avo/Timer.hpp>

#include <cmath>
#include <string>

#include "APlot2DGraph.hpp"
#include "APlot2D.hpp"

using Vec_2 = avo::Vec_2<Scalar>;


/// определение "сцены"
class Scene
{
  using Graphs = std::vector<APlot2D_graph>;
  using Legend = std::vector<std::string>;

  Graphs graphs; // набор примитивов, представляющий графики
  Legend legend; // названия графиков (параллельный массив)
  std::vector<unsigned> drawing_legend; // индексы выводимых надписей
  Bbox_2  bbox;   // бокс, в который вписана сетка (для автопозиционирования камеры)

  avo::Camera_2 camera;  // камера
  avo::Coordinate_grid_2<double> cgrid; // координатная сетка
  mutable avo::Bitmap_text_drawer_2 text_drawer; // объект, выводящий текст
  int mouse_x, mouse_y; // координаты мыши

public:

  /// изобразить "сцену"
  void paint(int plot_mode) const
  {
    cgrid.paint();
    for (auto &g : graphs)
      g.paint(plot_mode);

    draw_text();
  }

  /// получить бокс
  const Bbox_2& box() const
  {
    return bbox;
  }

  /// добавить график
  void add_graph(APlot2D_graph &&graph, std::string name = "<unknown>")
  {
    if (graphs.empty())
      bbox = graph.box();
    else
      bbox += graph.box();
    graphs.emplace_back(std::move(graph));
    legend.emplace_back(std::move(name));
  }


  /// приблизить камеру
  void camera_zoom_in()
  {
    camera.zoom_in(0.125);
    camera_changed();
  }

  /// "отодвинуть" камеру
  void camera_zoom_out()
  {
    camera.zoom_out(0.125);
    camera_changed();
  }

  /// передвинуть камеру
  void camera_move(const Vec_2 &delta)
  {
    camera.move(delta * camera.box().extents());
    camera_changed();
  }

  /// показать всю занятую область
  void camera_reset()
  {
    camera.reset(box());
    camera_changed();
  }

  /// изменён размер окна
  void reshape(int w, int h)
  {
    // изменить окно камеры
    camera.force_aspect((double)w / h);
    camera.apply();
    camera_changed();
    text_drawer.set_viewport(w, h);
  }

  /// установить координаты мыши
  void mouse_position(int x, int y)
  {
    mouse_x = x;
    mouse_y = y;
  }

private:

  /// обновить координатную сетку при изменении камеры
  void camera_changed()
  {
    cgrid.update(camera.box());
    text_drawer.set_coordinates(camera.box());
    camera.apply();
    glutPostRedisplay();  // уведомить FreeGlut, что следует перерисовать сцену
  }


  /// отображение координат и легенды графиков
  void draw_text() const
  {
    using std::to_string;
    auto smaller_box = camera;
    smaller_box.zoom_in(0.05); // чтобы не пропадали надписи на краях при округлении "наружу"
    auto &box = smaller_box.box();

    text_drawer.font(GLUT_BITMAP_HELVETICA_10);
    // размеры координатной сетки
    {
      // сгенерировать строки
      const auto left   = to_string(box.pmin.x),
                 bottom = to_string(box.pmin.y),
                 right  = to_string(box.pmax.x),
                 top    = to_string(box.pmax.y);

      // вывести текст средствами FreeGlut
      const auto center = 0.5 * (box.pmax + box.pmin);

      avo::LIGHT_GREY.select();
      text_drawer.view(box.pmin.x, center.y).put(left)
                 .view(center.x, box.pmin.y).put(bottom)
                 .view(box.pmax.x, center.y).put(right, avo::Alignment::toLeft)
                 .view(center.x, box.pmax.y).put(top, avo::Alignment::toDown)
                 ;
    }

    // координаты мыши
    {
      double mx, my; // "координаты" мыши
      text_drawer.pixels_to_view(mouse_x, mouse_y, mx, my);

      // сгенерировать строки
      const auto xy = to_string(mx) + '\n' + to_string(my);
    
      // вывести текст средствами FreeGlut
      avo::AZURE.select();
      text_drawer.view(box.pmin.x, box.pmax.y).put(xy, avo::Alignment::toDown);
    }

    // легенда
    text_drawer.font(GLUT_BITMAP_9_BY_15);
    {
      double y_step = 0., not_used;
      text_drawer.ortho_text_extents("", not_used, y_step);

      double y = box.pmax.y;
      for (Graphs::size_type i = 0; i < graphs.size(); ++i)
      {
        graphs[i].color().select();
        text_drawer.view(box.pmax.x, y).put(legend[i], avo::Alignment::toLeft | avo::Alignment::toDown);
        y -= y_step;
      }
    }
  }
};


//
// глобальные объекты

Scene scene;

/// текущая высота окна
static int cur_wnd_height;
/// текущий режим рисования графиков: "точки", "линии", "линии и точки"
int plot_mode = 1;

//
// добавление графиков "извне"

plot::~plot()
{
  APlot2D_graph pg{ color, line_width };
  for (auto &b: branches)
    pg.add_branch(std::move(b));
  scene.add_graph(std::move(pg), title);
}


//
// колбеки FreeGlut

/// перерисовка
void repaint()
{
  glClear(GL_COLOR_BUFFER_BIT);
  scene.paint(plot_mode);
  glutSwapBuffers();
}


/// изменён размер окна
void reshape(int w, int h)
{
  scene.reshape(w, h);
  cur_wnd_height = h;
  // уведомить OpenGL о новых размерах области рисования
  glViewport(0, 0, w, h);
}


/// обработать нажатие "специальной" клавиши
void on_special_key(int key, int x, int y)
{
  Vec_2 delta; // 0, 0
  switch (key)
  {
  case GLUT_KEY_LEFT:   delta.x = -0.1; break;
  case GLUT_KEY_UP:     delta.y =  0.1; break;
  case GLUT_KEY_RIGHT:  delta.x =  0.1; break;
  case GLUT_KEY_DOWN:   delta.y = -0.1; break;
  case GLUT_KEY_HOME:   scene.camera_reset(); return;
  default:
    return; // ничего не делать, если была нажата какая-либо другая клавиша
  }

  scene.camera_move(delta);
}

/// обработать нажатие "обычной" клавиши
void on_key(unsigned char key, int x, int y)
{
  switch (key)
  {
  case 27: // код ESC (ASCII) --  выйти из цикла обработки событий
    return glutLeaveMainLoop();
  
  case '\b': // переключение режима рисования
    if (++plot_mode == 4)
      plot_mode = 1;
    glutPostRedisplay();
    break;

  case '+': case 'z': case 'Z': scene.camera_zoom_in();   break;
  case '-': case 'x': case 'X': scene.camera_zoom_out();  break;
  case 'w': case 'W': on_special_key(GLUT_KEY_UP, x, y); break;
  case 's': case 'S': on_special_key(GLUT_KEY_DOWN, x, y); break;
  case 'a': case 'A': on_special_key(GLUT_KEY_LEFT, x, y); break;
  case 'd': case 'D': on_special_key(GLUT_KEY_RIGHT, x, y); break;
  default:
    return; // do nothing for other keys
  }
}


/// отслеживаем движение мыши, чтобы отображать её координаты
void on_hover(int x, int y)
{
  scene.mouse_position(x, cur_wnd_height - y);
  glutPostRedisplay();
}

/// масштабирование колёсиком мыши
void on_mouse_wheel(int wheel, int direction, int x, int y)
{
  if (direction < 0)
    scene.camera_zoom_in();
  else if (direction > 0)
    scene.camera_zoom_out();
}

/// по возможности не грузим процессор зря :)
void idle()
{
  avo::sleep(0.05);
}


int run_aplot2d(int argc, char **argv)
{
  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_MULTISAMPLE);
  
  glutInitWindowPosition(10, 20);
  glutInitWindowSize(1000, 640);
  glutCreateWindow("APlot 2D");

  glClearColor(0.0f, 0.0f, 0.07f, 1.0f);

  glEnable(GL_BLEND); // прозрачность
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  
  glEnable(GL_COLOR_MATERIAL);
  glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);

  // включить массивы вершин (OpenGL 1.1)
  glEnableClientState(GL_VERTEX_ARRAY);

  // регистрация колбеков
  glutDisplayFunc  (repaint);
  glutReshapeFunc  (reshape);
  glutKeyboardFunc (on_key);
  glutSpecialFunc  (on_special_key);
  glutIdleFunc     (idle);

  glutPassiveMotionFunc(on_hover);
  glutMouseWheelFunc(on_mouse_wheel);

  // старт
  scene.camera_reset();
  glutMainLoop();
  return 0;
}
