﻿/*
 * This file is part of AV Orchid.
 * Copyright (c) 2014, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * @file APlot2DGraph.hpp
 * @brief класс для создания и рисования однопраметрического графика на плоскости
 */
#pragma once
#include <avo/Color.hpp>
#include <avo/Prim2.hpp>
#include <avo/Bbox2.hpp>
#include "PlotVertices2.hpp"

#include <vector>

using Bbox_2 = avo::Bbox_2<Scalar>;
using Prim_2 = avo::Prim_2<Scalar>;
using avo::Color;


class APlot2D_graph
{
public:

  using This = APlot2D_graph;
  enum { paint_lines = 1, paint_points = 2, paint_lines_and_points = 3 };

  APlot2D_graph(): graph_width(1.f) {}

  explicit APlot2D_graph(const Color &color, float width = 1.f)
    : graph_color(color),
      graph_color_tr(color.red, color.green, color.blue, 0.7f * color.alpha),
      graph_width(width)
  {}

  APlot2D_graph(Points_2 &&points, const Color &color, float width = 1.f)
    : APlot2D_graph(color, width)
  {
    add_branch(std::move(points));
  }

  APlot2D_graph(This &&other)
    : prims(std::move(other.prims)),
      bbox(other.bbox),
      graph_color(other.graph_color),
      graph_color_tr(other.graph_color_tr),
      graph_width(other.graph_width)
  {}

  void add_branch(Points_2 &&points)
  {
    Bbox_2 pbox(points.cbegin(), points.cend());
    if (prims.empty())
      bbox = pbox;
    else
      bbox += pbox;

    prims.emplace_back();
    prims.back().vertices() = std::move(points);
  }

  void paint(int mode = paint_lines) const
  {
    if (mode & paint_lines)
    {
      (mode & paint_points? graph_color_tr: graph_color).select();
      glLineWidth(graph_width);
      paint_prims(GL_LINE_STRIP);
    }

    if (mode & paint_points)
    {
      graph_color.select();
      glPointSize(mode & paint_lines? 1.5f * graph_width: graph_width);
      paint_prims(GL_POINTS);
    }
  }

  const Color& color() const { return graph_color; }
  const Bbox_2& box() const { return bbox; }

private:

  using Prims = std::vector<Prim_2>;
  Prims prims;
  Bbox_2 bbox;
  Color graph_color, graph_color_tr;
  float graph_width;

  void paint_prims(GLenum mode) const
  {
    for (auto &p: prims)
      p.paint(mode);
  }
};


