﻿/*
 * This file is part of AV Orchid.
 * Copyright (c) 2013, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * @file Astar.hpp
 * @brief Реализация алгоритма поиска пути A*
 * @author Кувшинов Д.Р. https://evetro.wordpress.com/
 */
#pragma once
#include <cstddef>

#include <set>
#include <vector>

#include <functional>
#include <type_traits>


/// Класс-компаратор, реализующий косвенное сравнение
/**
  Компаратор двух индексов, сравнивающий их в первую очередь по
  соответствующим значениям некоторого массива с помощью компаратора Cmp1,
  если же эти значения эквивалентны относительно Cmp1, то 
  сравниваются непосредственно сами индексы с помощью компаратора Cmp2.
  Второе сравнение нужно для того, чтобы набор индексов мог включать
  любой индекс не более одного раза, допуская при этом совпадение
  значений по разным индексам (эти значения нужны для упорядочивания).
*/
template<class T, class Arr, class Cmp1, class Cmp2 = std::less<T>>
class Indirect_comparator
{
public:
  Indirect_comparator()
    : arr(nullptr), cmp1(), cmp2() {}

  explicit Indirect_comparator(const Arr &a, Cmp1 cmp1 = Cmp1(), Cmp2 cmp2 = Cmp2())
    : arr(&a), cmp1(cmp1), cmp2(cmp2) {}

  bool operator()(const T a, const T b) const
  {
    const auto &ra = (*arr)[a]; // ссылки вводятся, чтобы сократить запись
    const auto &rb = (*arr)[b];
    return cmp1(ra, rb)
       || !cmp1(rb, ra) && cmp2(a, b); // && имеет приоритет выше, чем ||
  }

private:
  const Arr *arr;
  Cmp1 cmp1;
  Cmp2 cmp2;
};


/// индекс вершины (для краткости часто отождествляется с вершиной)
using Vertex_index = std::size_t;


/// Политика генерации событий демонстрации алгоритма, используемая по умолчанию. Ничего не делает.
struct Astar_default_demo_policy
{
  void open_vertex(Vertex_index v, float score) {}
  void close_vertex(Vertex_index v) {}
};


/// Алгоритм поиска кратчайшего пути A*
/**
  Алгоритм A* осуществляет поиск кратчайшего пути между двумя заданными вершинами
  с использованием эвристики (быстро оценивающей расстояние между любыми двумя вершинами графа)
  для выбора приоритетного направления движения на каждом шаге.

  Для перебора соседей используется вектор наборов соседей AdjVec,
  для получения "длины" ребра -- бинарный функтор WeightMatrix.
*/
template
  <
    class Previous
  , class WeightMatrix
  , class AdjVec
  , class Heuristic
  , class DemoPolicy = Astar_default_demo_policy
  >
bool astar
  (
    Previous &Pr             ///< набор предшественников, позволяющий восстановить построенный путь, должен иметь размер >= N
  , const WeightMatrix &W    ///< длины рёбер
  , const AdjVec &A          ///< представление графа в виде вектора наборов соседей
  , const Vertex_index start ///< индекс начальной вершины
  , const Vertex_index goal  ///< индекс конечной вершины
  , const Vertex_index N     ///< количество вершин в графе
  , Heuristic f              ///< эвристика -- функтор, оценивающий расстояние между парой вершин
  , DemoPolicy demo = DemoPolicy() ///< генератор событий для визуализации алгоритма  
  )
{
  // удобные синонимы типов
  using Distance  = std::decay_t<decltype(f(start, goal))>;
  using Array     = std::vector<Distance>;
  using Comp      = Indirect_comparator<Vertex_index, Array, std::less<Distance>>;
  using Adjacency = typename AdjVec::value_type;    // набор соседей

  // инициализация массивов
  Array gscore(N), fscore(N);
  gscore[start] = Distance(0);
  fscore[start] = f(start, goal);
  Pr[start] = start;

  // наборы "закрытых" и "открытых" вершин
  std::vector<bool> closedset(N);
  std::set<Vertex_index, Comp> openset{Comp(fscore)}; // упорядочивать в силу fscore

  // начать с начальной вершины,
  openset.insert(start);
    // -- demo --
    demo.open_vertex(start, fscore[start]);

  // продолжать, пока есть "открытые"
  while (!openset.empty())
  {
    // получить индекс следующей вершины, закончить, если это конечная вершина
    const Vertex_index x = *openset.begin();
    if (x == goal)
      return true;

    // переместить вершину из набора "открытых" в набор "закрытых"
    openset.erase(openset.begin());
    closedset[x] = true;
      // -- demo --
      demo.close_vertex(x);

    // исследовать соседей вершины
    for (auto y: A[x])  // индекс очередного соседа
    {
      if (!closedset[y]) // если его нет в наборе "закрытых", то
      {
        // оценка длины пути, проходящего по ребру (x, y)
        const Distance possible_gscore = gscore[x] + W(x, y);

        const auto p = openset.find(y);
        const bool yNotFound = p == openset.end() // есть ли эта вершина среди "открытых"?
                 , possibleIsLower = possible_gscore < gscore[y]; // может ли вход из x сократить
                                                                  // длину пути, проходящего через y?

        // если то или иное -- истина, то
        if (yNotFound || possibleIsLower)
        {
          // удалить вершину y из "открытых", если y там была раньше
          if (!yNotFound) // (чтобы корректно переупорядочить вершины из-за изменения fscore)
            openset.erase(p);
          
          // путь, включающий y, пойдёт туда через x, пересчитать массивы
          Pr[y] = x;
          gscore[y] = possible_gscore;
          fscore[y] = possible_gscore + f(y, goal);
          // поместить y на новое место в соответствии с fscore
          openset.insert(y);
            // -- demo --
            demo.open_vertex(y, fscore[y]);
        }
      }
    }
  }

  // так и не удалось достичь конечной вершины
  return false;
}


/// Двунаправленный вариант алгоритма поиска кратчайшего пути A*
template
  <
    class Previous
  , class WeightMatrix
  , class AdjVec
  , class Heuristic
  , class DemoPolicy = Astar_default_demo_policy
  >
bool astar2
  (
    Previous &Pr             ///< набор предшественников, позволяющий восстановить построенный путь, должен иметь размер >= N
  , const WeightMatrix &W    ///< длины рёбер
  , const AdjVec &A          ///< представление графа в виде вектора наборов соседей
  , const Vertex_index start ///< индекс начальной вершины
  , const Vertex_index goal  ///< индекс конечной вершины
  , const Vertex_index N     ///< количество вершин в графе
  , Heuristic f              ///< эвристика -- функтор, оценивающий расстояние между парой вершин
  , DemoPolicy demo = DemoPolicy() ///< генератор событий для визуализации алгоритма
  )
{
  // удобные синонимы типов
  using Distance  = std::decay_t<decltype(f(start, goal))>;
  using Array     = std::vector<Distance>;
  using Comp      = Indirect_comparator<Vertex_index, Array, std::less<Distance>>;
  using Adjacency = typename AdjVec::value_type;    // набор соседей
  using Path      = std::vector<Vertex_index>;

  // инициализация массивов
  const auto INF_DIST = std::numeric_limits<Distance>::max();
  Array gscore1(N, INF_DIST), fscore1(N),
        gscore2(N, INF_DIST), fscore2(N);

  gscore1[start] = gscore2[goal] = Distance(0);
  fscore1[start] = fscore2[goal] = f(start, goal);
  Pr[start] = start;

  Path Pr2(N);
  Pr2[goal] = goal;

  // наборы "закрытых" и "открытых" вершин
  std::vector<bool> closedset1(N), closedset2(N);
  std::set<Vertex_index, Comp>
    openset1{Comp(fscore1)},
    openset2{Comp(fscore2)}; // упорядочивать в силу fscore

  // начать с начальной вершины,
  openset1.insert(start);
    demo.open_vertex(start, fscore1[start]); // -- demo --
  // и конечной вершины
  openset2.insert(goal);
    demo.open_vertex(goal, fscore2[goal]); // -- demo --

  // продолжать, пока есть "открытые"
  for (bool do_first = true, do_second = true; do_first || do_second;)
  {
    do_first  = !openset1.empty();
    do_second = !openset2.empty();

    // если доступны оба направления
    if (do_first && do_second)
    {
      // выбрать лучший ход из двух возможных
      if (fscore1[*openset1.begin()] < fscore2[*openset2.begin()])
        do_second = false;
      else
        do_first = false;
    }

    // первая часть
    if (do_first)
    {
      // получить индекс следующей вершины, закончить, если это конечная вершина
      const Vertex_index x = *openset1.begin();
      if (x == goal)
        return true;

      // переместить вершину из набора "открытых" в набор "закрытых"
      openset1.erase(openset1.begin());
      closedset1[x] = true;
        demo.close_vertex(x); // -- demo --

      // исследовать соседей вершины
      for (auto y: A[x])  // индекс очередного соседа
      {
        if (!closedset1[y]) // если его нет в наборе "закрытых", то
        {
          // оценка длины пути, проходящего по ребру (x, y)
          const Distance possible_gscore = gscore1[x] + W(x, y);

          const auto p = openset1.find(y);
          const bool yNotFound = p == openset1.end() // есть ли эта вершина среди "открытых"?
                   , possibleIsLower = possible_gscore < gscore1[y]; // может ли вход из x сократить
                                                                     // длину пути, проходящего через y?

          // если то или иное -- истина, то
          if (yNotFound || possibleIsLower)
          {
            // удалить вершину y из "открытых", если y там была раньше
            if (!yNotFound) // (чтобы корректно переупорядочить вершины из-за изменения fscore)
              openset1.erase(p);
          
            // путь, включающий y, пойдёт туда через x, пересчитать массивы
            Pr[y] = x;
            gscore1[y] = possible_gscore;
            fscore1[y] = possible_gscore + f(y, goal);
            // поместить y на новое место в соответствии с fscore
            openset1.insert(y);
              demo.open_vertex(y, fscore1[y]); // -- demo --
          }
        }
      }

      auto joint = std::make_pair(INF_DIST, N);
      for (auto y: A[x])
      {
        if (closedset2[y])
        {
          const auto dist = gscore1[y] + gscore2[y];
          if (dist < joint.first)
            joint = std::make_pair(dist, y);
        }
      }

      if (joint.second < N)
      {
        // сбросить всё содержимое openset2 в closedset2
        for (auto y: openset2)
        {
          closedset2[y] = true;
            demo.close_vertex(y); // -- demo --
        }

        for (auto x: openset1)
        {
          for (auto y: A[x])
            if (closedset2[y])
            {
              const auto dist = gscore1[x] + W(x, y) + gscore2[y];
              if (dist < joint.first)
              {
                joint = std::make_pair(dist, y);
                Pr[y] = x;
              }
            }
        }

        // развернуть второй участок пути, добавив его в путь-результат
        for (auto v = joint.second, w = Pr2[v]; v != w; v = w, w = Pr2[w])
          Pr[w] = v;
        return true;
      }
    }

    // вторая часть
    if (do_second)
    {
      // получить индекс следующей вершины, закончить, если это конечная вершина
      const Vertex_index x = *openset2.begin();
      if (x == start)
      {
        for (auto v = x, w = Pr2[x]; v != w; v = w, w = Pr2[w])
          Pr[w] = v;
        return true;
      }

      // переместить вершину из набора "открытых" в набор "закрытых"
      openset2.erase(openset2.begin());
      closedset2[x] = true;
        demo.close_vertex(x); // -- demo --

      // исследовать соседей вершины
      for (auto y: A[x])  // индекс очередного соседа
      {
        if (!closedset2[y]) // если его нет в наборе "закрытых", то
        {
          // оценка длины пути, проходящего по ребру (x, y)
          const Distance possible_gscore = gscore2[x] + W(x, y);

          const auto p = openset2.find(y);
          const bool yNotFound = p == openset2.end() // есть ли эта вершина среди "открытых"?
                   , possibleIsLower = possible_gscore < gscore2[y]; // может ли вход из x сократить
                                                                    // длину пути, проходящего через y?

          // если то или иное -- истина, то
          if (yNotFound || possibleIsLower)
          {
            // удалить вершину y из "открытых", если y там была раньше
            if (!yNotFound) // (чтобы корректно переупорядочить вершины из-за изменения fscore)
              openset2.erase(p);

            // путь, включающий y, пойдёт туда через x, пересчитать массивы
            Pr2[y] = x;
            gscore2[y] = possible_gscore;
            fscore2[y] = possible_gscore + f(start, y);
            // поместить y на новое место в соответствии с fscore
            openset2.insert(y);
              demo.open_vertex(y, fscore2[y]); // -- demo --
          }
        }
      }
    }
  }

  // так и не удалось достичь конечной вершины
  return false;
}
