﻿/*
 * This file is part of AV Orchid.
 * Copyright (c) 2013, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * @file Scene.hpp
 * @brief Классы событий визуализации, реализация DemoPolicy, класс сцены, склеивающий прочие компоненты
 * @author Кувшинов Д.Р. https://evetro.wordpress.com/
 */
#pragma once
#include <deque>
#include <memory>
#include <iostream>

#include "Dijkstra.hpp"     // реализация алгоритма Дейкстры
#include "Astar.hpp"        // реализация алгоритма A*
#include "TerrainGraph.hpp" // средства рисования ландшафта и граф ландшафта


/// интерфейс события визуализации
struct IEvent
{
  virtual ~IEvent() {}
  /// событие может быть применено к объекту Terrain для внесения изменений в его внешний вид
  virtual void apply(Terrain&) = 0;
};


/// событие "вершина помещена в открытое множество" ("открыта")
class Opened_event: public IEvent
{
public:

  Opened_event(Vertex_index i, float score)
    : i(i), score(score) {}

  /// калибровка оценки (отображает исходное значение в отрезок [0, 1] )
  void adjust(float minscore, float delta)
  {
    (score -= minscore) /= delta;
  }

  /// применение изменяет цвет
  void apply(Terrain &terr) override
  {
    auto &c = terr.color(i);
    c.red   = 1.f;
    c.green = 1.f - score;
  }

private:
  Vertex_index i;
  float score;
};


/// событие "вершина помещена в закрытое множество" ("закрыта")
class Closed_event: public IEvent
{
  Vertex_index i;

public:
  Closed_event(Vertex_index i)
    : i(i) {}

  void apply(Terrain &terr) override
  {
    terr.color(i).red = 0.0f;
  }
};


/// событие "вершина включена в восстанавливаемый путь"
class Path_event: public IEvent
{
  Vertex_index i;

public:
  Path_event(Vertex_index i): i(i) {}

  void apply(Terrain &terr) override
  {
    terr.color(i) = avo::WHITE;
  }  
};


/// очередь событий (объектов IEvent)
using Event_queue = std::deque<std::unique_ptr<IEvent>>;

/// реализация политики DemoPolicy для функции astar
class Astar_demo
{
public:

  /// привязать объект очереди и переменные, в которые будут записаны получаемые в процессе оценки расстояний
  Astar_demo(Event_queue &Q, float &minscore, float &maxscore)
    : Q(Q), minscore(minscore), maxscore(maxscore) {}

  /// создать событие "вершина открыта"
  void open_vertex(Vertex_index v, float score)
  {
    if (maxscore < score)
      maxscore = score;
    else if (score < minscore)
      minscore = score;
    Q.emplace_back(new Opened_event(v, score)); // std::make_unique?
  }

  /// создать событие "вершина закрыта"
  void close_vertex(Vertex_index v) 
  {
    Q.emplace_back(new Closed_event(v));
  }

  /// создать событие "вершина добавлена в путь"
  void nextPathElement(Vertex_index v)
  {
    Q.emplace_back(new Path_event(v));
  }

private:
  Event_queue &Q;
  float &minscore, &maxscore;
};


/// класс сцены, объединяющий все компоненты программы в единое целое
class Scene
{
  static const std::size_t N = 50 ///< количество строк в карте высот
                         , M = 50 ///< количество столбцов в карте высот
                         ;

  const float xStep ///< шаг сетки по оси 0x
            , yStep ///< шаг сетки по оси 0y
            ;

  /// ландшафт
  Terrain terrain;

  /// граф ландшафта
  Terrain_graph G;

  const avo::Color Gcolor     ///< цвет рёбер графа
                 , ApathColor ///< цвет пути, построенного A*
                 , DpathColor ///< цвет пути, построенного алгоритмом Дейкстры
                 ;

  using Path = std::vector<Vertex_index>;
  Path Apath ///< путь, построенный A*
     , Dpath ///< путь, построенный алгоритмом Дейкстры
     ;

  /// события
  Event_queue Q;

  /// множитель эвристики алгоритма A*
  float factor;
  /// выбранный алгоритм (обычный A* или двунаправленный A*)
  enum { normalAstar, bidirectionalAstar }
    current_algorithm;

public:

  bool paintG     ///< признак "рисовать граф ландшафта"
     , paintApath ///< признак "рисовать путь A*"
     , paintDpath ///< признак "рисовать путь алгоритма Дейкстры"
     ;

  Scene()
    : paintG(false), paintApath(false), paintDpath(false)
    , xStep(10.0f),  yStep(10.0f)
    , factor(1.0f)
    , Gcolor     (0.7f , 0.5f, 0.0f)
    , ApathColor (0.1f , 0.7f, 1.0f)
    , DpathColor (0.25f, 1.0f, 0.1f)
    , terrain(N, M + 1)
    , current_algorithm(normalAstar)
  {}


  /// размер ландшафта вдоль оси 0x
  float x_size() const
  {
    return xStep * N;
  }

  /// размер ландшафта вдоль оси 0y
  float y_size() const
  {
    return yStep * M;
  }

  
  /// построить случайный ландшафт и пути на нём
  void new_terrain_and_paths()
  {
    // построить карту высот случайного ландшафта
    Height_map HM;
    make_landscape(HM, N, M, 20, 4, 8);

    // получить ландшафт и граф ландшафта из карты высот
    terrain.from_height_map(HM, xStep, yStep);
    G.from_terrain(terrain);
    prepare_colors(G, terrain);

    // построить пути
    make_Dpath();
    make_Apath();
  }


  /// отрисовка сцены
  void paint() const
  {
    terrain.paint();
  
    if (paintG)
    {
      Gcolor.select();
      G.paint();
    }
  
    if (paintApath)
    {
      ApathColor.select();
      G.paint(Apath, G.vertices() - 1);
    }

    if (paintDpath)
    {
      DpathColor.select();
      G.paint(Dpath, G.vertices() - 1);
    }
  }

  /// применить следующее событие из очереди
  bool next_event()
  {
    if (Q.empty())
      return false;
    Q.front()->apply(terrain);
    Q.pop_front();
    return true;
  }

  /// применить другой множитель эвристики A* (приводит к перепостроению пути A*)
  void change_factor(float f)
  {
    if (f != factor)
    {
      factor = f;
      prepare_colors(G, terrain);
      make_Apath();
    }
  }

  void switch_algorithm()
  {
    switch (current_algorithm)
    {
      case normalAstar: current_algorithm = bidirectionalAstar; break;
      case bidirectionalAstar: current_algorithm = normalAstar; break;
      default: {}
    }

    prepare_colors(G, terrain);
    make_Apath();
  }

private:

  /// установить цвета треугольников в соответствии с весами рёбер
  void prepare_colors(Terrain_graph &graph, Terrain &terr)
  {
    const Vertex_index N = graph.vertices() - 1;
    const float maxprice = graph.max_price(N);

    for (Vertex_index i = 0; i <= N; ++i)
    {
      const float ratio = graph.estimate_price(i, N) / maxprice;
      terr.color(i) = avo::Color(0.7f - 0.3f * ratio, 0.0f, ratio);
    }
  }

  /// подготовить путь для разметки алгоритмом построения пути
  void prepare_path(Path &path, const Vertex_index size)
  {
    path.resize(size);
    for (Vertex_index i = 0; i < size; ++i)
      path[i] = i;
  }


  /// вычислить стоимость пути
  double path_weight(const Weights_for_graph &W, const Path &P, Vertex_index v)
  {
    double weight = 0.0;
    while (true)
    {
      const auto u = P[v];
      if (u == v)
        return weight;
      weight += W(u, v);
      v = u;
    }
  }


  /// построить путь алгоритмом Дейкстры (без визуализации)
  void make_Dpath()
  {
    prepare_path(Dpath, G.vertices());
     
    // построить пути алгоритмом Дейкстры из вершины 0, содержимое D не используется
    std::vector<float> D(G.vertices());
    Weights_for_graph wfg(G);
    dijkstra(Dpath, D, wfg, 0, G.vertices());

    // стоимость пути
    std::cout << "\nDijkstra path = " << path_weight(wfg, Dpath, G.vertices() - 1) << std::endl;
  }


  /// построить путь алгоритмом A*, отправляя события визуализации в очередь
  void make_Apath()
  {
    Q.clear();
    const auto N = G.vertices(), goal = N - 1;
    prepare_path(Apath, N);

    Heuristic_for_graph pfg(G, factor);
    Weights_for_graph wfg(G);

    float minscore = std::numeric_limits<float>::max()
        , maxscore = 0.0f;

    Astar_demo demo(Q, minscore, maxscore);

    // построить путь из 0 в (G.vertices()-1)
    switch (current_algorithm)
    {
      case normalAstar:
        astar(Apath, wfg, G, 0, goal, N, pfg, demo);
        break;
      case bidirectionalAstar:
      default:
        astar2(Apath, wfg, G, 0, goal, N, pfg, demo);
        break;
    }

    // отобразить оценки стоимости в [0, 1] для событий открытия вершин
    float delta = maxscore - minscore;
    if (delta == 0.0f)
    {
      delta = minscore;
      minscore *= 0.5f;
    }

    for (auto &p: Q)
      if (auto *e = dynamic_cast<Opened_event*>(p.get()))
        e->adjust(minscore, delta);

    // события "прохождения пути"
    for (Vertex_index v = goal;;)
    {
      Q.emplace_back(new Path_event(v));
      if (v == Apath[v])
        break;
      v = Apath[v];
    }

    // стоимость пути
    std::cout << "A* path = " << path_weight(wfg, Apath, goal) << std::endl;
  }
};
