﻿/*
* This file is part of AV Orchid.
* Copyright (c) 2016, Dmitri R. Kuvshinov <evetro.here@gmail.com>
*
* Permission to use, copy, modify, and/or distribute this software for
* any purpose with or without fee is hereby granted, provided that the
* above copyright notice and this permission notice appear in all copies.
*
* THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
* WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
* MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
* ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
* WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
* ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
* OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/
/**
* @file Plotting2D.cpp
* @brief реализация интерфейса рисования графика
*/
#include <GL/freeglut.h>

#include "Plotting2D.hpp"
#include <avo/Camera2.hpp>
#include <avo/CoordinateGrid2.hpp>
#include <avo/Prim2.hpp>
#include <avo/TextDrawer.hpp>
#include <avo/Timer.hpp>

#include <cassert>
#include <string>
#include <thread>
#include <mutex>
#include <condition_variable>


namespace
{

  using Mutex = std::mutex;
  using Lock = std::unique_lock<Mutex>;
  using Bbox_2 = avo::Bbox_2<double>;
  using Prim_2 = avo::Prim_2<double>;

  // rt == rendering thread
  namespace rt
  {

    enum Command
    {
      Do_none,                  // нет команд
      Do_exit,                  // закончить работу
      Do_setup_plotting_window, // изменить параметры окна
      Do_start_plotting,        // начать новый график
      Do_plotting_view,         // установить новые координаты отображаемой области
      Do_clear_plotting,        // удалить графики
      Do_plotting_view_all,     // сбросить область отображения (как при нажатии Home)
    } command = Do_none;

    bool is_done, already_closed;
    Mutex mtx;
    std::thread thr;
    std::condition_variable cv;

    void body(); // функция, выполняемая потоком thr
    bool check_command(); // функция, выполняющая команду

    struct Detacher
    {
      ~Detacher()
      {
        {
          Lock lock(mtx);
          command = Do_exit;
        }

        if (thr.joinable())
          thr.detach();
      }
    } detacher;


    namespace data
    {
      // setup_plotting_window
      std::string title;
      int window_width, window_height;

      // start_plotting
      Color vertex, edge;
      float vertex_px, edge_px;

      // plotting_view
      Bbox_2 view_box;

      // plot
      std::vector<Vec_2> positions;
    }


    bool launch()
    {
      if (already_closed)
        return false;
      // Если поток не был запущен, запустить его.
      if (thr.get_id() == std::thread::id())
        thr = std::thread(body);
      return true;
    }


    bool perform(Command cmd)
    {
      Lock lock(mtx);
      if (!launch())
        return false;

      is_done = false;
      command = cmd;

      // Ждём уведомления о том, что команда выполнена.
      cv.wait(lock, [] { return is_done || already_closed; });
      return !already_closed;
    }

  }

  // Данные сцены
  namespace scene
  {

    // Графики
    struct Graph
    {
      Prim_2 prim;
      Color vertex, edge;
      float vertex_px, edge_px;

      void draw(int mode) const
      {
        if (mode & 1)
        {
          edge.select();
          glLineWidth(edge_px);
          prim.paint(GL_LINE_STRIP);
        }

        if (mode & 2)
        {
          vertex.select();
          glPointSize(vertex_px);
          prim.paint(GL_POINTS);
        }
      }
    };

    // Графики
    std::vector<Graph> graphs;

    void new_graph()
    {
      float def_edge_px = 1.f, def_vertex_px = 1.f;
      if (!graphs.empty())
      {
        def_edge_px = graphs.back().edge_px;
        def_vertex_px = graphs.back().vertex_px;
      }

      if (graphs.empty() || !graphs.empty() && !graphs.back().prim.vertices().empty())
        scene::graphs.emplace_back();

      auto &g = scene::graphs.back();
      g.edge = rt::data::edge;
      g.vertex = rt::data::vertex;
      g.edge_px = rt::data::edge_px > 0.f ? rt::data::edge_px : def_edge_px;
      g.vertex_px = rt::data::vertex_px > 0.f ? rt::data::vertex_px : def_vertex_px;
    }

    // Вспомогательные данные
    Bbox_2 view_box;      // область отображения
    avo::Camera_2 camera; // камера
    avo::Coordinate_grid_2<double> cgrid;  // координатная сетка
    avo::Bitmap_text_drawer_2 text_drawer; // объект, выводящий текст
    int mouse_x, mouse_y; // координаты мыши

    int cur_wnd_height;  // текущая высота окна
    int plot_mode = 3;   // текущий режим рисования графиков: 1 "линии", 2 "точки", 3 "линии и точки"

  }


  //
  // колбеки FreeGlut

  /// обновить координатную сетку при изменении камеры
  void camera_changed()
  {
    scene::cgrid.update(scene::camera.box());
    scene::text_drawer.set_coordinates(scene::camera.box());
    scene::camera.apply();
    glutPostRedisplay();  // уведомить FreeGlut, что следует перерисовать сцену
  }

  /// приблизить камеру
  void camera_zoom_in()
  {
    scene::camera.zoom_in(0.125);
    camera_changed();
  }

  /// "отодвинуть" камеру
  void camera_zoom_out()
  {
    scene::camera.zoom_out(0.125);
    camera_changed();
  }

  /// передвинуть камеру
  void camera_move(const Vec_2 &delta)
  {
    scene::camera.move(delta * scene::camera.box().extents());
    camera_changed();
  }

  /// показать всю занятую область
  void camera_reset()
  {
    if (scene::view_box != scene::camera.box())
    {
      scene::camera.reset(scene::view_box);
      camera_changed();
    }
  }


  /// установить координаты мыши
  void mouse_position(int x, int y)
  {
    scene::mouse_x = x;
    scene::mouse_y = y;
  }

  /// отображение координат и легенды графиков
  void draw_text()
  {
    using std::to_string;
    auto smaller_box = scene::camera;
    smaller_box.zoom_in(0.05); // чтобы не пропадали надписи на краях при округлении "наружу"
    auto &box = smaller_box.box();

    scene::text_drawer.font(GLUT_BITMAP_HELVETICA_10);
    // размеры координатной сетки
    {
      // сгенерировать строки
      const auto
        left = to_string(box.pmin.x),
        bottom = to_string(box.pmin.y),
        right = to_string(box.pmax.x),
        top = to_string(box.pmax.y);

      // вывести текст средствами FreeGlut
      const auto center = 0.5 * (box.pmax + box.pmin);

      avo::LIGHT_GREY.select();
      scene::text_drawer.view(box.pmin.x, center.y).put(left)
        .view(center.x, box.pmin.y).put(bottom)
        .view(box.pmax.x, center.y).put(right, avo::Alignment::toLeft)
        .view(center.x, box.pmax.y).put(top, avo::Alignment::toDown)
        ;
    }

    // координаты мыши
    {
      double mx, my; // "координаты" мыши
      scene::text_drawer.pixels_to_view(scene::mouse_x, scene::mouse_y, mx, my);

      // сгенерировать строки
      const auto xy = to_string(mx) + '\n' + to_string(my);

      // вывести текст средствами FreeGlut
      avo::AZURE.select();
      scene::text_drawer.view(box.pmin.x, box.pmax.y).put(xy, avo::Alignment::toDown);
    }
  }


  // Выход из цикла (закрытие окна).
  void post_leave_event_loop()
  {
    rt::already_closed = true;
    rt::cv.notify_one();
    glutLeaveMainLoop();
  }

  // Событие: перерисовка.
  void on_repaint()
  {
    glClear(GL_COLOR_BUFFER_BIT);
    scene::cgrid.paint();

    for (auto &graph : scene::graphs)
      graph.draw(scene::plot_mode);

    draw_text();
    glutSwapBuffers();

    rt::check_command();
  }


  // Событие: изменён размер окна.
  void on_reshape(int w, int h)
  {
    scene::camera.force_aspect((double)w / h);
    camera_changed();
    scene::text_drawer.set_viewport(w, h);
    scene::cur_wnd_height = h;
    // уведомить OpenGL о новых размерах области рисования
    glViewport(0, 0, w, h);
  }


  // Событие: нажатие "специальной" клавиши
  void on_special_key(int key, int x, int y)
  {
    Vec_2 delta; // 0, 0
    switch (key)
    {
    case GLUT_KEY_LEFT:   delta.x = -0.1; break;
    case GLUT_KEY_UP:     delta.y = +0.1; break;
    case GLUT_KEY_RIGHT:  delta.x = +0.1; break;
    case GLUT_KEY_DOWN:   delta.y = -0.1; break;
    case GLUT_KEY_HOME:   camera_reset(); return;
    default:
      return; // ничего не делать, если была нажата какая-либо другая клавиша
    }

    camera_move(delta);
  }

  // Событие: нажатие "обычной" клавиши
  void on_key(unsigned char key, int x, int y)
  {
    switch (key)
    {
    case 27: // код ESC (ASCII) --  выйти из цикла обработки событий
      return post_leave_event_loop();

    case '\b': // переключение режима рисования
      if (++scene::plot_mode == 4)
        scene::plot_mode = 1;
      glutPostRedisplay();
      break;

    case '+': case 'z': case 'Z': camera_zoom_in();   break;
    case '-': case 'x': case 'X': camera_zoom_out();  break;
    case 'w': case 'W': on_special_key(GLUT_KEY_UP, x, y); break;
    case 's': case 'S': on_special_key(GLUT_KEY_DOWN, x, y); break;
    case 'a': case 'A': on_special_key(GLUT_KEY_LEFT, x, y); break;
    case 'd': case 'D': on_special_key(GLUT_KEY_RIGHT, x, y); break;
    default:
      return; // do nothing for other keys
    }
  }


  // Событие: движение мыши (чтобы отображать её координаты).
  void on_hover(int x, int y)
  {
    mouse_position(x, scene::cur_wnd_height - y);
    glutPostRedisplay();
  }

  // Событие: масштабирование колёсиком мыши.
  void on_mouse_wheel(int wheel, int direction, int x, int y)
  {
    if (direction < 0)
      camera_zoom_in();
    else if (direction > 0)
      camera_zoom_out();
  }


  // Когда нет событий, по возможности не грузим процессор зря :)
  void on_idle()
  {
    if (!rt::check_command())
      avo::sleep(0.03); // <= 33 fps
    glutPostRedisplay();
  }


  namespace rt
  {
    void body()
    {
      // FreeGlut init
      int argc = 1;
      char *argv[] = { "noname", nullptr };
      glutInit(&argc, argv);
      glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_MULTISAMPLE);
      glutSetOption(GLUT_ACTION_ON_WINDOW_CLOSE, GLUT_ACTION_GLUTMAINLOOP_RETURNS);

      glutInitWindowPosition(0, 0);
      glutInitWindowSize(600, 600);
      glutCreateWindow("");

      glClearColor(0.0f, 0.0f, 0.07f, 1.0f);

      glEnable(GL_BLEND); // прозрачность
      glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

      glEnable(GL_COLOR_MATERIAL);
      glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);

      // включить массивы вершин (OpenGL 1.1)
      glEnableClientState(GL_VERTEX_ARRAY);

      // регистрация колбеков
      glutDisplayFunc     (on_repaint);
      glutReshapeFunc     (on_reshape);
      glutKeyboardFunc    (on_key);
      glutSpecialFunc     (on_special_key);
      glutIdleFunc        (on_idle);

      glutPassiveMotionFunc(on_hover);
      glutMouseWheelFunc  (on_mouse_wheel);

      camera_reset();
      glutMainLoop();
      
      // работа окончена
      rt::already_closed = true;
      rt::cv.notify_one();
    }


    // Проверить наличие команды и выполнить её, если есть.
    bool check_command()
    {
      Lock lock(rt::mtx);

      // Добавить точки, если есть.
      if (!rt::data::positions.empty())
      {
        if (scene::graphs.empty())
          scene::new_graph();

        auto &v = scene::graphs.back().prim.vertices();
        auto &p = rt::data::positions;

        scene::view_box.grow(Bbox_2(p.begin(), p.end()));
        v.insert(v.end(), p.begin(), p.end());
        p.clear();
        return true;
      }

      // Выполнить команду.
      bool do_exit = false;
      switch (rt::command)
      {
      case rt::Do_none:
        return false;

      case rt::Do_exit:
        do_exit = true;
        break;

      case rt::Do_setup_plotting_window:
        glutSetWindowTitle(rt::data::title.c_str());
        glutReshapeWindow(rt::data::window_width, rt::data::window_height);
        break;

      case rt::Do_start_plotting:
        scene::new_graph();
        break;

      case rt::Do_plotting_view_all:
        camera_reset();
        break;

      case rt::Do_plotting_view:
        scene::view_box = rt::data::view_box;
        camera_reset();
        break;

      case rt::Do_clear_plotting:
        scene::view_box = Bbox_2();
        scene::graphs.clear();
        break;

      default:
        assert(!"Plotting2D: impossible command");
      }

      rt::command = rt::Do_none;
      rt::is_done = true;

      lock.unlock();
      rt::cv.notify_one();

      if (do_exit)
        post_leave_event_loop();
      return true;
    }
  }

}


// Создать окно, подготовить внутренние структуры данных для начала рисования.
bool setup_plotting_window(const char *title, int window_width, int window_height)
{
  rt::data::title = title;
  rt::data::window_width = window_width;
  rt::data::window_height = window_height;
  return perform(rt::Do_setup_plotting_window);
}

// Начать рисование нового графика.
bool start_plotting(Color vertex, float vertex_px, Color edge, float edge_px)
{
  rt::data::vertex = vertex;
  rt::data::edge = edge;
  rt::data::vertex_px = vertex_px;
  rt::data::edge_px = edge_px;
  return perform(rt::Do_start_plotting);
}

// Получить область отображения всего графика.
bool plotting_view_all()
{
  return perform(rt::Do_plotting_view_all);
}

// Установить координаты отображаемой области.
bool plotting_view(const Vec_2 &pmin, const Vec_2 &pmax)
{
  rt::data::view_box = Bbox_2(pmin, pmax);
  return perform(rt::Do_plotting_view);
}

// Передать следующую точку для размещения в графике.
bool plot(const Vec_2 &pos)
{
  Lock lock(rt::mtx);
  if (!rt::launch())
    return false;

  rt::data::positions.push_back(pos);
  return true;
}

// Удалить все графики.
bool clear_plotting()
{
  return perform(rt::Do_clear_plotting);
}

// Проверить, закрыто ли окно.
bool is_already_closed()
{
  return rt::already_closed;
}
