﻿/*
 * This file is part of AV Orchid.
 * Copyright (c) 2013, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * Пример: Collision3D
 * Автор: Кувшинов Д.Р. https://evetro.wordpress.com/
 * Используемые библиотеки: FreeGlut, OpenGL.
 *
 * Симуляция движения и столкновения тяжёлых шаров в пространстве.
 * Шары притягиваются друг к другу под действием классического закона тяготения.
 * Нажатие SPACE/ENTER создаёт новые шары.
 * Нажатие ESC/Q -- выход из программы.
 * ADZX и стрелки -- поворот камеры.
 * WS -- приближение, удаление камеры.
 * IJKLOP -- перемещение камеры в трёх осях.
 */

// TASK: Реализовать алгоритм компенсационного суммирования в виде класса, 
// объекты которого имитируют число с плавающей точкой, 
// но при добавлении или вычитании значений используют компенсацию.
// Компенсационное суммирование: https://ru.wikipedia.org/wiki/Алгоритм_Кэхэна
// Внедрить его в процесс накопления ускорений и последующего интегрирования.
// Проанализировать результат.

#include <GL/freeglut.h>
#include <cstdlib>        // rand/srand
#include <iostream>       

#include <avo/Basics.hpp>
#include <avo/Timer.hpp>
#include <avo/Vec3.hpp>     // 3-мерный вектор
#include <avo/Color.hpp>    // RGBA-цвет в 128-битном формате (компоненты типа float)
#include <avo/SphericalCamera.hpp> // камера в сферических координатах ("турельного" типа)

using avo::Color;
using avo::Timer;
using avo::Vec_3d;
using avo::random;

//
// базовые определения
//

/* 
 * Вектор и камера вынесены в отдельные файлы, так что в этой секции размещён только Ball.
 * Замечу, что здешний Vec3 значительно "навороченнее", чем Vec_2, так как взят из другого проекта.
 * В данном примере уровень использования функциональности Vec3 минимален.
 */

/// твёрдый тяжёлый шар
struct Ball
{
  Vec_3d p;      ///< текущая позиция
  Vec_3d v;      ///< текущая скорость
  Vec_3d a_old;  ///< старое ускорение (для метода Velocity Verlet)
  Vec_3d a;      ///< накопленное ускорение (сбрасывается при вычислении следующей позиции)
  double m;     ///< масса
  double r;     ///< радиус
  Color color;  ///< цвет (учитывается значение альфа-компонента -- в качестве коэффициента прозрачности)

  /// базовый конструктор (вначале ускорение нулевое)
  explicit Ball
    (
      const Vec_3d &pos  = Vec_3d(),
      const Vec_3d &vel  = Vec_3d(),
      double       mass = 1.0, 
      double       rad  = 10.0,
      const Color &col  = avo::YELLOW
    )
      : p(pos), v(vel), m(mass), r(rad), color(col), a(), a_old()
  {}

  /// нарисовать шар
  void paint() const
  {
    glLoadIdentity();   // установить единичную матрицу в качестве матрицы модели
    glTranslated(p.x, p.y, p.z); // домножить матрицу модели на матрицу перемещения в (p.x, p.y, p.z)
    glScaled(r, r, r);  // домножить матрицу модели на матрицу масштабирования в r раз

    color.select(); // выставить нужный цвет
    glCallList(DISPLAY_LIST);  // вызвать список отрисовки (модель привязана к списку отрисовки)
  }

  /// вычислить позицию шара через time_step секунд
  void move(double time_step)
  {
    // используется метод Velocity Verlet: http://en.wikipedia.org/wiki/Velocity_Verlet#Velocity_Verlet 
    const auto half_step = 0.5 * time_step;
    const Vec_3d ha = half_step * a;
    v += half_step * a_old + ha;
    p += time_step * (v + ha);
    a_old = a;
    a = Vec_3d(); // обнулить a
  }

  /// список отрисовки (display list), отвечающий шару
  static GLuint DISPLAY_LIST;
};

/// вывести текстовое представление шара в поток вывода
std::ostream& operator<<(std::ostream &os, const Ball &c)
{
  return os << "position: "    << c.p
            << "\nvelocity: "  << c.v
            << "\nmass: "      << c.m
            << "\nradius: "    << c.r   << std::endl;
}


//
// константы
// 

/// общее количество шаров
const size_t N = 100;

/// коэффициент восстановления при ударе
const double k = 0.8;

/// гравитационная постоянная
const double G = 4.0;

/// коэффициент "вязкого" трения (пропорционально квадрату радиуса и скорости)
const double F = 0.0;


//
// глобальные переменные
//

/// собственно гравитирующие тела (шары)
Ball C[N];

/// таймер, используемый для замера временного шага между кадрами
Timer<> timer;

/// счётчик кадров
int frame;

/// стартовый импульс
Vec_3d total_impulse;

/// объект камеры
avo::Spherical_camera camera;

// static-данные структуры/класса требуется явно размещать в некоторой единице трансляции (.cpp файле)
GLuint Ball::DISPLAY_LIST;


//
// вспомогательные функции
//

inline Vec_3d random3(double x0, double x1, double y0, double y1, double z0, double z1)
{
  return Vec_3d(random(x0, x1), random(y0, y1), random(z0, z1));
}

/// сгенерировать новый набор псевдослучайных шаров
void new_balls()
{
  std::cout << "\nGenerating new balls:\n";

  // солнце
  {
    Ball &b = C[0];
    b.p = Vec_3d();
    b.v = Vec_3d();
    b.r = 30.;
    b.m = 0.;
    b.color = avo::WHITE;
  }

  total_impulse = Vec_3d();
  for (size_t i = 1; i < N; ++i)
  {
    Ball &b = C[i]; // заполним поля C[i]
    // кинематические параметры шара
    b.p = random3(-600.0, 600.0, -600.0, 600.0, -600.0, 600.0); // рандомная позиция
    // вращение вокруг начала координат
    b.v = cross(b.p, random3(-100.0, 100.0, -100.0, 100.0, -100.0, 100.0)) / b.p.length();
    
    // свойства шара
    b.r = random(4.0, 20.0); // radius
    b.m = random(2.0, 6.0) * b.r * b.r * b.r; // масса, случайная "плотность" вещества

    total_impulse += b.m * b.v;
    C[0].m += b.m;
    
    // случайный цвет
    b.color.red   = random(0.f, 1.f);
    b.color.green = random(0.f, 1.f);
    b.color.blue  = random(0.f, 1.f);
    // прозрачность
    b.color.alpha = random(0.3f, 1.f);

    if (N < 100)
      std::cout << i << ")\n" << b;
  }

  C[0].m *= 3.; // солнце тяжелее всех планет вместе взятых в 3 раза
  std::cout << "sun mass == " << C[0].m << std::endl;
}


/// вычислить и применить гравитационное притяжение для заданной пары шаров
inline void gravity(Ball &b1, Ball &b2)
{
  Vec_3d r = b2.p - b1.p;
  auto d = r.length();
  
  // притяжение работает только для непересекающихся шаров
  const auto r_sum = b1.r + b2.r;
  if (d < r_sum) return;

  r *= G / (d * d * d);
  b1.a += b2.m * r;
  b2.a -= b1.m * r;
}


/// применить "сопротивление вязкой среды" (пропорционально скорости и площади фронтальной проекции)
inline void friction(Ball &b)
{
  b.a -= F * b.r * b.r / b.m * b.v;
}


/// проверить наличие и рассчитать столкновение двух шаров
void collide(Ball &a, Ball &b)
{
      Vec_3d dp = b.p - a.p;
  const auto dplen = dp.length(),
             rsum = a.r + b.r;

  if (dplen <= rsum) // есть ли столкновение?
  {
    // следует ли пересчитывать скорости? (двигаются ли шары навстречу друг другу)
    if (0.0 < dot(dp, a.v - b.v)) // dot -- скалярное произведение
    {
      const auto inv_mass = 1.0 / (a.m + b.m);
      const Vec_3d impulse = a.m * a.v + b.m * b.v;
      const Vec_3d elastic_v = k * (a.v - b.v);
      a.v = inv_mass * (impulse - b.m * elastic_v);
      b.v = inv_mass * (impulse + a.m * elastic_v);
    }

    // необходимо развести кружки, чтобы они перестали контактировать
    if (dplen != 0.0)
    {
      const auto dr = 1.00000001 * rsum - dplen;
      if (dr > 0.0)
      {
        // разведём по вектору AtoB в пропорции, отвечающей импульсам тел
        const auto impulseA = a.m * a.v.length(),
                   impulseB = b.m * b.v.length(),
                   alpha = impulseA / (impulseA + impulseB);

        dp *= dr / dplen;
        // линейная интерполяция
        a.p -= (1.0 - alpha) * dp;
        b.p +=        alpha  * dp;
      }
    }
    else // центры шаров совпали -- патологическая ситуация
    {
      // развести как угодно, например, по 0x
      const auto dx = 0.5 * rsum;
      if (a.p.x < b.p.x)
        a.p.x -= dx, b.p.x += dx;
      else
        a.p.x += dx, b.p.x -= dx;
    }
  }
}


//
// обработчики событий GLUT
//

/// изменение размеров окна, функция получает "новые размеры"
void reshape(int w, int h)
{
  // новая матрица проекции
  glMatrixMode(GL_PROJECTION);
  glPopMatrix();    // удалить предыдущую матрицу из стека матриц (стек матриц используется Spherical_camera)
  glLoadIdentity(); // загрузить единичную матрицу
  /* угол зрения по вертикали, соотношение сторон экрана, минимальная дальность, максимальная дальность */
  gluPerspective(75, double(w) / double(h), 16.0, 8192.0); // домножить на матрицу перспективной коррекции
  glPushMatrix();   // затолкнуть текущую матрицу в стек матриц (копируется, текущая остаётся на вершине)
  camera.apply();   // применить к текущей матрице матрицу камеры

  // установить координаты окна просмотра (вся клиентская область)
  glViewport(0, 0, w, h);
}


/// перерисовка
void repaint()
{
  // очистить окно, закрасив его цветом очистки, сбросить буфер глубины
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  // нарисовать все шары
  for (const auto &c : C)
    c.paint();
  glutSwapBuffers(); // обменять буферы кадра, чтобы вывести в окно новый кадр (двойная буферизация)
}


/// вычислить состояние сцены на следующий кадр
void frame_move()
{
  const auto MAX_STEP = 0.05;
  const int INTEGRATOR_ROUNDS = 4;

  double h = timer.step(); // время, прошедшее с предыдущего кадра
  if (h <= 0.0) return;
  // будем "тормозить", если компьютер не справляется, ограничив сверху шаг по времени
  if (h > MAX_STEP) h = MAX_STEP;

  for (int i = 0; i < N; ++i) // "треугольный" цикл
  {
    Ball &Ci = C[i];
    for (size_t j = i + 1; j < N; ++j)
      collide(Ci, C[j]); // проверить и рассчитать столкновение
  }

  // делаем дополнительные шаги интегрирования для данного временного шага (это делать необязательно)
  h /= INTEGRATOR_ROUNDS;

  // применить силы,
  // здесь бы могли пригодиться какие-нибудь продвинутые алгоритмы
  // либо для ускорения работы, либо для улучшения качества (например, компенсационное суммирование)
  for (int round = 0; round < INTEGRATOR_ROUNDS; ++round)
  {
    // вычисляем новые ускорения шаров
    for (int i = 0; i < N; ++i)
    {
      Ball &Ci = C[i];
      for (size_t j = i + 1; j < N; ++j)
        gravity(Ci, C[j]);
      friction(Ci);
    }

    // интегрируем ускорения, получая новые позиции
    for (auto &c : C)
    {
      c.move(h);
      if (c.p.dotSquare() > 1e+9) // если шар улетел слишком далеко от начала координат,
      {                           // то возвращаем его в окрестность начала координат
        total_impulse -= c.m * c.v;
        c.p = random3(-300.0, 300.0, -300.0, 300.0, -300.0, 300.0);
        c.v = Vec_3d();            // с нулевой скоростью
      }
    }
  }

  // статистика
  // в ti накапливается суммарное количество движения (total impulse -- "полный импульс")
  Vec_3d ti;
  // вычисляем и выводим полный импульс каждый 512-й кадр
  if ((frame++ & 511) == 0)
  {
    for (const auto &c : C)
      ti += c.m * c.v;
    std::cout << "imp. rel. err. = " << (ti - total_impulse).length() / total_impulse.length() << std::endl;
  }

  // уведомим GLUT о необходимости перерисовать сцену
  glutPostRedisplay();
}


/// обработать нажатие "обычной" клавиши (x, y -- координаты мыши в окне на момент нажатия)
void on_key(unsigned char key, int x, int y)
{
  switch (key)
  {
  case 'w': case 'W': // ближе к началу координат
    camera.forward();
    break;
  case 's': case 'S': // дальше от начала координат
    camera.backward();
    break;
  case 'a': case 'A': // повернуть влево
    camera.left();
    break;
  case 'd': case 'D': // повернуть вправо
    camera.right();
    break;
  case 'z': case 'Z': // повернуть вверх
    camera.up();
    break;
  case 'x': case 'X': // повернуть вниз
    camera.down();
    break;

  // IJKL-OP for moving origin
  case 'i': case 'I': // передвинуть начало координат вперёд
    camera.origin_forward(100.);
    break;
  case 'k': case 'K': // передвинуть начало координат назад
    camera.origin_forward(-100.);
    break;
  case 'j': case 'J': // передвинуть начало координат влево
    camera.origin_right(-70.);
    break;
  case 'l': case 'L': // передвинуть начало координат вправо
    camera.origin_right(70.);
    break;
  case 'o': case 'O': // передвинуть начало координат вверх
    camera.origin_up(70.);
    break;
  case 'p': case 'P': // передвинуть начало координат вниз
    camera.origin_up(-70.);
    break;

  case 27: case 'q': case 'Q': case 'й': case 'Й':
    glutLeaveMainLoop(); // покинуть цикл обработки сообщений
    return; // и покинуть данную функцию...

  case ' ': case '\r': case '\n':
    new_balls(); // сделать новые шарики
    break;
  
  default:
    return; // остальные игнорируются
  }

  camera.apply(); // применить матрицу камеры к матрице проекции
  glutPostRedisplay();
}


/// обработать нажатие "специальной" клавиши (x, y -- координаты мыши в окне на момент нажатия)
void on_special_key(int key, int x, int y)
{
  // движение камеры аналогично WASD, только с помощью стрелок на клавиатуре
  switch (key)
  {
  case GLUT_KEY_LEFT:
    camera.left();
    break;
  case GLUT_KEY_UP:
    camera.up();
    break;
  case GLUT_KEY_RIGHT:
    camera.right();
    break;
  case GLUT_KEY_DOWN:
    camera.down();
    break;
  default:
    return; // остальные игнорируются
  }

  camera.apply(); // применить матрицу камеры к матрице проекции
  glutPostRedisplay();
}


//
// main
//

int main(int argc, char **argv)
{
  // базовая инициализация
  glutInit(&argc, argv);  // обработка параметров командной строки
  std::srand((unsigned)(avo::seconds() * 1e+6) % 15485863);

  // создать окно с привязанным контекстом OpenGL
  // режим:
  glutInitDisplayMode(GLUT_RGBA        // 32-битный цвет RGBA,
                    | GLUT_ALPHA       // поддержка альфа-канала для имитации прозрачности
                    | GLUT_DEPTH       // поддержка буфера глубины
                    | GLUT_MULTISAMPLE // сглаживание мультисемплингом (FS MSAA)
                    | GLUT_DOUBLE);    // двойная буферизация
  glutInitWindowPosition(0, 0);   // координаты верхнего левого угла окна на экране
  glutInitWindowSize(800, 600);   // полные ширина и высота окна
  glutCreateWindow("Ball collision simulation"); // создать окно, в функцию передаётся заголовок создаваемого окна

  // параметры OpenGL для данного примера
  glClearColor(0.0, 0.0, 0.0, 1.0); // установить цвет очистки (фон) -- чёрный, непрозрачный
  glEnable(GL_COLOR_MATERIAL);      // выбирать цвет материала командой glColor
  glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE); // этот цвет -- с обеих сторон полигона, ambient и diffuse компоненты

    glEnable(GL_BLEND); // включить интерпретацию альфа канала как коэффициента смешивания цветов (для имитации прозрачности)
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA); // выбрать формулу смешивания
    /* sfactor == GL_SRC_ALPHA -- домножить цвет нового пикселя на его альфу, 
       dfactor == GL_ONE_MINUS_SRC_ALPHA -- домножить цвет старого пикселя на (1 - альфа нового),
       результирующий цвет получается сложением этих двух цветов */

  glEnable(GL_CULL_FACE);  // включить отбрасывание полигонов, смотрящих "от" камеры (повёрнутых обратной стороной)
  glEnable(GL_DEPTH_TEST); // использовать буфер глубины по назначению
  //glEnable(GL_DEPTH_CLAMP);
  glEnable(GL_NORMALIZE);  // форсировать нормирование нормалей (при масштабировании сфер происходит масштабирование и нормалей)
  glEnable(GL_LIGHTING);   // включить повершинное освещение
  glEnable(GL_LIGHT0);     // включить источник света №0

  // сгенерировать модель шара (полигональная аппроксимация сферы)
  Ball::DISPLAY_LIST = glGenLists(1);        // получить идентификатор нового списка отрисовки
  glNewList(Ball::DISPLAY_LIST, GL_COMPILE); // инициализировать этот список (режим "компиляции" команд), начать запись команд
  glutSolidSphere(1.0, 16, 8);       // выполнить рисование "кружка" (16 разбиений по "параллелям", 8 разбиения по "меридианам")
  glEndList();                       // закончить запись команд, скомпилировать список

  // привязка функций-обработчиков событий
  glutIdleFunc(frame_move);            // обработка движения
  glutDisplayFunc(repaint);            // перерисовка
  glutReshapeFunc(reshape);            // изменение размеров окна
  glutKeyboardFunc(on_key);         // обработка нажатий "обычных" клавиш
  glutSpecialFunc(on_special_key);  // обработка нажатий "специальных" клавиш

  // старт
  std::cout << "Starting Collision...\n" << "k = " << k << "\n";
  new_balls();
  timer.start();  // инициализация таймера
  glutMainLoop(); // вход в цикл обработки событий
  std::cout << "\nbye!";
  return 0;
}
