﻿/*
 * This file is part of AV Orchid.
 * Copyright (c) 2014, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * Пример: ColorClock
 * Автор: Кувшинов Д.Р. https://evetro.wordpress.com/
 * Используемые библиотеки: FreeGlut, OpenGL.
 *
 * Цветные "электронные" часы. Показывают системное время, цвет зависит от времени.
 * Особенности кода для Windows введены для того, чтобы не создавать окна консоли.
 */
#ifdef _MSC_VER
#define _CRT_SECURE_NO_WARNINGS
#endif
#include <GL/freeglut.h>

#include <cmath>   // fmod
#include <sstream>
#include <iomanip> // put_time
#include <chrono>  // дата-время C++11
#include <thread>  // sleep

#include <avo/Color.hpp>
using namespace avo;


//
// логика

float current_hour; // текущий час, в промежутке [0, 24]
float current_hour_shift; // сдвиг времени (для прокрутки цветов)
std::string current_time_string; // время, отображаемое в данный момент

// время в данный момент (абсолютное)
inline std::time_t now()
{
  return std::time(nullptr);
}

// схема получения строкового представления времени в Стандартной библиотеке C++, хм, странная...
// fmt передаётся put_time
std::string format_time(std::time_t t, const char *fmt)
{
  std::ostringstream os;
  os << std::put_time(std::localtime(&t), fmt); // строковое представление t
  return os.str();
}

// выполнять всё время
void frame_move()
{
  // спать ~30мс
  std::this_thread::sleep_for(std::chrono::milliseconds(30));
  // проверить текущее время
  auto formatted_t = format_time(now(), "%H:%M:%S");
  if (current_time_string != formatted_t) // изменилось?
  {
    current_time_string = formatted_t;
    // получить текущий час в плавающей точке
    const auto t  = now();
    const auto ts = std::localtime(&t);
    current_hour = ts->tm_hour + ts->tm_min / 60.f + ts->tm_sec / 3600.f;
    glutPostRedisplay(); // требуется перерисовка
  }
}

/// обработать нажатие "обычной" клавиши (x, y -- координаты мыши в окне на момент нажатия)
void on_key(unsigned char key, int x, int y)
{
  switch(key)
  {
  case 27: case 'q': case 'Q':
    glutLeaveMainLoop(); // покинуть цикл обработки сообщений
    break;
  case '+': // прокрутка часов для просмотра цветов
    if (++current_hour_shift == 25.f) current_hour_shift = 0.f;
    glutPostRedisplay();
    break;
  case '-':
    if (--current_hour_shift == -1.f) current_hour_shift = 24.f;
    glutPostRedisplay();
    break;
  default:
    return; // остальные коды игнорируются
  }
}


// перерисовка
void repaint()
{
  // цвет фона на каждый час
  static const Color background[25] =
  {
    { 0.2f, 0.0f, 0.5f },  // 0
    { 0.0f, 0.0f, 0.3f },  // 1
    { 0.0f, 0.0f, 0.1f },  // 2
    { 0.0f, 0.1f, 0.2f },  // 3
    { 0.0f, 0.3f, 0.3f },  // 4
    { 0.0f, 0.5f, 0.3f },  // 5
    { 0.0f, 0.6f, 0.4f },  // 6
    { 0.1f, 0.7f, 0.8f },  // 7
    { 0.2f, 0.7f, 1.0f },  // 8
    { 0.3f, 0.8f, 1.0f },  // 9
    { 0.5f, 1.0f, 1.0f }, // 10
    { 0.7f, 1.0f, 1.0f }, // 11
    { 1.0f, 1.0f, 1.0f }, // 12
    { 0.7f, 0.7f, 1.0f }, // 13
    { 0.5f, 0.5f, 1.0f }, // 14
    { 0.6f, 0.5f, 0.7f }, // 15
    { 0.8f, 0.7f, 0.5f }, // 16
    { 1.0f, 0.9f, 0.2f }, // 17
    { 1.0f, 0.7f, 0.1f }, // 18
    { 1.0f, 0.5f, 0.0f }, // 19
    { 1.0f, 0.2f, 0.4f }, // 20
    { 0.6f, 0.0f, 0.8f }, // 21
    { 0.4f, 0.0f, 0.7f }, // 22
    { 0.3f, 0.0f, 1.0f }, // 23
    { 0.2f, 0.0f, 0.5f }   // 0
  };

  // цвет шрифта на каждый час
  static const Color foreground[25] =
  {
    { 1.0f, 1.0f, 0.3f },  // 0
    { 1.0f, 0.6f, 0.0f },  // 1
    { 1.0f, 0.6f, 0.0f },  // 2
    { 1.0f, 0.8f, 0.2f },  // 3
    { 1.0f, 0.8f, 0.3f },  // 4
    { 1.0f, 1.0f, 0.6f },  // 5
    { 1.0f, 1.0f, 0.6f },  // 6
    { 0.8f, 1.0f, 0.7f },  // 7
    { 0.8f, 1.0f, 0.6f },  // 8
    { 0.8f, 1.0f, 0.5f },  // 9
    { 0.5f, 0.0f, 0.0f }, // 10
    { 0.0f, 0.3f, 0.0f }, // 11
    { 0.0f, 0.2f, 0.0f }, // 12
    { 0.0f, 0.2f, 0.0f }, // 13
    { 0.0f, 0.0f, 0.0f }, // 14
    { 0.0f, 0.0f, 0.0f }, // 15
    { 0.0f, 0.0f, 0.0f }, // 16
    { 0.0f, 0.2f, 1.0f }, // 17
    { 0.0f, 0.0f, 0.5f }, // 18
    { 0.0f, 0.3f, 0.6f }, // 19
    { 0.0f, 0.3f, 0.5f }, // 20
    { 0.1f, 0.5f, 0.9f }, // 21
    { 0.6f, 0.7f, 1.0f }, // 22
    { 1.0f, 1.0f, 1.0f }, // 23
    { 1.0f, 1.0f, 0.3f }   // 0
  };

  // сделать фон
  // по поводу метода вычисления цвета,
  // см. http://ru.wikipedia.org/wiki/%D0%9B%D0%B8%D0%BD%D0%B5%D0%B9%D0%BD%D0%B0%D1%8F_%D0%B8%D0%BD%D1%82%D0%B5%D1%80%D0%BF%D0%BE%D0%BB%D1%8F%D1%86%D0%B8%D1%8F
  // (линейная интерполяция для равноотстоящих узлов)
  const float selector = fmod(current_hour + current_hour_shift, 25.f);
  piecewise_sample_array(background, 25, selector).select_clear();
  glClear(GL_COLOR_BUFFER_BIT);

  // выбрать цвет шрифта
  piecewise_sample_array(foreground, 25, selector).select();
  // нарисовать строку
  glLoadIdentity();
  const auto text = (unsigned char*)current_time_string.c_str();
  const auto text_width  = glutStrokeLength(GLUT_STROKE_MONO_ROMAN, text);
  const auto text_height = glutStrokeHeight(GLUT_STROKE_MONO_ROMAN);
  gluOrtho2D(-0.01 * text_width , 1.01 * text_width, -0.1 * text_height, 0.75 * text_height);
  glutStrokeString(GLUT_STROKE_MONO_ROMAN, text);

  glFinish();
  glutSwapBuffers();
}


int main(int argc, char **argv)
{
  // базовая инициализация
  glutInit(&argc, argv);

  // создать окно с привязанным контекстом OpenGL
  // режим:
  glutInitDisplayMode(GLUT_RGBA        // 32-битный цвет RGBA,
                    | GLUT_MULTISAMPLE // сглаживание мультисемплингом (FS MSAA)
                    | GLUT_DOUBLE);    // двойная буферизация
  glutInitWindowPosition(0, 0);    // координаты верхнего левого угла окна на экране
  glutInitWindowSize(500, 140);    // полные ширина и высота окна
  glutCreateWindow("Color Clock"); // создать окно, в функцию передаётся заголовок создаваемого окна

  // параметры OpenGL для данного примера
  glEnable(GL_COLOR_MATERIAL);     // выбирать цвет материала командой glColor
  glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE); // этот цвет -- с обеих сторон полигона, ambient и diffuse компоненты
  glLineWidth(4.f);

  // привязка функций-обработчиков событий
  glutIdleFunc(frame_move); // обработка движения
  glutDisplayFunc(repaint); // перерисовка
  glutKeyboardFunc(on_key); // вызывать для обработки нажатий "обычных" клавиш

  glutMainLoop(); // вход в цикл обработки событий
  return 0;
}
