#version 110

// Noise shader demo.
// The following code was taken from
// https://github.com/ashima/webgl-noise/blob/master/src/classicnoise2D.glsl
// (MIT license).

//
// GLSL textureless classic 2D noise "cnoise",
// with an RSL-style periodic variant "pnoise".
// Author: Stefan Gustavson (stefan.gustavson@liu.se)
// Version: 2011-08-22
//
// Many thanks to Ian McEwan of Ashima Arts for the
// ideas for permutation and gradient selection.
//
// Copyright (c) 2011 Stefan Gustavson. All rights reserved.
// Distributed under the MIT license. See LICENSE file.
// https://github.com/ashima/webgl-noise
//

vec4 mod289(vec4 x)
{
  return x - floor(x * (1.0 / 289.0)) * 289.0;
}

vec4 permute(vec4 x)
{
  return mod289(((x*34.0)+1.0)*x);
}

vec4 taylorInvSqrt(vec4 r)
{
  return 1.79284291400159 - 0.85373472095314 * r;
}

vec2 fade(vec2 t) {
  return t*t*t*(t*(t*6.0-15.0)+10.0);
}

// Classic Perlin noise
float cnoise(vec2 P)
{
  vec4 Pi = floor(P.xyxy) + vec4(0.0, 0.0, 1.0, 1.0);
  vec4 Pf = fract(P.xyxy) - vec4(0.0, 0.0, 1.0, 1.0);
  Pi = mod289(Pi); // To avoid truncation effects in permutation
  vec4 ix = Pi.xzxz;
  vec4 iy = Pi.yyww;
  vec4 fx = Pf.xzxz;
  vec4 fy = Pf.yyww;

  vec4 i = permute(permute(ix) + iy);

  vec4 gx = fract(i * (1.0 / 41.0)) * 2.0 - 1.0 ;
  vec4 gy = abs(gx) - 0.5 ;
  vec4 tx = floor(gx + 0.5);
  gx = gx - tx;

  vec2 g00 = vec2(gx.x,gy.x);
  vec2 g10 = vec2(gx.y,gy.y);
  vec2 g01 = vec2(gx.z,gy.z);
  vec2 g11 = vec2(gx.w,gy.w);

  vec4 norm = taylorInvSqrt(vec4(dot(g00, g00), dot(g01, g01), dot(g10, g10), dot(g11, g11)));
  g00 *= norm.x;
  g01 *= norm.y;
  g10 *= norm.z;
  g11 *= norm.w;

  float n00 = dot(g00, vec2(fx.x, fy.x));
  float n10 = dot(g10, vec2(fx.y, fy.y));
  float n01 = dot(g01, vec2(fx.z, fy.z));
  float n11 = dot(g11, vec2(fx.w, fy.w));

  vec2 fade_xy = fade(Pf.xy);
  vec2 n_x = mix(vec2(n00, n01), vec2(n10, n11), fade_xy.x);
  float n_xy = mix(n_x.x, n_x.y, fade_xy.y);
  return 2.3 * n_xy;
}

// end of file
///////////////////////////////////////////////////////////////////////////////////////////////////
// shader parameters

uniform sampler1D palette;
uniform float t;
uniform float pixel_units_coef;
uniform float width, height;

void main(void)
{
  vec2 p = gl_FragCoord.xy / vec2(width, height) * 10.0;
  // adapts main part from
  // https://github.com/ashima/webgl-noise/blob/master/demo/common/noisedemoMain.frag
    
  // Perturb the texcoords with three components of noise
  vec2 uv = p + 0.1*vec2(cnoise(p + vec2(pixel_units_coef, t)),
		0.5 * (cnoise(p + vec2(43.0, 17.0 + t)) + cnoise(p + vec2(-17.0, -43.0 + t))));
  // Six components of noise in a fractal sum
  float n = cnoise(uv - vec2(0.0, t));
  n += 0.5 * cnoise(uv * 2.0 - vec2(0.0, t*1.4));
  n += 0.25 * cnoise(uv * 4.0 - vec2(0.0, t*2.0));
  n += 0.125 * cnoise(uv * 8.0 - vec2(0.0, t*2.8));
  n += 0.0625 * cnoise(uv * 16.0 - vec2(0.0, t*4.0));
  n += 0.03125 * cnoise(uv * 32.0 - vec2(0.0, t*5.6));
  n = n * 0.7;
  gl_FragColor = texture1D(palette, 0.5 * (1.0 + n));
}
