/*
 * This file is part of AV Orchid.
 * Copyright (c) 2015, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * Пример: FIRFilter
 * Автор: Кувшинов Д.Р. https://evetro.wordpress.com/
 * Используемые библиотеки: FreeGlut, OpenGL.
 *
 * Рисует графики "сигнала" (жёлтый) и его производной (оранжевый),
 * полученный с помощью FIR-фильтра, вычисляющий производную с точностью O(dx4).
 *
 * Нажатие
 *  * Цифра -- выбор функции, порождающей сигнал.
 *  * Space -- пауза.
 *  * Enter -- сброс сцены.
 *  * ESC/Q -- выход из программы.
 */
#include <GL/freeglut.h>
#include "FIRModel.hpp"

#include <avo/Timer.hpp>   // sleep, seconds
#include <avo/Vec2.hpp>    // 2D-вектор
#include <avo/Color.hpp>   // RGBA-цвет
#include <avo/Camera2.hpp> // 2D-"камера"

#include <cassert>
#include <cmath>
#include <vector>


using avo::Vec_2f;
using avo::Color;
using avo::Camera_2;

using Positions = std::vector<Vec_2f>;


/// Двумерный примитив.
struct V2f
{
  Positions positions;  ///< позиции вершин
  Color color;          ///< общий цвет примитива
  GLfloat line_width;   ///< толщина линий примитива
  GLenum mode;          ///< тип примитива

  explicit V2f(GLenum mode = GL_LINE_STRIP, Color color = avo::WHITE, GLfloat line_width = 1.0f)
    : mode(mode), color(color), line_width(line_width)
  {}

  /// Заполнить Y-отсчётами с равным шагом по 0X, начиная с индекса from и до конца.
  /// Возвращает количество записанных значений.
  std::size_t graphic(const float y[], std::size_t y_sz, std::size_t from, float x_step)
  {
    const auto sz = positions.size();
    assert(from < sz);
    const auto
      sz_left = sz - from,
      sz_write = y_sz < sz_left? y_sz: sz_left;

    float x = from != 0? (positions[from - 1].x + x_step): 0.f;
    for (std::size_t i = 0; i < sz_write; ++i)
    {
      positions[i + from] = Vec_2f(x, y[i]);
      x += x_step;
    }

    return sz_write;
  }

  /// Нарисовать.
  void paint() const
  {
    if (!positions.empty())
    {
      color.select();
      glLineWidth(line_width);
      glVertexPointer(2, GL_FLOAT, sizeof(Vec_2f), positions.data());
      glDrawArrays(mode, 0, positions.size());
    }
  }
};


/// Количество отображаемых отсчётов "сигнала" (размер буфера).
const unsigned BUFFER_SIZE = 300;
const float MODEL_TIME_STEP = 0.01f;

/// Структура, собирающая все визуальные "запчасти" вместе.
struct Scene
{
  V2f signal;   ///< последовательность отсчётов сигнала
  V2f filtered; ///< последовательность отсчётов выхода FIR-фильтра
  V2f vertical_grid;   ///< вертикальные линии координатной сетки
  V2f horizontal_grid; ///< горизонтальные линии координатной сетки
  V2f axis;     ///< горизонтальная ось

  FIR_model model; ///< данные модели (сигналы)

  Scene()
    : signal(GL_LINE_STRIP, avo::YELLOW, 5.0f),
      filtered(GL_LINE_STRIP, avo::ORANGE, 5.0f),
      vertical_grid(GL_LINES, avo::AZURE),
      horizontal_grid(GL_LINES, avo::AZURE),
      axis(GL_LINES, avo::WHITE, 3.0f),
      model(BUFFER_SIZE, MODEL_TIME_STEP)
  {
    signal.positions.reserve(BUFFER_SIZE);
    filtered.positions.reserve(BUFFER_SIZE);
    axis.positions.resize(2);
  }

  /// обновить время
  void update(float t)
  {
    const auto new_samples = model.next_time(t);

  }

  /// нарисовать сцену (переадресация geometry)
  void paint() const
  {
    vertical_grid.paint();
    horizontal_grid.paint();
    axis.paint();
    signal.paint();
    filtered.paint();
  }
}
scene; ///< глобальный объект сцены

float signal[BUFFER_SIZE]; ///< глобальный массив "сигнал"
float filtrs[BUFFER_SIZE]; ///< результат работы фильтра

float (*signal_function)(float); ///< текущая функция-генератор "сигнала"
const float horizontal_step = 0.01f; ///< смещение по горизонтали за один кадр

bool paused = false; ///< пауза

avo::Camera_2 camera;  ///< глобальный объект "камеры"
float left_border = 0; ///< x-координата левого края видимой области


/// Тестовая функция: "периодический" x^2.
float x2_saw(float x)
{
  const float x_mod = std::fmodf(x * 0.5f, 1.f);
  return 2.0f * x_mod * x_mod - 1.0f;
}

/// Тестовая функция: "периодический" модуль.
float pabs(float x)
{
  const float x_mod = std::fmodf(x, 1.f) - 0.5f;
  return std::abs(x_mod);
}


/// FIR-фильтр: свёртка пяти значений, производная, O(h4), сдвиг на два отсчёта.
template <class Float, class InIt, class OutIt>
OutIt derivative_4(Float step, InIt from, InIt to, OutIt dest)
{
  const Float
    bx = Float(1) / (12 * step),
    by = Float(2) / (3 * step);

  if (from == to) return dest;
  auto f0 = *from;
  if (++from == to) return dest;
  auto f1 = *from;
  if (++from == to) return dest;
  auto f2 = *from;
  if (++from == to) return dest;
  auto f3 = *from;
  if (++from == to) return dest;
  auto f4 = *from;
  
  while (true)
  {
    *dest = bx * (f0 - f4) - by * (f1 - f3);
    ++dest;

    if (++from == to)
      return dest;

    f0 = f1;
    f1 = f2;
    f2 = f3;
    f3 = f4;
    f4 = *from;
  }
}



/// Сброс сцены на начальное состояние.
void reset_scene()
{
  const float vert_bound = 1.2f, horz_bound = 3.6f;
  camera.reset(Camera_2::Box(0, horz_bound, -vert_bound, vert_bound));
  camera.apply();

  // горизонтальная координатная сетка по 0.1
  auto &hg = scene.horizontal_grid.positions;
  hg.clear();
  for (int yi = (int)(vert_bound / 0.1f) + 4, i = -yi; i < yi; ++i)
  {
    hg.emplace_back(0.f, i * 0.1f);
    hg.emplace_back(horz_bound + 1.f, i * 0.1f);
  }

  scene.axis.positions[0].x = 0.f;
  scene.axis.positions[1].x = horz_bound + 1.f;

  // вертикальная сетка
  auto &vg = scene.vertical_grid.positions;
  vg.clear();
  for (int xi = (int)(horz_bound / 0.1f) + 4, i = 0; i < xi; ++i)
  {
    vg.emplace_back(i * 0.1f, -vert_bound - 1.f);
    vg.emplace_back(i * 0.1f, +vert_bound + 1.f);
  }

  // сигнал
  auto &sv = scene.signal.positions;
  sv.clear();
  for (int i = 0; i < BUFFER_SIZE; ++i)
  {
    const auto x = i * horizontal_step;
    sv.emplace_back(x, signal_function(x));
  }

  // результат фильтра
  auto &fv = scene.filtered.positions;
  fv.resize(sv.size() - 4);
  {
    derivative_4(horizontal_step, sv.begin(), sv.end(),
      fv.begin());
    // x-координаты -- нули, поправим их
    for (int i = 0, imax = fv.size(); i < imax; ++i)
      fv[i].x = (i + 4) * horizontal_step;
  }
}


// обработчики событий FreeGlut

/// Перерисовка сцены.
void repaint()
{
  glClear(GL_COLOR_BUFFER_BIT); // очистить окно, закрасив его цветом очистки
  scene.paint();
  glFinish(); // дождаться выполнения всех OpenGL команд
  glutSwapBuffers(); // обменять буферы кадра, чтобы вывести в окно новый кадр (двойная буферизация)
}

/// Обновить состояние сцены.
void frame_move()
{
  // "спать" 20 миллисекунд, ограничивая таким образом перерисовку 1000/20 = 50 кадрами в секунду
  avo::sleep(0.02); // на деле это может быть далеко не 20мс, а как захочет ОС :)
  if (!paused)
  {
    const auto next_border = left_border + horizontal_step;
    // сдвинуть горизонтальную сетку
    const Vec_2f move_vec(horizontal_step, 0);
    for (auto &v: scene.horizontal_grid.positions)
      v += move_vec;
    scene.axis.positions[0] += move_vec;
    scene.axis.positions[1] += move_vec;

    // поправить вертикальную сетку
    if ((long)(10.f * left_border) != (long)(10.f * next_border))
    {
      auto &vg = scene.vertical_grid.positions;
      vg.erase(vg.begin(), vg.begin() + 2);
      vg.emplace_back(vg[vg.size() - 2]);
      vg.emplace_back(vg[vg.size() - 2]);
      vg[vg.size() - 2].x += 0.1f;
      vg.back().x += 0.1f;
    }

    // продолжить сигнал
    auto &sv = scene.signal.positions;
    {
      sv.erase(sv.begin());
      const auto x = sv.back().x + horizontal_step;
      sv.emplace_back(x, signal_function(x));
    }    

    // продолжить фильтрованный сигнал
    {
      auto &fv = scene.filtered.positions;
      fv.erase(fv.begin());
      fv.emplace_back();
      derivative_4(horizontal_step, sv.end() - 5, sv.end(), fv.end() - 1);
      fv.back().x = sv.back().x;
    }

    // завершение
    left_border = next_border;
    camera.move(horizontal_step, 0);
    camera.apply();
    glutPostRedisplay(); // попросим GLUT выполнить перерисовку
  }
}

/// Обработать нажатие "обычной" клавиши (x, y -- координаты мыши в окне на момент нажатия).
void on_key(unsigned char key, int x, int y)
{
  switch(key)
  {
  case 27: case 'q': case 'Q':
    glutLeaveMainLoop(); // покинуть цикл обработки сообщений
    break;

  case '\n': case '\r':
    reset_scene();
    glutPostRedisplay();
    break;

  case ' ':
    paused = !paused;
    break;

  case '1':
    signal_function = std::sinf;
    break;

  case '2':
    signal_function = std::log10f;
    break;

  case '3':
    signal_function = pabs;
    break;

  case '4':
    signal_function = x2_saw;
    break;

  default:
    return; // остальные коды игнорируются
  }
}


//
// main
//

int main(int argc, char **argv)
{
  // базовая инициализация
  glutInit(&argc, argv); // обработка параметров командной строки

  // создать окно с привязанным контекстом OpenGL
  // режим: 32-битный цвет RGBA, двойная буферизация DOUBLE, попытаться использовать полноэкранное сглаживание
  glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_MULTISAMPLE);
  glutInitWindowPosition(0, 0); // координаты верхнего левого угла окна на экране
  glutInitWindowSize(900, 600); // полные ширина и высота окна
  glutCreateWindow("FIR Filter");  // создать окно, в функцию передаётся заголовок создаваемого окна

  // параметры OpenGL для данного примера
  glEnable(GL_COLOR_MATERIAL); // выбирать цвет материала командой glColor (через Color::select)
  glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE); // этот цвет -- с обеих сторон полигона, ambient и diffuse компоненты
  glClearColor(0.0f, 0.0f, 0.0f, 0.0f); // установить цвет очистки (фон) -- чёрный
  glEnable(GL_LINE_SMOOTH);  // включить "сглаживание линий" (отдельно от полноэкранного сглаживания)

  // включить рисование примитивов, заданных массивами
  glEnableClientState(GL_VERTEX_ARRAY); // включить массив координат вершин

  // привязка функций-обработчиков событий
  glutIdleFunc(frame_move); // вызывать, когда нет событий
  glutDisplayFunc(repaint); // вызывать для отрисовки кадра
  glutKeyboardFunc(on_key); // вызывать для обработки нажатий "обычных" клавиш

  // старт
  signal_function = std::sinf;
  reset_scene();
  glutMainLoop(); // вход в цикл обработки событий
  return 0;
}
