/*
 * This file is part of AV Orchid.
 * Copyright (c) 2015, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * @file FIRModel.hpp
 * @author Кувшинов Д.Р.
 * @brief Исходный и обработанный сигнал.
 */
#pragma once
#include <cstddef>


class FIR_model
{
  struct FIR_model_state; // pimpl
  FIR_model_state *state;

  FIR_model(FIR_model&) = delete;
  FIR_model& operator=(FIR_model&) = delete;

  void cleanup();

public:

  ~FIR_model() { cleanup(); }
  
  /// Размер буфера (в отсчётах), временной шаг дискретизации сигнала.
  /// Замечание: "при прочих равных" для операций со временем рекомендуется использовать double.
  FIR_model(std::size_t bufsz, float timestep);

  // Вспомогательные определения.

  FIR_model()
    : state(nullptr) {}

  FIR_model(FIR_model &&other)
    : state(other.state)
  {
    other.state = nullptr;
  }

  FIR_model& operator=(FIR_model &&other)
  {
    if (this != &other)
    {
      cleanup();
      state = other.state;
      other.state = nullptr;
    }
    return *this;    
  }

  /// Проверка на наличие состояния.
  bool is_valid() const { return state != nullptr; }

  // Основной функционал.

  /// Функция-генератор сигнала (на вход принимает время).
  typedef float (*Signal_generator)(float t);

  /// Выбрать функцию-генератор сигнала.
  void select_signal(Signal_generator);

  /// Установить следующий момент времени (кадра). Вычисляет новые отсчёты.
  /// @return количество новых отсчётов (максимум, сколько можно забрать take_samples)
  std::size_t next_time(float);

  /// Виды буферов.
  enum class Buffer : unsigned { Signal, Filtered };

  /// Узнать количество доступных для чтения элементов в буфере.
  std::size_t available_samples(Buffer) const;

  /// "Забрать" из заданного буфера заданное число готовых отсчётов.
  /// @return указатель на начало буфера, nullptr, если samples > доступного числа.
  const float* take_samples(Buffer, std::size_t samples);
};

