﻿/*
 * This file is part of AV Orchid.
 * Copyright (c) 2013, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * Пример: GLPrimitives
 * Автор: Кувшинов Д.Р. https://evetro.wordpress.com/
 * Используемые библиотеки: FreeGlut, OpenGL.
 *
 * Рисует случайно сгенерированный движущийся набор цветных точек ("вершин"), соединённых
 * в GL-примитив некоторого вида (выбирается цифровой клавишей):
 *   1 -- GL_POINTS
 *   2 -- GL_LINES
 *   3 -- GL_LINE_STRIP
 *   4 -- GL_LINE_LOOP
 *   5 -- GL_TRIANGLES
 *   6 -- GL_TRIANGLE_STRIP
 *   7 -- GL_TRIANGLE_FAN
 *   8 -- GL_QUADS
 *   9 -- GL_QUAD_STRIP
 *
 * Вершины двигаются внутри коробки (0, 0, 0) -- (1, 1, 1), "отскакивая" от стенок.
 *
 * Нажатие BACKSPACE переключает режим закраски между FLAT и SMOOTH.
 * Нажатие +/- добавляет/удаляет вершины.
 *
 * Клавиши F10, F11 и F12 переключают режим растеризации полигонов (точки, линии, зарашенные).
 *
 * Нажатие ENTER/SPACE генерирует новый набор псевдослучайных вершин.
 * Нажатие ESC/Q -- выход из программы.
 */
#include <GL/freeglut.h>
#include <avo/Basics.hpp> // random

#include <avo/Timer.hpp>
#include <avo/Color.hpp> // RGBA-цвет в 128-битном формате (компоненты типа float)
#include <avo/Vec3.hpp>  // 3D-вектор

using avo::Color;
using avo::Vec_3f;
using avo::random;

//
// вспомогательные функции
//

/// сгенерировать некоторый случайный яркий цвет
inline Color random_bright_color()
{
  return Color(random(), random(), random()).bright();
}


/// вершина: координаты, скорость, цвет
struct Vertex
{
  Vec_3f pos, vel; ///< позиция и скорость
  Color  col;      ///< цвет

  /// движение, шаг на h секунд
  void move(float h)
  {
    // собственно движение
    pos += h * vel;
    // отражение от стенок
    if      (pos.x <= 0.f)  vel.x =  abs(vel.x);
    else if (1.f <= pos.x)  vel.x = -abs(vel.x);
    if      (pos.y <= 0.f)  vel.y =  abs(vel.y);
    else if (1.f <= pos.y)  vel.y = -abs(vel.y);
    if      (pos.z <= 0.f)  vel.z =  abs(vel.z);
    else if (1.f <= pos.z)  vel.z = -abs(vel.z);
  }

  /// сгенерировать случайную вершину
  void make_random()
  {
    pos = Vec_3f(random(), random(), random());
    vel = Vec_3f(random(), random(), random()).normalize() * 0.05f;
    col = random_bright_color();
  }
};


//
// константы и глобальные переменные
// 

/// максимально возможное количество вершин примитива
const int N_MAX = 2400;
/// текущее количество вершин
int N;

/// вершины
Vertex vertex[N_MAX];

/// последний момент отрисовки
double t0;

/// режим рисования геометрического примитива OpenGL
GLenum prim_mode;
/// признак: используется градиентная заливка треугольников
bool smooth_shading;

/// генерируем случайные вершины, заполняя массив vertex с помощью make_random
void make_random_vertices()
{
  for (auto &v: vertex)
    v.make_random();

  N = 24;
  t0 = avo::seconds();
}


//
// обработчики событий GLUT
//

/// изменение размеров окна, функция получает "новые размеры"
void reshape(int w, int h)
{
  // новая матрица проекции
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity(); // загрузить единичную матрицу
  /* угол зрения по вертикали, соотношение сторон экрана, минимальная дальность, максимальная дальность */
  gluPerspective(60, double(w) / double(h), 0.25, 2.0); // домножить на матрицу перспективной коррекции
  gluLookAt
  (
      0.5, 0.5, -0.625  // позиция камеры
    , 0.5, 0.5, 0.5     // точка, на которую смотрит камера
    , 0.0, 1.0, 0.0     // вектор "вверх"
  );

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  // установить координаты окна просмотра (вся клиентская область)
  glViewport(0, 0, w, h);
}

/// перерисовка сцены
void repaint()
{
  // очистить окно, закрасив его цветом очистки
  glClear(GL_COLOR_BUFFER_BIT);
  glBegin(prim_mode);
  for (int i = 0; i < N; ++i)
  {
    vertex[i].col.select();
    glVertex3fv(&vertex[i].pos.x); // все три координаты идут в памяти подряд,
    // поэтому можно их не выписывать на месте, а передать адрес первой
  }
  glEnd();

  glFinish(); // дождаться выполнения всех OpenGL команд
  glutSwapBuffers(); // обменять буферы кадра, чтобы вывести в окно новый кадр (двойная буферизация)
}


/// вызывается из главного цикла, когда нет событий для обработки
void frame_move()
{
  // "спать" 8 миллисекунд, ограничивая таким образом перерисовку 1000/8 = 125 кадрами в секунду
  avo::sleep(0.008); // на деле это может быть далеко не 8мс, а как захочет ОС :)
  const double t1 = avo::seconds(); // новый момент времени
  const float h = (float)(t1 - t0); // шаг
  t0 = t1;
  for (int i = 0; i < N; ++i) // передвинуть все вершины
    vertex[i].move(h);
  glutPostRedisplay(); // попросим GLUT выполнить перерисовку
}


/// обработать нажатие "обычной" клавиши (x, y -- координаты мыши в окне на момент нажатия)
void on_key(unsigned char key, int x, int y)
{
  switch(key)
  {
  case 27: case 'q': case 'Q':
    glutLeaveMainLoop(); // покинуть цикл обработки сообщений
    break;

  case ' ': case '\r': case '\n':
    make_random_vertices();
    glutPostRedisplay();
    break;

  case '-':
    if (--N < 1) N = 1;
    break;

  case '=': case '+':
    if (++N > N_MAX) N = N_MAX;
    break;

  case '1': prim_mode = GL_POINTS;          break;
  case '2': prim_mode = GL_LINES;           break;
  case '3': prim_mode = GL_LINE_STRIP;      break;
  case '4': prim_mode = GL_LINE_LOOP;       break;
  case '5': prim_mode = GL_TRIANGLES;       break;
  case '6': prim_mode = GL_TRIANGLE_STRIP;  break;
  case '7': prim_mode = GL_TRIANGLE_FAN;    break;
  case '8': prim_mode = GL_QUADS;           break;
  case '9': prim_mode = GL_QUAD_STRIP;      break;

  case '\b':
    glShadeModel( (smooth_shading = !smooth_shading)? GL_SMOOTH: GL_FLAT);
    break;

  default:
    return; // остальные коды игнорируются
  }
}


/// обработать нажатие "специальной" клавиши (x, y -- координаты мыши в окне на момент нажатия)
void on_special(int key, int x, int y)
{
  switch (key)
  {
    // выбор режима растеризации полигонов
  case GLUT_KEY_F10: glPolygonMode(GL_FRONT_AND_BACK, GL_POINT);  break; // точки
  case GLUT_KEY_F11: glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);   break; // линии
  case GLUT_KEY_F12: glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);   break; // нормальный (закрашивать)
  default:
    return; // остальные коды игнорируются
  }
}


//
// main
//

int main(int argc, char **argv)
{
  // базовая инициализация
  glutInit(&argc, argv); // обработка параметров командной строки
  std::srand((unsigned)(avo::seconds() * 1e+6) % 15485863);
  smooth_shading = true;

  // создать окно с привязанным контекстом OpenGL
  // режим: 32-битный цвет RGBA, двойная буферизация DOUBLE, попытаться использовать полноэкранное сглаживание
  glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_MULTISAMPLE);
  glutInitWindowPosition(0, 0); // координаты верхнего левого угла окна на экране
  glutInitWindowSize(800, 600); // полные ширина и высота окна
  glutCreateWindow("GLPrimitives");  // создать окно, в функцию передаётся заголовок создаваемого окна

  // параметры OpenGL для данного примера
  glEnable(GL_COLOR_MATERIAL); // выбирать цвет материала командой glColor (через Color::select)
  glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE); // этот цвет -- с обеих сторон полигона, ambient и diffuse компоненты
  glClearColor(0.0f, 0.0f, 0.0f, 0.0f); // установить цвет очистки (фон) -- чёрный
  glPointSize(5.f); // установить диаметр точки -- 5 пикселей
  glEnable(GL_POINT_SMOOTH); // включить "сглаживание точек" (отдельно от полноэкранного сглаживания)
    // -- без этой команды на некоторых устройствах/версиях драйверов точки будут отображаться не кружками, а квадратами
  glEnable(GL_LINE_SMOOTH);  // включить "сглаживание линий" (отдельно от полноэкранного сглаживания)
 
  // привязка функций-обработчиков событий
  glutIdleFunc(frame_move);     // вызывать, когда нет событий
  glutReshapeFunc(reshape);     // изменение размеров окна
  glutDisplayFunc(repaint);     // вызывать для отрисовки кадра
  glutKeyboardFunc(on_key);     // вызывать для обработки нажатий "обычных" клавиш
  glutSpecialFunc(on_special);  // вызывать для обработки нажатий "специальных" клавиш

  // старт
  make_random_vertices();
  glutMainLoop(); // вход в цикл обработки событий
  return 0;
}
