﻿/*
 * This file is part of AV Orchid.
 * Copyright (c) 2013, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * Пример: GLUPoly
 * Автор: Кувшинов Д.Р. https://evetro.wordpress.com/
 * Используемые библиотеки: FreeGlut, OpenGL.
 *
 * Демонстрация использования GLU для триангуляции многоугольников.
 * Управление "камерой": стрелки и WASD, +/- Z/X для увеличения/уменьшения масштаба.
 * Колёсико мыши изменяет масштаб (как +/-).
 * HOME -- вернуть камеру в положение по умолчанию.
 * SPACE/ENTER -- загрузка файла (название надо ввести в консоли).
 * F3 -- включить/выключить отображение размеров видимого поля.
 * F4 -- включить/выключить отображение координат мыши.
 * ESC -- выход из программы.
 */
#include "Mesh2.hpp"

#include <avo/CoordinateGrid2.hpp>
#include <avo/TextDrawer.hpp>
#include <avo/Camera2.hpp>
#include <avo/Timer.hpp>

#include <cmath>
#include <string>
//#include <functional>
#include <algorithm>
#include <iostream>
#include <fstream>

using avo::Color;
using avo::Camera_2;

/// определение "сцены"
class Scene
{
  Mesh_2 mesh; // набор примитивов, представляющий многоугольник
  Bbox_2 mbox; // бокс, в который вписана сетка (для автопозиционирования камеры)

public:

  /// загрузить набор контуров из файла fn
  void load(const std::string &fn)
  {
    std::ifstream f(fn);
    if (!f)
      std::cerr << "File \"" << fn << "\" can't be loaded." << std::endl;
    else
      load(f);
  }

  /// считать набор контуров из потока
  void load(std::istream &is)
  {
    using namespace std;

    int n = 0; // количество контуров
    is >> n;

    if (n > 0)
    {
      cout << n << " contours.\n";
      vector<Contour> ctrs(n);

      // прочитать из файла первый контур и сформировать бокс
      is >> ctrs[0];
      auto   verts = ctrs[0].size();
      Scalar area  = ctrs[0].area(),
             perim = ctrs[0].perimeter();
      mbox = ctrs[0].box();

      // прочитать оставшиеся контуры
      for (int i = 1; i < n && is.good(); ++i)
      {
        is >> ctrs[i];
        verts += ctrs[i].size();
        area  += ctrs[i].area();
        perim += ctrs[i].perimeter();
        mbox  += ctrs[i].box();
      }

      // сводка по загруженным контурам
      cout << verts << " vertices, area = " << area << ", perimeter = " << perim;
      cout << "\nbox: min = " << mbox.pmin << "; max = " << mbox.pmax << endl;

      // ликвидировать пустые контуры
      ctrs.erase
        (
          remove_if
            (
              ctrs.begin(), ctrs.end(), // mem_fun не хочет работать, так пусть будет лямбда :)
              [](const Contour &c) { return c.empty(); } // std::mem_fn(&Contour::empty)
            ),
          ctrs.end()
        );

      // построить сетку
      try
      {
        mesh = Mesh_2(ctrs.begin(), ctrs.end());
        mesh.background.alpha = 0.4f;
      }
      catch (...)
      {
        cerr << "Failed to create new mesh." << endl;
      }
    }
    else
      cout << "No contours found." << endl;
  }


  /// изобразить "сцену"
  void paint() const
  {
    mesh.paint();
  }

  /// получить бокс
  const Bbox_2& box() const
  {
    return mbox;
  }
};


//
// глобальные объекты

Scene scene;     ///< многоугольники
Camera_2 camera; ///< камера
avo::Coordinate_grid_2<double> cgrid;  ///< координатная сетка
avo::Bitmap_text_drawer_2 text_drawer; ///< объект, выводящий текст

bool drawing_box_coordinates   = true; ///< рисовать координаты бокса
bool drawing_mouse_coordinates = true; ///< рисовать координаты мыши

/// текущая высота окна
int cur_wnd_height;

/// координаты мыши
int mouse_x, mouse_y;

//
// вспомогательные функции

/// обновить координатную сетку при изменении камеры
void camera_changed()
{
  cgrid.update(camera.box());
  text_drawer.set_coordinates(camera.box());
  camera.apply();
  glutPostRedisplay();  // уведомить FreeGlut, что следует перерисовать сцену
}

/// приблизить камеру
void camera_zoom_in()
{
  camera.zoom_in(0.125);
  camera_changed();
}

/// "отодвинуть" камеру
void camera_zoom_out()
{
  camera.zoom_out(0.125);
  camera_changed();
}

/// отображение координат
void draw_coordinates()
{
  using std::to_string;
  if (!drawing_box_coordinates && !drawing_mouse_coordinates)
    return;

  auto smaller_box = camera;
  smaller_box.zoom_in(0.05); // чтобы не пропадали надписи на краях при округлении "наружу"
  auto &box = smaller_box.box();

  if (drawing_box_coordinates)
  {
    // сгенерировать строки
    const auto left   = to_string(box.pmin.x),
               bottom = to_string(box.pmin.y),
               right  = to_string(box.pmax.x),
               top    = to_string(box.pmax.y);

    // вывести текст средствами FreeGlut
    const auto center = 0.5 * (box.pmax + box.pmin);

    avo::LIGHT_GREY.select();
    text_drawer.view(box.pmin.x, center.y).put(left)
               .view(center.x, box.pmin.y).put(bottom)
               .view(box.pmax.x, center.y).put(right, avo::Alignment::toLeft)
               .view(center.x, box.pmax.y).put(top, avo::Alignment::toDown)
               ;
  }

  if (drawing_mouse_coordinates)
  {
    double mx, my; // "координаты" мыши
    text_drawer.pixels_to_view(mouse_x, mouse_y, mx, my);

    // сгенерировать строки
    const auto xy = to_string(mx) + '\n' + to_string(my);
    
    // вывести текст средствами FreeGlut
    avo::AZURE.select();
    text_drawer.view(box.pmin.x, box.pmax.y).put(xy, avo::Alignment::toDown);
  }
}


//
// колбеки FreeGlut

/// перерисовка
void repaint()
{
  glClear(GL_COLOR_BUFFER_BIT);
  cgrid.paint();
  scene.paint();
  draw_coordinates();
  glutSwapBuffers();
}


/// изменён размер окна
void reshape(int w, int h)
{
  // изменить окно камеры
  camera.force_aspect((double)w / h);
  camera.apply();
  camera_changed();
  text_drawer.set_viewport(w, h);
  cur_wnd_height = h;

  // уведомить OpenGL о новых размерах области рисования
  glViewport(0, 0, w, h);
}


/// обработать нажатие "специальной" клавиши
void on_special_key(int key, int x, int y)
{
  Vec_2 delta; // 0, 0
  switch (key)
  {
  case GLUT_KEY_LEFT:   delta.x = -0.1; break;
  case GLUT_KEY_UP:     delta.y =  0.1; break;
  case GLUT_KEY_RIGHT:  delta.x =  0.1; break;
  case GLUT_KEY_DOWN:   delta.y = -0.1; break;
  case GLUT_KEY_HOME:   camera.reset(scene.box()); break;
  case GLUT_KEY_F3:     
    drawing_box_coordinates = !drawing_box_coordinates;
    return glutPostRedisplay();

  case GLUT_KEY_F4:
    drawing_mouse_coordinates = !drawing_mouse_coordinates;
    return glutPostRedisplay();

  default:
    return; // ничего не делать, если была нажата какая-либо другая клавиша
  }

  delta *= camera.box().extents();
  camera.move(delta);
  camera_changed();
}

/// обработать нажатие "обычной" клавиши
void on_key(unsigned char key, int x, int y)
{
  switch (key)
  {
  case 27: // код ESC (ASCII) --  выйти из цикла обработки событий
    return glutLeaveMainLoop();

  case '+': case 'z': case 'Z': camera_zoom_in();  break;
  case '-': case 'x': case 'X': camera_zoom_out(); break;
  case 'w': case 'W': on_special_key(GLUT_KEY_UP, x, y); break;
  case 's': case 'S': on_special_key(GLUT_KEY_DOWN, x, y); break;
  case 'a': case 'A': on_special_key(GLUT_KEY_LEFT, x, y); break;
  case 'd': case 'D': on_special_key(GLUT_KEY_RIGHT, x, y); break;

  case ' ': case '\r': case '\n':
    {
      // загрузить файл (ввести название в консоль)
      std::string fn;
      std::cout << "Enter the input file name:\n";
      std::getline(std::cin, fn);
      scene.load(fn);

      camera.reset(scene.box());  
      camera_changed();
    }
    break;
  
  default:
    return; // do nothing for other keys
  }
}


/// отслеживаем движение мыши, чтобы отображать её координаты
void on_hover(int x, int y)
{
  mouse_x = x;
  mouse_y = cur_wnd_height - y;
  glutPostRedisplay();
}

/// масштабирование колёсиком мыши
void on_mouse_wheel(int wheel, int direction, int x, int y)
{
  if (direction < 0)
    camera_zoom_in();
  else if (direction > 0)
    camera_zoom_out();
}

/// по возможности не грузим процессор зря :)
void idle()
{
  avo::sleep(0.05);
}


//
// main

int main(int argc, char **argv)
{
  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_MULTISAMPLE);
  
  glutInitWindowPosition(10, 20);
  glutInitWindowSize(1000, 640);
  glutCreateWindow("GLUPoly");

  glClearColor(0.0f, 0.0f, 0.07f, 1.0f);

  glEnable(GL_BLEND); // прозрачность
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  
  glEnable(GL_COLOR_MATERIAL);
  glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);

  // включить массивы вершин (OpenGL 1.1)
  glEnableClientState(GL_VERTEX_ARRAY);

  // регистрация колбеков
  glutDisplayFunc  (repaint);
  glutReshapeFunc  (reshape);
  glutKeyboardFunc (on_key);
  glutSpecialFunc  (on_special_key);
  glutIdleFunc     (idle);

  glutPassiveMotionFunc(on_hover);
  glutMouseWheelFunc(on_mouse_wheel);

  // старт
  std::cout << "Starting GLUPoly...\n";
  scene.load("input.txt");
  camera.reset(scene.box());
  glutMainLoop();
  std::cout << "\nbye!";
  return 0;
}
