/*
 * This file is part of AV Orchid.
 * Copyright (c) 2013, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * @file   BoundedArray.hpp
 * @brief  Bounded_array container: contains a static array and currently used size field.
 * @author Dmitri Kuvshinov
 *
 * Bounded_array may provide more performance than std::vector or std::valarray
 * when maximal size is known and small (due to not using dynamic allocation at all).
 *
 * The class is modelled against std::array and std::vector, however it is not intended to be an aggregate type.
 * Bounded_array does not used allocators. Proper alignment is supported.
 *
 * Status: testing pending. 
 * This implementation may still be too complex.
 *
 * @todo test, this class is not tested well
 * @todo refactor trivial/non-trivial almost duplicated parts
 */
#pragma once
#include <cstddef>
#include <iterator>
#include <memory>
#include <utility>
#include <type_traits>
#include <stdexcept>
#include <algorithm>
#include <initializer_list>

// should be placed near ("" means implementation-defined behavior, so problems are possible)
#include "memmove.hpp"


namespace avo
{

// implementation details
namespace impl
{

  // storage with padding
  template <class SizeType, std::size_t Padding, std::size_t Size>
  struct Bounded_array_storage_with_padding
  {
    SizeType sz;
  private:
    char padding[Padding];
  public:
    char bin[Size];
  };

  // storage without padding
  template <class SizeType, std::size_t Size>
  struct Bounded_array_storage_with_padding<SizeType, 0, Size>
  {
    SizeType sz;
    char bin[Size];
  };

  // provides possibly-aligned storage (inserts padding if needed)
  template <class T, class SizeType, std::size_t Align, std::size_t Size, bool SizeChecksOn>
  struct Bounded_array_storage
    : Bounded_array_storage_with_padding
      <
        SizeType,
        (Align <= sizeof(SizeType)? 0: Align - sizeof(SizeType)),
        Size * sizeof(T)
      >
  {
    /// get a pointer to the array
    T* data()
    {
      return reinterpret_cast<T*>(bin);
    }

    /// get a const pointer to the array
    const T* data() const
    {
      return reinterpret_cast<const T*>(bin);
    }

  protected:
    // aux staff
    
    /// throw size underflow (std::out_of_range)
    static void error_size_underflow()
    {
      throw std::out_of_range("Bounded_array: size underflow");
    } 

    /// throw size overflow (std::out_of_range)
    static void error_size_overflow()
    {
      throw std::out_of_range("Bounded_array: size overflow");
    }

    /// throw index or iterator out of range (std::out_of_range)
    static void error_index_out_of_range()
    {
      throw std::out_of_range("Bounded_array: index or iterator out of range");
    }

    /// check if size is possible (against maximal one) and throw if not (SizeChecksOn = false disables this check)
    void check_size(const SizeType size) const
    {
      // the compiler should optimize out calls to this function when
      // SizeChecksOn is false and reduce SizeChecksOn check when it's true
      if (SizeChecksOn && Size < static_cast<std::size_t>(size))
        error_size_overflow();
    }

    /// check if index is possible according to the current size and throw if not
    void check_index(const SizeType index) const
    {
      if (static_cast<std::size_t>(sz) <= static_cast<std::size_t>(index))
        error_index_out_of_range();  
    }
  };


  /////////////////////////////////////////////////////////////////////////////////////////////////
  // implementation of Bounded_array memory operations for trivial types
  template <class T, std::size_t MaxSize, class SizeType, bool SizeChecksOn>
  class Bounded_array_trivial
    : protected Bounded_array_storage
        <T, SizeType, std::alignment_of<T>::value, MaxSize, SizeChecksOn>
  {
    using Base = Bounded_array_storage
      <T, SizeType, std::alignment_of<T>::value, MaxSize, SizeChecksOn>;
    using This = Bounded_array_trivial
      <T, MaxSize, SizeType, SizeChecksOn>;

  protected:
    ~Bounded_array_trivial() {} // = default;

  public:

    /// creates an empty array
    Bounded_array_trivial()
    {
      sz = 0;
    }

    /// copies elements from {}
    template <class T>
    explicit Bounded_array_trivial(std::initializer_list<T> inl)
    {
      assign(inl.begin(), inl.end());
    }

    /// do a simple bitwise copy
    Bounded_array_trivial(const This &other)
    {
      copy_from(other);
    }

    /// do a simple bitwise copy
    This& operator=(const This &other)
    {
      copy_from(other);
      return *this;
    }

    /// copies elements from {}
    This& operator=(std::initializer_list<T> inl)
    {
      assign(inl.begin(), inl.end());
      return *this;
    }


    using Base::data;

    /// fill the array with values from [begin, end) (may throw if the range is too big)
    template <class InIt>
    void assign(InIt begin, InIt end)
    {
      if (begin == end)
      {
        sz = 0;
        return;
      }

      T *d = data();
      const auto die = d + MaxSize; // ultimate end
      for (;;)
      {
        *d++ = *begin;
        if (++begin == end)
          break;

        if (d == die)
        {
          sz = MaxSize;
          error_size_overflow();
        }
      }

      sz = d - data();
    }

    /// produce n copies of value if possible (may throw if n is too big)
    void assign_n(const SizeType n, const T &value)
    {
      check_size(n);
      sz = n;
      std::fill_n(data(), n, value);
    }

    /// swap contents with a similar array (linear time)
    void swap(This &other)
    {
      // it still feels like pointer juggling, yeah
      const auto sz1 = sz,
                 sz2 = other.sz,
                 sz0 = std::min(sz1, sz2);

      // swap sizes
      other.sz = sz1;
      sz = sz2;

      const auto d1 = data(),
                 d2 = other.data(),
                 d1tail = d1 + sz0,
                 d2tail = d2 + sz0;

      // swap main part
      std::swap_ranges(d1, d1tail, d2);
      // and copy the tail of the bigger array
      if (sz1 < sz2)
        std::copy(d2tail, d2 + sz2, d1tail);
      else if (sz2 < sz1)
        std::copy(d1tail, d1 + sz1, d2tail);
    }

    /// try to resize, new elements are not assigned (contradicts std::vector behaviour)
    void resize(const SizeType new_size)
    {
      const auto size = sz;
      if (size < new_size)
        grow(new_size - size);
      else
        shrink(size - new_size);
    }

    /// try to resize, assign new elements if growing to new_value
    void resize(const SizeType new_size, const T &new_value)
    {
      const auto size = sz;
      if (size < new_size)
      {
        const auto delta = new_size - size;
        grow(delta);
        std::fill_n(data() + size, delta, new_value);        
      }
      else
        shrink(size - new_size);
    }

    /// just sets size = 0
    void clear()
    {
      sz = 0;
    }

    /// removes the last element (throws if SizeChecksOn and size is zero)
    void pop_back()
    {
      shrink(1);
    }

    /// inserts one new element in the end of the array (throws if SizeChecksOn and maximal size is reached)
    void push_back(const T &new_value)
    {
      *buy() = new_value;
    }

    void push_back(T &&moved_value)
    {
      emplace_back(std::move(moved_value));
    }

    /// constructs one new element in the end of the array from args (throws if SizeChecksOn and maximal size is reached)
    template <class ...Args>
    void emplace_back(Args&&... args)
    {
      new((void*)buy()) T(std::forward<Args>(args)...);
    }


    //
    // more std::vector features: insert, emplace, erase

    /// inserts a copy of value at position pos
    T* insert(const T *pos, const T &value)
    {
      return emplace(pos, value);
    }

    T* insert(const T *pos, T &&value)
    {
      return emplace(pos, std::move(value));
    }

    /// inserts n copies of value in pos, shifts tail elements (linear in time), may throw
    T* insert(const T *pos, const SizeType n, const T &value)
    {
      const auto d = data();
      const auto i = pos - d;
      check_index(i);
      const auto dsrc = d + i,
                 ddst = dsrc + n,
                 dend = d + sz;

      grow(n);
      memmove(dsrc, dend, ddst);
      std::fill_n(dsrc, n, value);
      return ddst;
    }

    /// inserts copies of elements from [begin, end), shifts tail elements (linear in time), may throw
    template <class InIt>
    T* insert(const T *pos, InIt begin, InIt end)
    {
      // static dispatch according to InIt category
      return insert_impl(pos, begin, end, typename
        std::iterator_traits<InIt>::iterator_category());
    }

    /// constructs an element from args at pos, shifts tail elements (linear in time), may throw
    template <class... Args>
    T* emplace(const T *pos, Args&&... args)
    {
      const auto d = data();
      const auto i = pos - d;
      check_index(i);
      const auto dsrc = d + i,
                 ddst = dsrc + 1,
                 dend = d + sz;
      
      grow(1);
      memmove(dsrc, dend, ddst);
      new((void*)dsrc) T(std::forward<Args>(args)...);
      return ddst;
    }


    /// erases one element at pos, shifts tail elements (linear in time), may throw
    T* erase(const T *pos)
    {
      const auto d = data(),
                 dend = d + sz;
      const auto i = pos - d;
      check_index(i);
      shrink(1);
      const auto ddst = d + i,
                 dsrc = ddst + 1;
      memmove(dsrc, dend, ddst);
      return const_cast<T*>(pos);
    }

    /// erases elements, which are in [begin, end), shifts tail elements (linear in time), may throw
    T* erase(const T *begin, const T *end)
    {
      if (end <= begin)
        return const_cast<T*>(end);

      const auto d = data(),
                 dend = d + sz;
   
      check_index(begin - d);
      if (dend < end)
        error_index_out_of_range();

      // it seems it's safe to proceed...
      if (dend != end)
      {
        const auto dsrc = const_cast<T*>(end),
                   ddst = const_cast<T*>(begin);
        memmove(dsrc, dend, ddst);
      }
      
      shrink(end - begin);
      return const_cast<T*>(begin);
    }


  private:

    /// increase size by delta elements
    void grow(const SizeType delta)
    {
      const auto nsz = sz + delta;
      check_size(nsz);
      sz = nsz;
    }

    /// add one element in end and return its address
    T* buy()
    {
      T *p = data() + sz;
      grow(1);
      return p;
    }

    /// decreases size by delta elements
    void shrink(const SizeType delta)
    {
      const auto size = sz;
      if (size < delta)
        error_size_underflow();
      sz = size - delta;
    }

    /// copy contents from other, overwrites this one
    void copy_from(const This &other)
    {
      const auto size = other.sz;
      sz = size;
      const auto d = other.data();
      std::copy(d, d + size, data());
    }

    /// insert range implementation for random access iterators
    template <class RanIt>
    T* insert_impl(const T *pos, RanIt begin, RanIt end, std::random_access_iterator_tag)
    {
      const auto d = data();
      const auto n = end - begin;
      const auto i = pos - d;
      check_index(i);
      const auto dsrc = d + i,
                 ddst = dsrc + n,
                 dend = d + sz;

      grow(n);
      memmove(dsrc, dend, ddst);
      std::copy(begin, end, dsrc);
      return ddst;
    }

    /// insert range implementation for input iterators
    template <class InIt>
    T* insert_impl(const T *pos, InIt begin, InIt end, std::input_iterator_tag)
    {
      if (begin == end)
        return const_cast<T*>(pos);

      // there's something to insert
      const auto size = sz;
      const auto dbegin = data(),
                 dend = dbegin + sz,
                 die = dbegin + MaxSize;

            auto d = dend;
      const auto i = pos - dbegin;
      check_index(i);

      // first, try to insert it in the end
      for (;;)
      {
        if (d == die)
        {
          sz = MaxSize;
          error_size_overflow();
        }

        *d++ = *begin;
        if (++begin == end)
        {
          sz = d - dbegin;
          break;
        }
      }

      // second, rotate the inserted range to pos
      const auto dpos = dbegin + i;
      std::rotate(dpos, dend, d);
      return dpos + (d - dend);
    }
  };


  /////////////////////////////////////////////////////////////////////////////////////////////////
  // implementation of Bounded_array memory operations for non-trivial types
  template <class T, std::size_t MaxSize, class SizeType, bool SizeChecksOn>
  class Bounded_array_non_trivial
    : protected Bounded_array_storage
        <T, SizeType, std::alignment_of<T>::value, MaxSize, SizeChecksOn>
  {
    using Base = Bounded_array_storage
      <T, SizeType, std::alignment_of<T>::value, MaxSize, SizeChecksOn>;
    using This = Bounded_array_non_trivial
      <T, MaxSize, SizeType, SizeChecksOn>;

  protected:

    ~Bounded_array_non_trivial()
    {
      clear();
    }

  public:

    // constructors

    Bounded_array_non_trivial()
    {
      sz = 0;
    }

    Bounded_array_non_trivial(const This &other)
    {
      sz = 0;
      copy_from(other);
    }

    Bounded_array_non_trivial(This &&other)
    {
      sz = 0;
      move_from(other);
    }

    template <class T>
    Bounded_array_non_trivial(std::initializer_list<T> inl)
    {
      sz = 0;
      assign(inl.begin(), inl.end());
    }

    // assignments

    This& operator=(const This &other)
    {
      copy_from(other);
      return *this;
    }

    This& operator=(This &&other)
    {
      move_from(other);
      return *this;
    }

    template <class T>
    This& operator=(std::initializer_list<T> inl)
    {
      assign(inl.begin(), inl.end());
      return *this;
    }


    using Base::data;

    template <class InIt>
    void assign(InIt begin, InIt end)
    {
      clear();
      if (begin == end)
        return;

      T *d = data();
      const auto du  = d + sz, // uninitialized start here
                 die = d + MaxSize; // ultimate end

      while (d != du)
      {
        *d++ = *begin;
        if (++begin == end)
        {
          shrink(du - d);
          return;
        }
      }

      // here begin != end but d == du
      for (;;)
      {
        new((void*)d++) T(*begin);
        if (++begin == end)
        {
          sz = d - data();
          return;
        }

        if (d == die)
        {
          sz = MaxSize;
          error_size_overflow();
        }
      }
    }


    void assign_n(const SizeType n, const T &value)
    {
      check_size(n);
      const auto d0 = data(),
                 d1 = d0 + sz,
                 d2 = d0 + n;

      if (d1 < d2)
      {
        sz = n;
        std::fill(d0, d1, value); // initialized part
        std::uninitialized_fill(d1, d2, value); // uninitialized part
      }
      else
      {
        shrink(sz - n); // destroy excess elements
        std::fill(d0, d2, value);
      }
    }


    void swap(This &other)
    {
      // it still feels like pointer juggling, yeah
      const auto sz1 = sz,
                 sz2 = other.sz,
                 sz0 = std::min(sz1, sz2);

      // swap sizes
      other.sz = sz1;
      sz = sz2;

      const auto d1 = data(),
                 d2 = other.data(),
                 d1tail = d1 + sz0,
                 d2tail = d2 + sz0;

      // swap main part
      std::swap_ranges(d1, d1tail, d2);
      // and copy the tail of the bigger array
      if (sz1 < sz2)
        std::uninitialized_copy(d2tail, d2 + sz2, d1tail);
      else if (sz2 < sz1)
        std::uninitialized_copy(d1tail, d1 + sz1, d2tail);
    }


    void resize(const SizeType new_size, const T &new_value = T())
    {
      const auto size = sz;
      if (size < new_size)
      {
        const auto delta = new_size - size;
        grow(delta);
        std::uninitialized_fill_n(data() + size, delta, new_value);        
      }
      else
        shrink(size - new_size);
    }


    void clear()
    {
      resize(0);
    }

    void pop_back()
    {
      shrink(1);
    }

    void push_back(const T &new_value)
    {
      emplace_back(new_value);
    }

    void push_back(T &&new_value)
    {
      emplace_back(std::move(new_value));
    }

    template <class... Args>
    void emplace_back(Args&&... args)
    {
      new((void*)buy()) T(std::forward<Args>(args)...);
    }


    //
    // more std::vector features: insert, emplace, erase

    /// same as emplace(pos, value)
    T* insert(const T *pos, const T &value)
    {
      return emplace(pos, value);
    }

    T* insert(const T *pos, T &&value)
    {
      return emplace(pos, std::move(value));
    }

    /// inserts n copies of value in pos, shifts tail elements (linear in time), may throw
    T* insert(const T *pos, const SizeType n, const T &value)
    {
      const auto d = data();
      const auto i = pos - d;
      check_index(i);
      const auto dsrc = d + i,
                 ddst = dsrc + n,
                 dend = d + sz;

      grow(n); // doesn't initialize
      uninitialized_memmove(dsrc, dend, ddst, do_not_destroy);
      std::fill_n(dsrc, n, value); // initialized, just assign
      return ddst;
    }

    /// inserts copies of elements from [begin, end), shifts tail elements (linear in time), may throw
    template <class InIt>
    T* insert(const T *pos, InIt begin, InIt end)
    {
      // static dispatch according to InIt category
      return insert_impl(pos, begin, end, typename
        std::iterator_traits<InIt>::iterator_category());
    }

    /// constructs an element from args at pos, shifts tail elements (linear in time), may throw
    template <class... Args>
    T* emplace(const T *pos, Args&&... args)
    {
      const auto d = data();
      const auto i = pos - d;
      check_index(i);
      const auto dsrc = d + i,
                 ddst = dsrc + 1,
                 dend = d + sz;
      
      grow(1);
      uninitialized_memmove(dsrc, dend, ddst); // destroy one object
      new((void*)dsrc) T(std::forward<Args>(args)...);
      return ddst;
    }


    /// erases one element at pos, shifts tail elements (linear in time), may throw
    T* erase(const T *pos)
    {
      const auto d = data(),
                 dend = d + sz;
      const auto i = pos - d;
      check_index(i);
      const auto ddst = d + i,
                 dsrc = ddst + 1;
      initialized_memmove(dsrc, dend, ddst);
      shrink(1);
      return const_cast<T*>(pos);
    }

    /// erases elements, which are in [begin, end), shifts tail elements (linear in time), may throw
    T* erase(const T *begin, const T *end)
    {
      if (end <= begin)
        return const_cast<T*>(end);

      const auto d = data(),
                 dend = d + sz;
   
      check_index(begin - d);
      if (dend < end)
        error_index_out_of_range();

      // it seems it's safe to proceed...
      if (dend != end)
      {
        const auto dsrc = const_cast<T*>(end),
                   ddst = const_cast<T*>(begin);
        initialized_memmove(dsrc, dend, ddst);
      }
      
      shrink(end - begin);
      return const_cast<T*>(begin);
    }

  private:

    // doesn't construct elements, only allocates space for them
    void grow(const SizeType delta)
    {
      const auto nsz = sz + delta;
      check_size(nsz);
      sz = nsz;
    }

    // allocates one more element (doesn't construct it) and returns its address
    T* buy()
    {
      T *p = data() + sz;
      grow(1);
      return p;
    }

    // destroys delta elements if possible
    void shrink(const SizeType delta)
    {
      const auto size = sz;
      if (size < delta)
        error_size_underflow();
      
      const auto bound = size - delta;
      sz = bound;
      destroy_range(data() + bound, data() + size);
    }

    void copy_from(const This &other)
    {
      const auto sz1 = sz,
                 sz2 = other.sz;

      const auto d1  = data();
      const auto d2  = other.data();

      if (sz2 < sz1)
      {
        shrink(sz1 - sz2);
        std::copy(d2, d2 + sz2, d1);  
      }
      else
      {
        sz = sz2;
        const auto d2u = d2 + sz1;
        std::copy(d2, d2u, d1);
        std::uninitialized_copy(d2u, d2 + sz2, d1 + sz1);
      }
    }

    void move_from(This &other)
    {
      const auto sz1 = sz,
                 sz2 = other.sz;

      const auto d1  = data();
      const auto d2  = other.data();

      if (sz2 < sz1)
      {
        shrink(sz1 - sz2);
        std::move(d2, d2 + sz2, d1);  
      }
      else
      {
        sz = sz2;
        const auto d2e = d2 + sz2;
        auto d2u = d2 + sz1,        
             d1d = d1 + sz1;

        std::move(d2, d2u, d1);
        while (d2u != d2e)
          new((void*)d1d++) T(std::move(*d2u++));
      }

      other.clear();
    }

    /// insert range implementation for random access iterators
    template <class RanIt>
    T* insert_impl(const T *pos, RanIt begin, RanIt end, std::random_access_iterator_tag)
    {
      const auto d = data();
      const auto n = end - begin;
      const auto i = pos - d;
      const auto size = static_cast<decltype(i + n)>(sz);
      check_index(i);
      grow(n);
      
      const auto dsrc = d + i,
                 ddst = dsrc + n,
                 dend = d + size;

      uninitialized_memmove(dsrc, dend, ddst, do_not_destroy);
      // after moving we can have two parts: initialized all elements and uninitialized "jumped over" elements
      if (size < i + n) // both parts are present
      {
        const auto init_size = size - i;
        const auto uninit = begin + init_size;
        std::uninitialized_copy(uninit, end, std::copy(begin, uninit, dsrc));
      }  
      else // only the initialized part is present
      {
        std::copy(begin, end, dsrc);
      }

      return ddst;
    }

    /// insert range implementation for input iterators
    template <class InIt>
    T* insert_impl(const T *pos, InIt begin, InIt end, std::input_iterator_tag)
    {
      if (begin == end)
        return const_cast<T*>(pos);

      // there's something to insert
      const auto size = sz;
      const auto dbegin = data(),
                 dend = dbegin + sz,
                 die = dbegin + MaxSize;

            auto d = dend;
      const auto i = pos - dbegin;
      check_index(i);

      // first, try to insert it in the end
      for (;;)
      {
        if (d == die)
        {
          sz = MaxSize;
          error_size_overflow();
        }

        new((void*)d++) T(*begin);
        if (++begin == end)
        {
          sz = d - dbegin;
          break;
        }
      }

      // second, rotate the inserted range to pos
      const auto dpos = dbegin + i;
      std::rotate(dpos, dend, d);
      return dpos + (d - dend);      
    }
  };

}


/// bounded array built over in-place stored static array class
template 
  <
    class T,                        ///< type of elements of the array
    std::size_t MaxSize,            ///< maximal size of the storage
    bool SizeChecksOn = true,       ///< correctness of size-changing operations is checked and exceptions are thrown
    class SizeType = std::size_t,   ///< type used to count used storage
    class DiffType = std::ptrdiff_t ///< type used to represent difference of iterators
  >
class Bounded_array
  : public std::conditional<std::is_pod<T>::value, // is good enough? duplicated below
      impl::Bounded_array_trivial<T, MaxSize, SizeType, SizeChecksOn>,
      impl::Bounded_array_non_trivial<T, MaxSize, SizeType, SizeChecksOn>>::type
{
public:

  // trivial STL-compatibility things...

  using value_type      = T;                  ///< type of stored elements (T)
  using reference       = value_type&;        ///< value_type&
  using const_reference = const value_type&;  ///< const value_type&
  using pointer         = value_type*;        ///< value_type*
  using const_pointer   = const value_type*;  ///< const value_type*
  using size_type       = SizeType;           ///< type representing size (SizeType)
  using difference_type = DiffType;           ///< type representing pointers difference

  using iterator         = pointer;        ///< the same as pointer
  using const_iterator   = const_pointer;  ///< the same as const_pointer
  using reverse_iterator = 
    std::reverse_iterator<iterator>;       ///< reverse iterator based upon iterator
  using const_reverse_iterator =
    std::reverse_iterator<const_iterator>; ///< reverse iterator based upon const_iterator

  // constructors and assignments

  using This = Bounded_array<T, MaxSize, SizeChecksOn, SizeType, DiffType>;
  using Base = typename
    std::conditional<std::is_pod<T>::value,
      impl::Bounded_array_trivial<T, MaxSize, SizeType, SizeChecksOn>,
      impl::Bounded_array_non_trivial<T, MaxSize, SizeType, SizeChecksOn>>::type;

  Bounded_array() {}
  Bounded_array(const This &other)
    : Base(other.to_base()) {}

  Bounded_array(This &&other)
    : Base(std::move(other.to_base())) {}

  template <class T>
  Bounded_array(std::initializer_list<T> inl)
    : Base(inl) {}


  This& operator=(const This &other)
  {
    to_base() = other.to_base();
    return *this;
  }

  This& operator=(This &&other)
  {
    to_base() = std::move(other.to_base());
    return *this;
  }

  This& operator=(std::initializer_list<T> inl)
  {
    to_base() = inl;
    return *this;
  }


  // inherited

  using Base::push_back;
  using Base::emplace_back;
  using Base::pop_back;
  using Base::insert;
  using Base::erase;
  using Base::emplace;
  using Base::assign;
  using Base::assign_n;
  using Base::clear;
  using Base::resize;
  
  /// swap with other contents
  void swap(This &other)
  {
    to_base().swap(other);
  }

  /// check if the array is empty
  bool empty() const
  {
    return sz == 0;
  }

  /// get size of the array
  size_type size() const
  {
    return sz;
  }

  /// get maximal possible size of this array
  size_type max_size() const
  {
    static_assert (static_cast<std::size_t>(static_cast<size_type>(MaxSize)) == MaxSize,
      "avo::Bounded_array error: MaxSize too big for SizeType provided");
    return static_cast<size_type>(MaxSize);
  }

  /// get a pointer to the array
  pointer data()
  {
    return reinterpret_cast<pointer>(bin);
  }

  /// get a const pointer to the array
  const_pointer data() const
  {
    return reinterpret_cast<const_pointer>(bin);
  }

  /// the beginning of the range of the stored objects
  iterator begin()
  {
    return data();
  }

  /// the beginning of the range of the stored objects, read-only
  const_iterator begin() const
  {
    return data();
  }

  /// the beginning of the range of the stored objects, read-only
  const_iterator cbegin() const
  {
    return begin();
  }

  /// the end of the range of the stored objects
  iterator end()
  {
    return data() + size();
  }

  /// the end of the range of the stored objects, read-only
  const_iterator end() const
  {
    return data() + size();
  }

  /// the end of the range of the stored objects, read-only
  const_iterator cend() const
  {
    return end();
  }

  /// the beginning of the reversed range of the stored objects
  reverse_iterator rbegin()
  {
    return reverse_iterator(end());
  }

  /// the beginning of the reversed range of the stored objects, read-only
  const_reverse_iterator rbegin() const
  {
    return const_reverse_iterator(end());
  }

  /// the beginning of the reversed range of the stored objects, read-only
  const_reverse_iterator crbegin() const
  {
    return rbegin();
  }

  /// the end of the reversed range of the stored objects
  reverse_iterator rend()
  {
    return reverse_iterator(begin());
  }

  /// the end of the reversed range of the stored objects, read-only
  const_reverse_iterator rend() const
  {
    return const_reverse_iterator(begin());
  }

  /// the end of the reversed range of the stored objects, read-only
  const_reverse_iterator crend() const
  {
    return rend();
  }

  
  /// get an element of the array by its index, unchecked
  reference operator[](const size_type index)
  {
    return (data())[index];
  }

  /// get an element of the array by its index, unchecked, read-only
  const_reference operator[](const size_type index) const
  {
    return (data())[index];
  }

  /// get an element of the array by its index, checked
  reference at(const size_type index)
  {
    check_index(index);
    return (*this)[index];
  }

  /// get an element of the array by its index, checked, read-only
  const_reference at(const size_type index) const
  {
    check_index(index);
    return (*this)[index];
  }
  
  /// access the first (index 0) element of the array
  reference front()
  {
    return *data();
  }

  /// access the first (index 0) element of the array, read-only
  const_reference front() const
  {
    return *data();
  }

  /// access the last element of the array
  reference back()
  {
    return (data())[size() - 1];
  }

  /// access the last element of the array, read-only
  const_reference back() const
  {
    return (data())[size() - 1];
  }

private:

  Base& to_base() { return *this; }
  const Base& to_base() const { return *this; }
};


/// swap contents (linear time)
template <class T, std::size_t MaxSize, bool SizeChecksOn, class SizeType, class DiffType>
inline void swap(
  Bounded_array<T, MaxSize, SizeChecksOn, SizeType, DiffType> &a,
  Bounded_array<T, MaxSize, SizeChecksOn, SizeType, DiffType> &b)
{
  a.swap(b);
}


/// elementwise operator== comparison
template <class T, std::size_t MaxSize, bool SizeChecksOn, class SizeType, class DiffType>
inline bool operator==(
  const Bounded_array<T, MaxSize, SizeChecksOn, SizeType, DiffType> &a,
  const Bounded_array<T, MaxSize, SizeChecksOn, SizeType, DiffType> &b)
{
  if (a.size() != b.size())
    return false;
  return std::equal(a.begin(), a.end(), b.begin());
}

template <class T, std::size_t MaxSize, bool SizeChecksOn, class SizeType, class DiffType>
inline bool operator!=(
  const Bounded_array<T, MaxSize, SizeChecksOn, SizeType, DiffType> &a,
  const Bounded_array<T, MaxSize, SizeChecksOn, SizeType, DiffType> &b)
{
  return !(a == b);
}

/// lexicographical comparison using operator<(T, T)
template <class T, std::size_t MaxSize, bool SizeChecksOn, class SizeType, class DiffType>
inline bool operator<(
  const Bounded_array<T, MaxSize, SizeChecksOn, SizeType, DiffType> &a,
  const Bounded_array<T, MaxSize, SizeChecksOn, SizeType, DiffType> &b)
{
  return std::lexicographical_compare(a.begin(), a.end(), b.begin(), b.end());
}

template <class T, std::size_t MaxSize, bool SizeChecksOn, class SizeType, class DiffType>
inline bool operator<=(
  const Bounded_array<T, MaxSize, SizeChecksOn, SizeType, DiffType> &a,
  const Bounded_array<T, MaxSize, SizeChecksOn, SizeType, DiffType> &b)
{
  return !(b < a);
}

template <class T, std::size_t MaxSize, bool SizeChecksOn, class SizeType, class DiffType>
inline bool operator>(
  const Bounded_array<T, MaxSize, SizeChecksOn, SizeType, DiffType> &a,
  const Bounded_array<T, MaxSize, SizeChecksOn, SizeType, DiffType> &b)
{
  return b < a;
}

template <class T, std::size_t MaxSize, bool SizeChecksOn, class SizeType, class DiffType>
inline bool operator>=(
  const Bounded_array<T, MaxSize, SizeChecksOn, SizeType, DiffType> &a,
  const Bounded_array<T, MaxSize, SizeChecksOn, SizeType, DiffType> &b)
{
  return !(a < b);
}

}
