﻿/*
 * This file is part of AV Orchid.
 * Copyright (c) 2014, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * @file  CoordinateGrid2.hpp
 * @brief класс, инкапсулирующий средства рисования координатной сетки на плоскости
 */
#pragma once
#include "Prim2.hpp"
#include "Bbox2.hpp"
#include "Color.hpp"
#include <cmath>

namespace avo
{

/// класс, инкапсулирующий средства рисования координатной сетки на плоскости
template <class Float>
class Coordinate_grid_2
{
public:
  using Bbox_2 = avo::Bbox_2<Float>;
  using Vec_2  = avo::Vec_2<Float>;
  using Prim_2 = avo::Prim_2<Float>;

  Color grid_color; ///< цвет линий координатной сетки
  Color axes_color; ///< цвет осей

  Coordinate_grid_2()
    : grid_color(GREY, 0.5f),
      axes_color(WHITE),
      grid(GL_LINES),
      axes(GL_LINES)
  {}


  /// пересчитать координатную сетку для заданного бокса
  void update(const Bbox_2 &box)
  {
    update_grid(box);
    update_axes(box);
  }

  /// нарисовать
  void paint() const
  {
    grid_color.select();
    grid.paint();
    axes_color.select();
    axes.paint();
  }

private:
  Prim_2 grid; ///< координатная сетка
  Prim_2 axes; ///< оси

  // пересчитать координатную сетку
  void update_grid(const Bbox_2 &box)
  {
    const Float
    h = box.height(),
    w = box.width(),
    scale = std::floor(std::log10(std::fmin(h, w))), // fmin из C99
    step  = std::pow(10.0, scale),
    invstep = Float(1.0) / step
    ;

    auto &gv = grid.vertices();
    gv.clear();

    const Float
      xbeg = step * std::ceil(box.pmin.x * invstep),
      xend = step * std::ceil(box.pmax.x * invstep)
      ;

    // вертикальные
    for (Float x = xbeg; x < xend; x += step)
    {
      gv.emplace_back(x, box.pmin.y);
      gv.emplace_back(x, box.pmax.y);
    }

    const Float
      ybeg = step * std::ceil(box.pmin.y * invstep),
      yend = step * std::ceil(box.pmax.y * invstep)
      ;

    // горизонтальные
    for (Float y = ybeg; y < yend; y += step)
    {
      gv.emplace_back(box.pmin.x, y);
      gv.emplace_back(box.pmax.x, y);
    }
  }

  // пересчитать оси
  void update_axes(const Bbox_2 &box)
  {
    auto &av = axes.vertices();
    av.clear();

    if (box.pmin.x < 0.0 && 0.0 < box.pmax.x)
    {
      av.emplace_back(0.0, box.pmin.y);
      av.emplace_back(0.0, box.pmax.y);
    }

    if (box.pmin.y < 0.0 && 0.0 < box.pmax.y)
    {
      av.emplace_back(box.pmin.x, 0.0);
      av.emplace_back(box.pmax.x, 0.0);
    }
  }
};

}
