﻿/*
 * This file is part of AV Orchid.
 * Copyright (c) 2014, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#pragma once
#include <cstddef>
#include <new>
#include <string>
#include <fstream>

namespace avo
{

/// прочитать содержимое файла с заданным именем в строку, ошибки передаются в поток err
std::string file_to_string(const char *fn, std::ostream &err)
{
  // попробуем открыть нужный файл
  std::ifstream input_file(fn, std::ios::binary);
  if (!input_file) // ошибка, файл открыть не удалось
  {
    err << "File \"" << fn << "\" can not be opened.\n";
    return {};
  }

  // установить позицию чтения файла в конец файла
  input_file.seekg(0, std::ios::end);
  // получить численное значение позиции чтения (т.е. размер файла в байтах)
  const std::streamsize input_file_size = input_file.tellg();

  // переменная, в которую будем считывать содержимое файла
  std::string result;
  if (input_file_size >= result.max_size() - 1) // ошибка, файл слишком велик
  {
    err << "File \"" << fn << "\" is too big (" << input_file_size << " bytes).\n";
    return {};
  }

  // подготовить строку нужного размера
  try
  {
    const auto str_size = static_cast<std::size_t>(input_file_size);
    result.reserve(str_size + 1);
    result.resize(str_size, '\0');
  }
  catch (const std::bad_alloc &e)
  {
    err << "Failed to load file \"" << fn << "\": not enough memory ("
        << e.what() << ").\n";
    return {};
  }

  // установить позицию чтения файла в начало
  input_file.seekg(std::ios::beg);
  // считать содержимое в строку (в подготовленное для этого место)
  // по стандарту C++11 символы строки хранятся в непрерывном массиве
  input_file.read(&result[0], input_file_size);
  return result; // перемещение без копирования в C++11
}

}
