﻿/*
 * This file is part of AV Orchid.
 * Copyright (c) 2014, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * @file Prim2.hpp
 * @author Dmitri Kuvshinov
 * @brief 2D GL1.1 primitive with floating point coordinates
 */
#pragma once
#include <GL/freeglut.h>
#include "Vec2.hpp"
#include <vector>

namespace avo
{

/// Двумерный геометрический примитив средствами GL1.1 (массив вершин).
template <class Float>
class Prim_2
{
  static_assert (std::is_same<Float, float>::value || std::is_same<Float, double>::value,
    "Prim2: only double and float are supported as a Scalar type.");

public:
  using Vec_2 = Vec_2<Float>;
  using Vertices = std::vector<Vec_2>;
  using This = Prim_2<Float>;

  /// по умолчанию -- просто набор точек
  explicit Prim_2(GLenum mode = GL_POINTS)
    : mode_(mode) {}

  /// инициализировать из последовательности, режим по умолчанию -- набор точек
  template <class ForIt>
  Prim_2(ForIt first, ForIt last, GLenum mode = GL_POINTS)
    : mode_(mode), data_(first, last) {}

  // compiler should define the following ctors and op=s automatically, but MSVC2010 doesn't do it

  Prim_2(const This &other)
    : mode_(other.mode_), data_(other.data_) {}
  Prim_2(This &&other)
    : mode_(other.mode_), data_(std::move(other.data_)) {}

  This& operator=(const This &other)
  {
    mode_ = other.mode_;
    data_ = other.data_;
    return *this;
  }

  This& operator=(This &&other)
  {
    mode_ = other.mode_;
    data_ = std::move(other.data_);
    return *this;
  }

  /// обменять содержимое
  void swap(This &other)
  {
    std::swap(mode_, other.mode_);
    data_.swap(other.data_);
  }

  /// получить доступ к вершинам примитива
  Vertices& vertices() { return data_; }
  const Vertices& vertices() const { return data_; }

  /// узнать текущий режим
  GLenum mode() const { return mode_; }
  /// установить другой режим
  void mode(GLenum new_mode) { mode_ = new_mode; }

  /// отрисовка примитива с произвольным режимом
  void paint(GLenum cur_mode) const
  {
    if (data_.empty())
      return;
    // Вообще, в большинстве случаев данный код будет приводить к постоянному преобразованию
    // массива double'ов в массив float'ов драйвером, что не есть хорошо с точки зрения производительности.
    glVertexPointer(2, (sizeof(Float) == sizeof(float)? GL_FLOAT: GL_DOUBLE), sizeof(Vec_2), &(data_[0]));
    glDrawArrays(cur_mode, 0, data_.size());
    // Альтернативы: преобразовать заранее самостоятельно (примитив хранит float'ы),
    // придумать хитрый способ обработки значений для достижения на float'ах точности отрисовки,
    // соответствующей double'ам.
  }

  /// отрисовка примитива с режимом по умолчанию
  void paint() const { paint(mode_); }

private:
  /// режим отрисовки примитива OpenGL (тот же смысл, что у параметра glBegin)
  GLenum mode_;
  /// вершины, составляющие примитив (только пары координат)
  Vertices data_;
};


template <class Float>
inline void swap(Prim_2<Float> &a, Prim_2<Float> &b)
{
  a.swap(b);
}

}
