/*
 * This file is part of AV Orchid.
 * Copyright (c) 2013, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * @file   memmove.hpp
 * @brief  typesafe memory move functions to replace std::memmove in some situations
 * @author Dmitri Kuvshinov
 *
 */
#pragma once

#include <cstring>
#include <memory>
#include <iterator>
#include <type_traits>
#include <utility>


namespace avo
{

/// typesafe memmove (bitwise, dispatches to std::memmove) for types with trivial copy
template <class T>
inline void memmove(T *begin, T *end, T *dest)
{
  static_assert(std::has_trivial_copy<T>::value,
    "avo::memmove: memmove must be applied to types with trivial copy");
  std::memmove(dest, begin, (end - begin) * sizeof(T));
}

/// destroy a range of objects
template <class It, class Alloc>
inline void destroy_range(It begin, It end, Alloc &&alloc)
{
  for (; begin != end; ++begin)
    alloc.destroy(alloc.address(*begin));
}

/// destroy a range using std::allocator
template <class It>
inline void destroy_range(It begin, It end)
{
  destroy_range(begin, end,
    std::allocator<typename std::iterator_traits<It>::value_type>());
}


/// select to destroy or not to destroy, and pronounce it clearly
enum DestroyOptions
{
  do_not_destroy,
  do_destroy
};

/// memory move into uninitialized range, accepts only random access iterators
/**
 * Option dopt may be used to explicitly suppress destruction of removed object,
 * which is the default action (so moving into uninitialized range leaves the 
 * non-overlapping part of source range uninitialized).
 */
template <class RanIt, class Alloc>
void uninitialized_memmove
  (RanIt begin, RanIt end, RanIt dest, Alloc &&alloc, DestroyOptions dopt = do_destroy)
{
  static_assert(
    std::is_same<
      typename std::iterator_traits<RanIt>::iterator_category,
      std::random_access_iterator_tag>::value,
    "avo::memmove: uninitialized_memmove accepts only random access iterators");

  // there are four possibilities:
  //   simplest one: end <= dest || dest + (end - begin) <= begin;
  //   move forward: begin < dest;
  //   move backward: dest < begin;
  //   no action: dest == begin.
  const auto size = end - begin;
  if (end <= dest || (dest + size) <= begin)
  {
    for (RanIt p = begin; p != end; ++dest, ++p)
      alloc.construct(alloc.address(*dest), std::move(*p));
    if (dopt == do_destroy)
      destroy_range(begin, end, std::forward<Alloc>(alloc));
  }
  else if (begin < dest)
  {
    const auto bound = end - dest;
    auto i = size - 1;
    for (; i >= bound; --i) // uninitialized part
      alloc.construct(alloc.address(dest[i]), std::move(begin[i]));
    for (; i > -1; --i) // initialized part
      dest[i] = std::move(begin[i]);
    if (dopt == do_destroy)
      destroy_range(begin, dest, std::forward<Alloc>(alloc));
  }
  else if (dest < begin)
  {
    const auto bound = begin - dest;
    auto i = (decltype(bound))0;
    for (; i < bound; ++i) // uninitialized part
      alloc.construct(alloc.address(dest[i]), std::move(begin[i]));
    for (; i < size; ++i) // initialized part
      dest[i] = std::move(begin[i]);
    if (dopt == do_destroy)
      destroy_range(dest + size, end, std::forward<Alloc>(alloc));
  }
  else
  {
    // do nothing, cause dest == begin here
  }
}

/// memory move into uninitialized range, accepts only random access iterators
/**
 * Option dopt may be used to explicitly suppress destruction of removed object,
 * which is the default action (so moving into uninitialized range leaves the 
 * non-overlapping part of source range uninitialized).
 */
template <class RanIt>
inline void uninitialized_memmove
  (RanIt begin, RanIt end, RanIt dest, DestroyOptions dopt = do_destroy)
{
  uninitialized_memmove(begin, end, dest,
    std::allocator<typename std::iterator_traits<RanIt>::value_type>(),
    dopt);
}


/// memory move into initialized range, accepts only random access iterators
/**
 * Option dopt may be used to explicitly demand destruction of removed object,
 * which is not the default action (so moving into initialized range leaves the 
 * non-overlapping part of source range initialized as well).
 */
template <class RanIt, class Alloc>
void initialized_memmove
  (RanIt begin, RanIt end, RanIt dest, Alloc &&alloc, DestroyOptions dopt = do_not_destroy)
{
  static_assert(
    std::is_same<
      typename std::iterator_traits<RanIt>::iterator_category,
      std::random_access_iterator_tag>::value,
    "BoundedArray: initialized_memmove accepts only random access iterators");

  // there are three possibilities:
  //   move forward: begin < dest;
  //   move backward: dest < begin;
  //   no action: dest == begin.
  const auto size = end - begin;
  if (begin < dest)
  {
    for (auto i = size - 1; i > -1; --i)
      dest[i] = std::move(begin[i]);
    if (dopt == do_destroy)
      destroy_range(begin, std::min(end, dest), std::forward<Alloc>(alloc));
  }
  else if (dest < begin)
  {
    for (auto i = (decltype(size))0; i < size; ++i)
      dest[i] = std::move(begin[i]);
    if (dopt == do_destroy)
      destroy_range(std::max(begin, dest + size), end, std::forward<Alloc>(alloc));
  }
  else
  {
    // do nothing, cause dest == begin here
  }
}

/// memory move into initialized range, accepts only random access iterators
/**
 * Option dopt may be used to explicitly demand destruction of removed object,
 * which is not the default action (so moving into initialized range leaves the 
 * non-overlapping part of source range initialized as well).
 */
template <class RanIt>
inline void initialized_memmove
  (RanIt begin, RanIt end, RanIt dest, DestroyOptions dopt = do_not_destroy)
{
  initialized_memmove(begin, end, dest,
    std::allocator<typename std::iterator_traits<RanIt>::value_type>(),
    dopt);
}

}
