﻿/*
 * This file is part of AV Orchid.
 * Copyright (c) 2013, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * Пример: Manipulator
 * Автор: Кувшинов Д.Р. https://evetro.wordpress.com/
 * Используемые библиотеки: FreeGlut, OpenGL.
 *
 * Имитация движения манипулятора.
 * Нажатие ENTER -- смена направления движения манипулятора.
 * WASDXZ и стрелки -- поворот камеры.
 * Нажатие ESC/Q -- выход из программы.
 */
#include <GL/freeglut.h>
#include <cstdlib>        // rand/srand
#include <iostream>

#include <avo/Basics.hpp>
#include <avo/Timer.hpp>
#include <avo/Color.hpp>    // RGBA-цвет в 128-битном формате (компоненты типа float)
#include <avo/SphericalCamera.hpp> // камера в сферических координатах ("турельного" типа)

using avo::Color;
using avo::random;

//
// базовые определения
//

/// "универсальная" функция "возведение в квадрат"
template <class X>
inline X square(const X x)
{ return x * x; }

/// интерполяция путём решения уравнения x'' = a sgn(2t - t0 - t1), x'(t0) = x'(t1) = 0, x(t0) = x0, x(t1) = x1
/*
 * Сначала (до середины временного отрезка происходит разгон с постоянным ускорением),
 * затем торможение с тем же ускорением.
 */
template <class X = double>
class Dd_interpolator
{
public:
  Dd_interpolator() 
    : x0(X()), x1(X()), accel_2(X()), t0(0.0), t1(0.0), t2(0.0) {}

  Dd_interpolator
  ( const double t0, const double t1, 
    const X     &x0, const X     &x1 )
    : t0(t0), t1(t1), x0(x0), x1(x1),
      t2      (0.5 * (t1 + t0)),
      accel_2 (t0 != t1? 2.0 * (x1 - x0) / square(t1 - t0) : 0.0) 
  {}

  /// вычислить позицию для момента t, условие t0 <= t <= t1 не проверяется
  X operator()(const double t) const
  {
    return t <= t2?
      x0 + square(t - t0) * accel_2: 
      x1 - square(t1 - t) * accel_2;
  }

private:
  X x0, x1, accel_2; // начальная и конечная позиции, ускорение пополам
  double t0, t1, t2; // начальный и конечный моменты времени, середина временного отрезка
};

/// математическая модель манипулятора
class Manipulator_model
{
public:
  /// состояние: 4 угла и 2 смещения
  struct State
  {
    double phi,   ///< поворот всей конструкции вокруг вертикальной оси
           psi,   ///< поворот "руки" вокруг призмы
           eta,   ///< поворот второго звена вокруг оси первого звена
           theta, ///< поворот "держателя"
           x,     ///< смещение первого звена относительно крепления
           y;     ///< смещение второго звена относительно крепления
  };

  Manipulator_model() {}

  Manipulator_model
  ( const double t0, const double t1,
    const State &s0, const State &s1 )
    : phi   (t0, t1, s0.phi,    s1.phi  ),
      psi   (t0, t1, s0.psi,    s1.psi  ),
      eta   (t0, t1, s0.eta,    s1.eta  ),
      theta (t0, t1, s0.theta,  s1.theta),
      x     (t0, t1, s0.x,      s1.x    ),
      y     (t0, t1, s0.y,      s1.y    )
  {}

  /// вычислить состояние для текущего t, условие t0 <= t <= t1 не проверяется
  State operator()(const double t) const
  {
    State result = { phi(t), psi(t), eta(t), theta(t), x(t), y(t) };
    return result;
  }

private:
  Dd_interpolator<double> phi, psi, eta, theta, x, y;
};


std::istream& operator>>(std::istream &is, Manipulator_model::State &s)
{
  return is >> s.phi >> s.psi >> s.x >> s.y;
}


/// класс, отвечающий за отрисовку манипулятора для заданного состояния модели
class Manipulator
{
public:
  /**
   *               wid1            wid2
   *               l1  l1len       l3 l3len
   *               j1              wid3
   * wid2 l2  wid3 j2          l2  j3 j3len
   *               j2len           l3
   *      l2len                    l4 l4len
   */
  struct Dimensions
  {
    double width[3];     ///< "диаметры": вертикального базового цилиндра, звена, крепления звена
    double link_len[4];  ///< длины "звеньев", включая базовый цилиндр и держатель
    double joint_len[2]; ///< длины креплений звеньев
  };

  Manipulator(const Dimensions &dims)
    : d(dims) {}

  /// инициализировать цвета и модели элементов манипулятора
  static void init()
  {
    // цвета
    LINK_COLOR  = Color(0.6f); // серый
    JOINT_COLOR = avo::ORANGE;

    // сгенерировать модель цилиндра (полигональная аппроксимация)
    CYLINDER = glGenLists(1);        // получить идентификатор нового списка отрисовки
    glNewList(CYLINDER, GL_COMPILE); // инициализировать этот список (режим "компиляции" команд), начать запись команд
    glutSolidCylinder(0.5, 1.0, 16, 4); // выполнить рисование "кружка" (16 разбиений по "параллелям", 8 разбиений по "направляющим")
    glEndList();                        // закончить запись команд, скомпилировать список

    // модель призмы
    PRISM = glGenLists(1);
    glNewList(PRISM, GL_COMPILE);
    const float rsqrt5 = std::sqrt(0.2f);
    const float v[6][3] = { {-0.5f, -0.5f,  0.f}, {-0.5f,  0.5f,  0.f},
          /* vertices */    { 0.5f,  0.5f,  0.f}, { 0.5f, -0.5f,  0.f},
                            {-0.5f,  0.f,  -1.f}, { 0.5f,  0.f,  -1.f} };

    glBegin(GL_QUADS);
    glNormal3f(0.f, 0.f, 1.f);
    glVertex3fv(&v[3][0]);
    glVertex3fv(&v[2][0]);
    glVertex3fv(&v[1][0]);
    glVertex3fv(&v[0][0]);

    glNormal3f(0.f, +2.f * rsqrt5, -rsqrt5);
    glVertex3fv(&v[1][0]);
    glVertex3fv(&v[2][0]);
    glVertex3fv(&v[5][0]);
    glVertex3fv(&v[4][0]);

    glNormal3f(0.f, -2.f * rsqrt5, -rsqrt5);
    glVertex3fv(&v[4][0]);
    glVertex3fv(&v[5][0]);    
    glVertex3fv(&v[3][0]);
    glVertex3fv(&v[0][0]);
    glEnd();

    glBegin(GL_TRIANGLES);
    glNormal3f(-1.f, 0.f, 0.f);
    glVertex3fv(&v[0][0]);
    glVertex3fv(&v[1][0]);
    glVertex3fv(&v[4][0]);

    glNormal3f(1.f, 0.f, 0.f);
    glVertex3fv(&v[2][0]);
    glVertex3fv(&v[3][0]);
    glVertex3fv(&v[5][0]);
    glEnd();
    glEndList();

    // модель держателя -- октаэдр
    HOLDER = glGenLists(1);
    glNewList(HOLDER, GL_COMPILE);
    glutSolidOctahedron(); // диаметр == 2
    glEndList();
  }

  /// нарисовать манипулятор для заданного состояния
  void paint(const Manipulator_model::State &state) const
  {
    glRotated(state.phi, 0., 0., 1.);
    glPushMatrix();
    // l1
    glScaled(d.width[0], d.width[0], d.link_len[0]);
    LINK_COLOR.select();
    glCallList(CYLINDER);
    glPopMatrix();
    glPushMatrix();
    // j1
    glScaled(d.width[0], d.width[0], d.width[0]);
    JOINT_COLOR.select();
    glCallList(PRISM);
    glPopMatrix();
    // j2
    glTranslated(0., 0., -d.width[0]);
    glRotated(state.psi, 1., 0., 0.);

    glRotated(-90., 1., 0., 0.);
    glTranslated(0., 0.5 * d.width[2], -0.5 * d.joint_len[0]);
    glPushMatrix();
    glScaled(d.width[2], d.width[2], d.joint_len[0]);
    glCallList(CYLINDER);
    glPopMatrix();
    // l2
    glTranslated(0., 0., -state.x);
    glPushMatrix();
    glScaled(d.width[1], d.width[1], d.link_len[1]);
    LINK_COLOR.select();
    glCallList(CYLINDER);
    glPopMatrix();
    // j3
    glRotated(-90., 1., 0., 0.);
    glRotated(state.eta, 0., -1., 0.);    
    glTranslated(0., -d.link_len[1] - 0.5 * d.width[2], -0.5 * d.joint_len[1]);
    glPushMatrix();
    glScaled(d.width[2], d.width[2], d.joint_len[1]);
    JOINT_COLOR.select();
    glCallList(CYLINDER);
    glPopMatrix();
    // l3
    glTranslated(0., 0., -state.y);
    glPushMatrix();
    glScaled(d.width[1], d.width[1], d.link_len[2]);
    LINK_COLOR.select();
    glCallList(CYLINDER);
    glPopMatrix();
    // l4
    glRotated(state.theta, 0., 0., 1.);
    glTranslated(0., 0., d.link_len[2]);
    glScaled(d.link_len[3], d.link_len[3], d.link_len[3]);
    glCallList(HOLDER);
  }

public:
  static GLuint CYLINDER, PRISM, HOLDER;
  static Color  LINK_COLOR, JOINT_COLOR;
  Dimensions d;
};


//
// глобальные переменные
//

Manipulator_model::State state;
Manipulator_model        model;
Manipulator::Dimensions dims  = { { 50, 25, 40 }, { 60, 280, 250, 30 }, { 60, 60 } };
Manipulator             manip (dims);

/// отрезок времени, на котором производится симуляция (в секундах системы пользователя)
double t0, t1;

/// объект камеры
avo::Spherical_camera camera;

/// static-данные требуется явно размещать в некоторой единице трансляции
GLuint Manipulator::CYLINDER,   Manipulator::PRISM,       Manipulator::HOLDER;
Color  Manipulator::LINK_COLOR, Manipulator::JOINT_COLOR;

//
// обработчики событий GLUT
//

/// изменение размеров окна, функция получает "новые размеры"
void reshape(int w, int h)
{
  // новая матрица проекции
  glMatrixMode(GL_PROJECTION);
  glPopMatrix();    // удалить предыдущую матрицу из стека матриц (стек матриц используется Spherical_camera)
  glLoadIdentity(); // загрузить единичную матрицу
  /* угол зрения по вертикали, соотношение сторон экрана, минимальная дальность, максимальная дальность */
  gluPerspective(75, double(w) / double(h), 16.0, 4096.0); // домножить на матрицу перспективной коррекции
  glPushMatrix();   // затолкнуть текущую матрицу в стек матриц (копируется, текущая остаётся на вершине)
  camera.apply();   // применить к текущей матрице матрицу камеры

  // установить координаты окна просмотра (вся клиентская область)
  glViewport(0, 0, w, h);
}

/// перерисовка
void repaint()
{
  // очистить окно, закрасив его цветом очистки, сбросить буфер глубины
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glLoadIdentity();
  manip.paint(state);
  glFinish(); // дождаться выполнения всех OpenGL команд
  glutSwapBuffers(); // обменять буферы кадра, чтобы вывести в окно новый кадр (двойная буферизация)
}


/// новая цель движения, t0 -- "опорный момент времени", должен быть установлен заранее
void reset_movement()
{
  t1 = t0 + random(5., 10.);
  const Manipulator_model::State s1 = 
  {
    /* phi */ random(-359., 359.),
    /* psi */ random( -30.,  30.),
    /* eta */ random(-359., 359.),
    /* th  */ random(-359., 359.),
    /* x   */ random(  0.,  200.),
    /* y   */ random(  0.,  180.) 
  };
  model = Manipulator_model(t0, t1, state, s1);
}


/// вычислить состояние сцены на следующий кадр
void frame_move()
{
  // текущее время
  const double t = avo::seconds();
  if (t <= t0) return;
  if (t <= t1)
  {
    state = model(t);
    // уведомим GLUT о необходимости перерисовать сцену
    glutPostRedisplay();
  }
  else
  {
    t0 = t  + random(0.5, 4.);
    reset_movement();
  }
}


/// обработать нажатие "обычной" клавиши (x, y -- координаты мыши в окне на момент нажатия)
void on_key(unsigned char key, int x, int y)
{
  switch (key)
  {
  case 'w': case 'W': case 'ц': case 'Ц': // ближе к началу координат
    camera.forward();
    break;

  case 's': case 'S': case 'ы': case 'Ы': // дальше от начала координат
    camera.backward();
    break;

  case 'a': case 'A': case 'ф': case 'Ф': // переместить влево
    camera.left();
    break;

  case 'd': case 'D': case 'в': case 'В': // переместить вправо
    camera.right();
    break;

  case 'z': case 'Z': case 'я': case 'Я':
    camera.up();
    break;

  case 'x': case 'X': case 'ч': case 'Ч':
    camera.down();
    break;

  case 27: case 'q': case 'Q': case 'й': case 'Й':
    glutLeaveMainLoop(); // покинуть цикл обработки сообщений
    return; // и покинуть данную функцию...

  case '\r': case '\n':
    t0 = avo::seconds(); // начиная с этого момента,
    reset_movement();    // двигаться по-другому
    break;

  default:
    return; // остальные игнорируются
  }

  camera.apply(); // применить матрицу камеры к матрице проекции
  glutPostRedisplay();
}


/// обработать нажатие "специальной" клавиши (x, y -- координаты мыши в окне на момент нажатия)
void on_special_key(int key, int x, int y)
{
  // движение камеры аналогично WASD, только с помощью стрелок на клавиатуре
  switch (key)
  {
  case GLUT_KEY_LEFT:
    camera.left();
    break;

  case GLUT_KEY_UP:
    camera.up();
    break;

  case GLUT_KEY_RIGHT:
    camera.right();
    break;

  case GLUT_KEY_DOWN:
    camera.down();
    break;

  default:
    return; // остальные игнорируются
  }

  camera.apply(); // применить матрицу камеры к матрице проекции
  glutPostRedisplay();
}


//
// main
//

int main(int argc, char **argv)
{
  // базовая инициализация
  glutInit(&argc, argv);  // обработка параметров командной строки
  std::srand((unsigned)(avo::seconds() * 1e+6) % 15485863);

  // создать окно с привязанным контекстом OpenGL
  // режим: 32-битный цвет RGBA, двойная буферизация DOUBLE, создать буфер глубины DEPTH, сглаживание мультисемплингом MULTISAMPLE
  glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH | GLUT_MULTISAMPLE);
  glutInitWindowPosition(0, 0);   // координаты верхнего левого угла окна на экране
  glutInitWindowSize(800, 600);   // полные ширина и высота окна
  glutCreateWindow("Manipulator motion"); // создать окно, в функцию передаётся заголовок создаваемого окна

  // параметры OpenGL для данного примера
  glClearColor(0.1f, 0.1f, 0.2f, 1.0f); // установить цвет очистки (фон)
  glMaterialf(GL_FRONT, GL_SHININESS, 50.0f);
  glEnable(GL_COLOR_MATERIAL);      // выбирать цвет материала командой glColor
  glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE); // этот цвет -- с обеих сторон полигона, ambient и diffuse компоненты

  glEnable(GL_CULL_FACE);  // включить отбрасывание полигонов, смотрящих "от" камеры (повёрнутых обратной стороной)
  glEnable(GL_DEPTH_TEST); // использовать буфер глубины по назначению
  glEnable(GL_NORMALIZE);  // форсировать нормирование нормалей (при масштабировании сфер происходит масштабирование и нормалей)
  glEnable(GL_LIGHTING);   // включить повершинное освещение
  glEnable(GL_LIGHT0);     // включить источник света №0
  glEnable(GL_LIGHT1);     // включить источник света №1

  // настроить источник света №1
  {
    float pos[] = { 0.25f, 0.5f, -1.0f, 0.0f };
    glLightfv(GL_LIGHT1, GL_POSITION, pos);    // позиция с w == 0 -- направление света
    float col[] = { 1.0f, 1.0f, 0.75f, 1.0f }; // цвет
    glLightfv(GL_LIGHT1, GL_DIFFUSE,  col);    // diffuse составляющая
  }

  // рассеянный свет, идущий "со всех сторон"
  {
    float ambientColor[] = { 0.1f, 0.1f, 0.125f, 1.0f };
    glLightModelfv(GL_LIGHT_MODEL_AMBIENT, ambientColor); // выбрать цвет
  }

  Manipulator::init();

  // привязка функций-обработчиков событий
  glutIdleFunc(frame_move);            // обработка движения
  glutDisplayFunc(repaint);            // перерисовка
  glutReshapeFunc(reshape);            // изменение размеров окна
  glutKeyboardFunc(on_key);         // обработка нажатий "обычных" клавиш
  glutSpecialFunc(on_special_key);  // обработка нажатий "специальных" клавиш

  // старт
  std::cout << "Starting Manipulator...\n";
  t0 = avo::seconds();
  glutMainLoop(); // вход в цикл обработки событий
  std::cout << "\nbye!";
  return 0;
}
