﻿/*
 * This file is part of AV Orchid.
 * Copyright (c) 2013, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * Пример: Plasma
 * Автор: Кувшинов Д.Р. https://evetro.wordpress.com/
 * Используемые библиотеки: FreeGlut, OpenGL.
 *
 * Эффект "плазмы". См. http://en.wikipedia.org/wiki/Plasma_effect
 * Демонстрирует использование табличных функций (без интерполяции),
 * шаблонов функций.
 *
 * Кроме эффекта плазмы демонстрирует билинейную интерполяцию:
 *  4 -- случайной карты;
 *  5 -- случайной карты температур с симуляцией процесса теплопереноса.
 *
 * Нажатие:
 * цифры -- выбор алгоритма;
 * P     -- показывать/скрыть палитру;
 * [     -- уменьшить число потоков OpenMP;
 * ]     -- увеличить число потоков OpenMP;
 * B     -- прогнать тест быстродействия без учёта вывода графики (геометрическое среднее трёх значений пикселей/с)
 * ESC/Q -- выход из программы.
 */
#include <GL/freeglut.h>

#include <cassert>
#include <cmath>
#include <iostream>
#include <algorithm>     // max

#include <memory>
#include <random>

#include <omp.h>

#include <avo/Pixel.hpp> // RGBA-цвет в 32-битном формате
#include <avo/Timer.hpp>

using avo::Pixel;

//
// Константы и глобальные переменные.
//

/// Размеры картинки: ширина и высота (для простоты зафиксированы).
int width = 800, height = 600;

/// Счётчик кадров (сбрасывается каждые 10с).
int frames;
/// Момент времени, с которого происходит накопление счётчика кадров.
double frames_reported_last;

/// Номер текущего кадра -- "время".
int t = 0;

/// Признак "показывать палитру".
bool show_palette;

/// Сама картинка.
Pixel *pixels;

/// Период целочисленной синусоиды.
const int SINE_PERIOD = 512;
// -- должен быть степенью двойки:
static_assert((SINE_PERIOD & (SINE_PERIOD - 1)) == 0, "Plasma: SINE_PERIOD must be power of 2");


/// Палитра, глобальный объект.
class Palette
{
public:
  /// Конструктор: заполнить палитру.
  Palette()
  {
    // Линейная модель, аналогично Hue в HSV
    // см. http://upload.wikimedia.org/wikipedia/commons/5/5d/HSV-RGB-comparison.svg?uselang=ru (здесь в обратном порядке)
    //     http://ru.wikipedia.org/wiki/HSV_(цветовая_модель)
    for (int i = 0; i < 64; ++i)
    {
      const uint8_t i4 = uint8_t(i * 4), inv_i4 = uint8_t(255 - i4);
      p[i      ]  = Pixel::rgb( 255   , 0     , i4     );  // красный  -> маджента 
      p[i +  64]  = Pixel::rgb( inv_i4, 0     , 255    );  // маджента -> синий
      p[i + 128]  = Pixel::rgb( 0     , i4    , 255    );  // синий    -> циан
      p[i + 192]  = Pixel::rgb( 0     , 255   , inv_i4 );  // циан     -> зелёный
      p[i + 256]  = Pixel::rgb( i4    , 255   , 0      );  // зелёный  -> жёлтый
      p[i + 320]  = Pixel::rgb( 255   , inv_i4, 0      );  // жёлтый   -> красный
    }
  }

  /// получить цвет по индексу
  Pixel operator[](const int index) const
  { return p[index]; }

private:
  Pixel p[384];
} palette; // конструктор вызывается автоматически


//
// Обработчики событий GLUT.
//

/// Перерисовка.
void repaint()
{        /* ширина,  высота, тип пикселя, тип компонента, указатель на картинку */
  glDrawPixels(width, height, GL_RGBA, GL_UNSIGNED_BYTE, pixels);

  // счётчик кадров в секунду, выводит среднее за 10с
  ++frames;
  const double dt = avo::seconds() - frames_reported_last;
  if (dt >= 10.0)
  {
    std::cout << frames / dt << '\n';
    frames = 0;
    frames_reported_last += dt;
  }

  glFinish(); // дождаться завершения обработки очереди команд OpenGL
}


/// Изменение размеров окна, функция получает новые размеры.
void reshape(int w, int h)
{
  // заменим старое содержимое pixels на новое
  delete[] pixels;
  pixels = new Pixel[(w < 384? 384: w) * h];

  // установим новые размеры картинки
  width   = w;
  height  = h;
    
  // установить новые координаты окна просмотра (вся клиентская область)
  glViewport(0, 0, w, h);
}


///////////////////////////////////////////////////////////////////////////////
// Plasma things

//
// Реализации синусоиды.
//


/// Непосредственно вычисляемая синусоида, отображает [0, SINE_PERIOD-1] -> [0, 255], продолжена периодически на все целые.
struct Sine // структура нужна, чтобы можно было просто передать тип шаблону plasma в качестве реализации
{           // "просто" функцию так передать нельзя
  int operator()(const int x) const
  {
    static const float PI2 = 6.283185307179586476925286766559f / SINE_PERIOD;
    const float nx = (x & (SINE_PERIOD - 1)) * PI2;
    return (int)(127.5f * (std::sin(nx) + 1.f));
  }
};

/// Табличная синусоида, отображает [0, SINE_PERIOD-1] -> [0, 255], продолжена периодически на все целые.
class Table_sine
{
public:
  static const int N = SINE_PERIOD;
  Table_sine()
  {
    Sine sine; // повторное использование ;)
    for (int i = 0; i < N; ++i)
      p[i] = sine(i);
  }

  int operator()(const int x) const
  { return p[x & (N - 1)]; }

private:
  int p[N];
};


//
// Реализации функции расстояния.
//

/// Вычисление евклидовой нормы по обычной формуле с использованием квадратного корня.
struct Sqrt_dist
{
  int operator()(int a, int b) const
  {
    return (int)std::sqrt(float(a * a + b * b));
  }
};

/// Грубая аппроксимация евклидовой нормы (изолиния -- восьмиугольник вместо окружности).
struct Crude_dist
{
  int operator()(int a, int b) const
  {
    a = std::abs(a);
    b = std::abs(b);
    return a < b? b + (a >> 1): a + (b >> 1);
  }
};


/// Рисование плазмы, использующее переданные функторы-реализации синуса и расстояния.
template <class SineImpl, class DistImpl>
void plasma()
{
  static const SineImpl sine;
  static const DistImpl dist;

  t++;
  // все константы взяты "с потолка"
  const int mwh = std::max(width, height) / 2,
            st1 = (mwh * sine(    t     +  17)) / 256,
            st2 = (mwh * sine(3 * t / 2 + 107)) / 256,
            st3 = (mwh * sine(3 * t / 8 + 210)) / 256,

  x0 = (st1 -     width / 7) * 3,   y0 = (2 * height / 3 - st3) * 2,
  x1 = (st2 + 3 * width / 4)    ,   y1 = (4 * height / 3 - st1) * 7,
  x2 = (6 * width / 7 - st3) * 2,   y2 = (    height / 2 - st1),
  x3 = (3 * width / 5 - st1)    ,   y3 = (st2 - st3) / 2;

  #pragma omp parallel for
  for (int y = 0; y < height; ++y)
  {
    Pixel *line = pixels + y * width;

    if (show_palette)
    {
      for (int x = 0; x < 384; ++x)
        line[x] = palette[x];
    }

    const int z0 = y - y0,  z1 = y - y1,
              z2 = y - y2,  z3 = y - y3;

    for (int x = show_palette? 384: 0; x < width; ++x)
    {	
      const int v0 = x - x0,  v1 = x - x1,
                v2 = x - x2,  v3 = x - x3;

      const int d0 = dist(v0, z0),
                d1 = dist(v1, z1),
                d2 = dist(v2, z2),
                d3 = dist(v3, z3);

      const int tv  = sine(d0) + sine(d1) + sine(d2) + sine(d3);
      line[x] = palette[(384 * tv) / 1024];
    }
  }

  glutPostRedisplay();
}

// TASK: вместо эффекта "плазмы" реализовать другой эффект, например, "бегущие волны".



///////////////////////////////////////////////////////////////////////////////
// Билинейная интерполяция, базовый и тестировочный код.
// https://ru.wikipedia.org/wiki/Билинейная_интерполяция

/// Базовый код для билинейной интерполяции.
namespace bilinear
{
  using namespace std;

  /// Класс "матрица", элементы которой интерполируются по двум направлениям.
  template <class T>
  class Matrix
  {
    unique_ptr<T[]> dt; // Привязан автоматически удаляемый динамический массив.
    size_t rs, cs, sz;  // Можно было бы использовать valarray<T>...

  public:
    Matrix()
      : rs(0), cs(0), sz(0) {}

    Matrix(size_t rows, size_t cols)
      : rs(0), cs(0), sz(0)
    {
      const auto szt = rows * cols;
      dt.reset(new T[szt]);
      rs = rows;
      cs = cols;
      sz = szt;
    }

    Matrix(const Matrix &other)
      : Matrix(other.rs, other.cs)
    {
      copy_n(other.data(), sz, data());
    }

    Matrix(Matrix &&other)
      : dt(move(other.dt)), rs(other.rs), cs(other.cs), sz(other.sz)
    {
      other.rs = 0;
      other.cs = 0;
      other.sz = 0;
    }

    Matrix& operator=(const Matrix &other)
    {
      if (this == &other)
        return *this;

      if (size() < other.size())
        return *this = Matrix<T>(other);

      rs = other.rs;
      cs = other.cs;
      sz = other.sz;
      copy_n(other.data(), sz, data());
      return *this;
    }

    Matrix& operator=(Matrix &&other)
    {
      if (this != &other)
      {
        dt = move(other.dt);
        rs = other.rs;
        cs = other.cs;
        sz = other.sz;
        other.rs = 0;
        other.cs = 0;
        other.sz = 0;
      }
      return *this;
    }

    Matrix& operator+=(const Matrix &other)
    {
      // Операторы += и т.п. здесь не используются,
      // данная реализация приведена в качестве примера.
      const auto as = size();      
      assert(as == other.size());
      const auto
        ap = data(),
        bp = other.data();

      for (size_t i = 0; i < as; ++i)
        ap[i] += bp[i];
      return *this;
    }

    bool empty() const { return sz == 0; }
    size_t size() const { return sz; }
    size_t rows() const { return rs; }
    size_t cols() const { return cs; }

    T* data() { return dt.get(); }
    const T* data() const { return dt.get(); }

    T& operator[](size_t idx)
    {
      assert(idx < sz);
      return dt[idx];
    }
    
    const T& operator[](size_t idx) const
    {
      assert(idx < sz);
      return dt[idx];
    }

    T* row(size_t idx)
    {
      assert(idx < rs);
      return data() + idx * cs;
    }
    
    const T* row(size_t idx) const
    {
      assert(idx < rs);
      return data() + idx * cs;
    }

    T& operator()(size_t i, size_t j)
    {
      assert(j < cs);
      return row(i)[j];
    }

    const T& operator()(size_t i, size_t j) const
    {
      assert(j < cs);
      return row(i)[j];
    }

    T* begin() { return data(); }
    T* end() { return begin() + size(); }

    const T* begin() const { return data(); }
    const T* end() const { return begin() + size(); }
  };


  template <class T1, class T2> inline
  bool operator==(const Matrix<T1> &a, const Matrix<T2> &b)
  {
    return a.rows() == b.rows() && a.cols() == b.cols()
        && equal(a.begin(), a.end(), b.begin());
  }

  template <class T1, class T2> inline
  bool operator!=(const Matrix<T1> &a, const Matrix<T2> &b)
  {
    return !(a == b);
  }


  /// Линейная интерполяция _0 и _1 с параметром u, который должен быть в диапазоне [0, 1].
  template <class T> inline
  T lerp(T u, T _0, T _1)
  {
    return _0 * (T(1) - u) + _1 * u;
  }

  /// Билинейная интерполяция: lerp(u, lerp(v, _00, _01), lerp(v, _10, _11)).
  /// Несмотря на морфему -линейная, данной функции соответствует квадрика.
  template <class T> inline
  T blerp(T u, T v, T _00, T _01, T _10, T _11)
  {
    const auto
      _0 = lerp(v, _00, _01),
      _1 = lerp(v, _10, _11);

    return lerp(u, _0, _1);
  }

  /// Вписать x в диапазон [min_x, max_x].
  template <class T> inline
  T clamp(T x, T min_x, T max_x)
  {
    return x < min_x? min_x
         : x > max_x? max_x
         : x;
  }


  /// Билинейная интерполяция карты.
  /// Выбирает 4 элемента, между которыми находится (x, y) и интерполирует дробными частями.
  /// m должна быть размера 2x2 или больше.
  template <class T> inline
  T sample(T x, T y, const Matrix<T> &m)
  {
    const ptrdiff_t
      x_size = m.cols(),
      y_size = m.rows();

    assert(x_size > 1 && y_size > 1);
    y = clamp(y, T(0), T(y_size - 1));
  
    auto i = min(ptrdiff_t(y), y_size - 2); // trunc
    y -= T(i); // now it's v

    const auto
      m0 = m.row(i),
      m1 = m0 + m.cols();

    x = clamp(x, T(0), T(x_size - 1));

    auto j = min(ptrdiff_t(x), x_size - 2); // trunc
    x -= T(j); // now it's u
    return blerp(x, y, m0[j], m1[j], m0[j+1], m1[j+1]);
  }


  /// Создать матрицу заданных размеров заполненную случайными значениями из диапазона [min_v, max_v]
  /// с равномерным распределением.
  template <class T>
  Matrix<T> make_random_matrix(size_t rows, size_t cols, T min_v = T(0), T max_v = T(1))
  {
    mt19937 rng(random_device{}());
    uniform_real_distribution<T> vg(min_v, max_v);

    Matrix<T> result(rows, cols);
    for (auto &item: result)
      item = vg(rng);
 
    return result;
  }
}


/// Карта температур для процесса симуляции теплопереноса.
bilinear::Matrix<float> temperatures;

/// Визуализация карты температур с помощью программно реализованной билинейной интерполяции.
void draw_temperatures()
{
  const float
    y_factor = float(temperatures.rows() - 1) / height,
    x_factor = float(temperatures.cols() - 1) / width;

  #pragma omp parallel for
  for (int y = 0; y < height; ++y)
  {
    Pixel *line = pixels + y * width;
    const auto yn = y_factor * y;

    if (show_palette)
    {
      for (int x = 0; x < 384; ++x)
        line[x] = palette[x];
    }

    for (int x = show_palette? 384: 0; x < width; ++x)
    {	
      const auto xn = x_factor * x;
      line[x] = palette[int(383.f * bilinear::sample(xn, yn, temperatures))];
    }
  }

  glutPostRedisplay();
}


///////////////////////////////////////////////////////////////////////////////
// Симуляция процесса теплопереноса.
// Использует для визуализации билинейную интерполяцию карты температур.

/// Коэффициенты теплопроводности (передачи энергии из ячейки в соседние).
const float heat_coeff[3][3] =
{
  { 1e-5, 2e-2, 1e-5 },
  { 1e-3, 0,    1e-3 },
  { 1e-5, 2e-2, 1e-5 }
};


/// Вспомогательная функция передачи энергии (температуры).
inline float conduct(float &to, float from, float coef)
{
  const float to_0 = to;
  if (to_0 > from)
    return 0.f;

  const float output = (from - to_0) * coef;
  to = to_0 + output;
  return output;
}


/// Один шаг симуляции теплопереноса.
void temperatures_step()
{
  const auto
    rows = temperatures.rows(),
    cols = temperatures.cols();

  bilinear::Matrix<float> nt = temperatures;
  // Simple, non-optimized (cellwise) algorithm.
  for (size_t i = 0; i < rows; ++i)
  {
    const auto
      tp = temperatures.row(i),
      np = nt.row(i);

    for (size_t j = 0; j < cols; ++j)
    {
      const float from = tp[j];
      float outcome = 0;
      if (i != 0)
      {
        if (j != 0)
          outcome += conduct(np[j-cols-1], from, heat_coeff[0][0]);
        outcome += conduct(np[j-cols], from, heat_coeff[0][1]);
        if (j != cols - 1)
          outcome += conduct(np[j-cols+1], from, heat_coeff[0][2]);
      }

      if (j != 0)
        outcome += conduct(np[j-1], from, heat_coeff[1][0]);
      if (j != cols - 1)
        outcome += conduct(np[j+1], from, heat_coeff[1][2]);

      if (i != rows - 1)
      {
        if (j != 0)
          outcome += conduct(np[j+cols-1], from, heat_coeff[2][0]);
        outcome += conduct(np[j+cols], from, heat_coeff[2][1]);
        if (j != cols - 1)
          outcome += conduct(np[j+cols+1], from, heat_coeff[2][2]);
      }

      np[j] -= outcome;
    }
  }

  temperatures = nt;
}


/// Симуляция теплопереноса + визуализация.
void conduction_demo()
{
  // Два вычислительных шага на один визуализационный.
  draw_temperatures();
  temperatures_step();
  temperatures_step();
}


///////////////////////////////////////////////////////////////////////////////
// Benchmark (uses Plasma)

/// Тест быстродействия для одного варианта функции plasma, возвращает кадры/с.
template <class SineImpl, class DistImpl>
double benchmark_one()
{
  // warming up
  const auto warming_up_frames = 100;
  for (auto i = warming_up_frames; i; --i)
    plasma<SineImpl, DistImpl>();
  
  // check time
  const double time_threshold = 5.0; // 5 seconds
  const auto one_run_rounds = 100;
  std::size_t rounds = 0;
  double t0 = avo::seconds(), t1;
  do for (auto i = one_run_rounds; i; --i, ++rounds)
    plasma<SineImpl, DistImpl>();
  while ((t1 = avo::seconds()) - t0 < time_threshold);

  return rounds / (t1 - t0);
}


/// Тест быстродействия -- геометрическое среднее мегапиксели/с для трёх вариантов 1, 2, 3.
double benchmark()
{
  return std::pow(
    benchmark_one<Sine, Sqrt_dist>() * 
    benchmark_one<Table_sine, Sqrt_dist>() *
    benchmark_one<Table_sine, Crude_dist>(),
    1./3.) * (1.0e-6 * width * height);
}


/// Вывести в консоль количество потоков в рабочей группе OpenMP.
void report_current_threads()
{
  std::cout << "\nThreads: " << omp_get_max_threads() << std::endl;
}


/// Обработать нажатие "обычной" клавиши (x, y -- координаты мыши в окне на момент нажатия).
void on_key(unsigned char key, int x, int y)
{
  switch(key) // код клавиши
  {
  case '1':
    std::cout << "\ndirect implementation\n";
    glutIdleFunc(plasma<Sine, Sqrt_dist>);
    break;

  case '2':
    std::cout << "\ntable implementation\n";
    glutIdleFunc(plasma<Table_sine, Sqrt_dist>);
    break;

  case '3':
    std::cout << "\ntable/crude dist implementation\n";
    glutIdleFunc(plasma<Table_sine, Crude_dist>);
    break;

  case '4':
    std::cout << "\nbilinear test\n";
    temperatures = bilinear::make_random_matrix<float>(20, 20);
    glutIdleFunc(draw_temperatures);
    break;

  case '5':
    std::cout << "\nconduction demo\n";
    temperatures = bilinear::make_random_matrix<float>(64, 64);
    glutIdleFunc(conduction_demo);
    break;

  case 'p': case 'P':
    show_palette = !show_palette;
    break;

  case 'b': case 'B':
    (std::cout << "\nBenchmark").flush();
    std::cout << " results: " << benchmark() << " Mpixels/s\n";
    break;

  case '[': // убрать поток из рабочей группы OpenMP
    {
      const int threads = omp_get_max_threads();
      if (threads > 1)
        omp_set_num_threads(threads - 1);
      report_current_threads();
    }
    break;

  case ']': // добавить поток в рабочую группу OpenMP
    {
      const int threads = omp_get_max_threads();
      if (threads < omp_get_num_procs())
        omp_set_num_threads(threads + 1);
      report_current_threads();
    }
    break;

  case 27: case 'q': case 'Q':
    glutLeaveMainLoop(); // покинуть цикл обработки сообщений
    break;

  default:
    return;
  }
}


//
// Точка входа.
//

int main(int argc, char **argv)
{
  // базовая инициализация
  glutInit(&argc, argv);  // обработка параметров командной строки

  // создать окно с привязанным контекстом OpenGL
  glutInitDisplayMode(GLUT_RGBA | GLUT_SINGLE); // режим: 32-битный цвет RGBA, одинарный буфер кадра SINGLE
  glutInitWindowPosition(0, 0);      // координаты верхнего левого угла окна на экране
  glutInitWindowSize(width, height); // полные ширина и высота окна
  glutCreateWindow("Plasma");        // создать окно, в функцию передаётся заголовок создаваемого окна

  // привязка функций-обработчиков событий
  glutIdleFunc(plasma<Table_sine, Sqrt_dist>);
  glutDisplayFunc(repaint);     // вызывать для отрисовки кадра
  glutKeyboardFunc(on_key); // вызывать для обработки нажатий "обычных" клавиш
  glutReshapeFunc(reshape);     // изменение размеров окна
  
  // старт
  pixels = new Pixel[width * height];
  frames_reported_last = avo::seconds();
  std::cout << "Starting Plasma\n";
  report_current_threads();
  glutMainLoop();
  std::cout << "\nbye!";
  return 0;
}
