﻿/*
 * This file is part of AV Orchid.
 * Copyright (c) 2013, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * Пример: Snow2D
 * Автор: Кувшинов Д.Р. https://evetro.wordpress.com/
 * Используемые библиотеки: FreeGlut, OpenGL.
 * См. также GLPrimitivesVA.
 *
 * Рисует падающий цветной "снег".
 * Движение всех "снежинок" вычисляется одинаково, всё различие между ними состоит лишь
 * в начальной точке траектории и времени начала падения (ну и цвет разный).
 * Движение состоит из двух фаз: "раскачивающееся" и "прямолинейное".
 * Фактически движение частицы (снежинки) управляется конечным автоматом, который может находиться
 * в трёх состояниях: 
 *   1) раскачивание влево (LeftWave), 
 *   2) раскачивание вправо (RightWave) и
 *   3) прямолинейное падение (Down).
 * Переход на прямолинейное падение происходит при достижении нижней (1/10) части окна.
 * Раскачивание представляет собой движение по ветви параболы.
 * Состояние частицы включает в себя также "стартовые" (для текущего состояния) время и позицию.
 *
 * Нажатие ENTER/SPACE генерирует новый набор псевдослучайных снежинок.
 * Нажатие ESC/Q -- выход из программы.
 */
#include <GL/freeglut.h>
#include <iostream>

#include <avo/Basics.hpp> // random
#include <avo/Timer.hpp>
#include <avo/Pixel.hpp>  // 32-битный RGBA-цвет
#include <avo/Vec2.hpp>   // 2D-вектор

using avo::Vec_2f;
using avo::Pixel;
using avo::random;

//
// Вспомогательные функции.
//

/// возведение в квадрат (иногда удобно иметь такую функцию, да)
inline float square(float x)
{
  return x * x;
}

/// сгенерировать некоторый случайный "яркий" цвет
inline Pixel random_bright_color()
{
  const avo::Pixel_component comp0 = random(0, 255),
                            comp1 = random(0, 255);
  switch (std::rand() % 3)
  {
  case 0: return Pixel::rgb(255, comp0, comp1);
  case 1: return Pixel::rgb(comp0, 255, comp1);
  case 2: return Pixel::rgb(comp0, comp1, 255);
  default: return Pixel::grey(255); // чтобы компилятор не писал warning'и
  }
}


//
// Реализация "снежинок".
//

/// данные, модель и логика снежинки
struct Snowflake
{
  Pixel color;     ///< цвет
  Vec_2f position; ///< координаты

  double t0; ///< начальное время для текущего режима
  Vec_2f p0; ///< начальная позиция для текущего режима
  enum Mode { LeftWave, RightWave, Down } mode; ///< текущее состояние (режим движения)

  /// сгенерировать случайные данные для снежинки
  /// начальное время выбирается из отрезка [t0min, t0max])
  void make_random(double t0min, double t0max)
  {
    color = random_bright_color();
    t0 = random(t0min, t0max);
    p0 = Vec_2f(random(0.0f, 1.0f),  // положение по горизонтали покрывает всю ширину окна
                random(1.1f, 1.4f)); // положение по вертикали всегда "над" окном в невидимой области (до 40% высоты окна)
    mode = std::rand() & 1? LeftWave: RightWave; // с равной вероятностью начинать раскачиваться влево или вправо
  }

  /// вычислить положение и состояние снежинки для следующего момента времени t
  void move(double t)
  {
    const float dt = (float)(t - t0);
    if (dt <= 0.f) // текущее время в прошлом относительно t0 -- стоим на месте
    {
      position = p0;
      return;
    }

    // двигаться в соответствии с текущим режимом
    switch (mode)
    {
    case LeftWave: // по параболе влево
      position = p0 - dt * Vec_2f(
                                  1.f / 60.f,                               // x
                                  (1.f / 100.f) * (9.f - square(dt - 3.0f)) // y
                                 );

      if (dt >= 4.0f) // не более 4 секунд
      { // переключиться на движение вправо
        mode = RightWave;
        t0 = t;
        p0 = position;
      }
      break;

    case RightWave: // по параболе вправо
      position = p0 - dt * Vec_2f(
                                  -1.f / 60.f,                              // x
                                  (1.f / 100.f) * (9.f - square(dt - 3.0f)) // y
                                 );

      if (dt >= 4.0f) // не более 4 секунд
      { // переключиться на движение влево
        mode = LeftWave;
        t0 = t;
        p0 = position;
      }
      break;

    case Down: // гиперболически замедляющееся падение по вертикали
      position = Vec_2f(
                        p0.x,
                        p0.y / (0.5f * dt + 1.0f)
                       );
      break;
    }

    // переключиться на падение, если достигли нижней 1/10 окна
    if (position.y <= 0.1f)
    {
      mode = Down;
      t0 = t;
      p0 = position;
    }
      
    // удаляем все снежинки, которые достигли нижней 1/1000 окна
    if (position.y <= 1e-3f)
      make_random(t, t + 2.0);

    // TASK: добавить переключение цвета при переключении режима, например,
    // LeftWave -- зелёный,
    // RightWave -- красный,
    // Down -- белый.

    // TASK: добавить отталкивание/притягивание снежинок от/к указателю мыши.
    // Получение координат мыши см., например, в примере BasicFrame или Collision2D.
  }
};


//
// Глобальные константы и переменные.
//

// TASK: добавить возможность изменения числа активных снежинок на экране во время работы программы,
//       выводить в консоль количество снежинок при его изменении.

/// общее количество активных снежинок
const int SNOWFLAKES = 20000;

/// массив снежинок
Snowflake snowflakes[SNOWFLAKES];


/// сгенерировать случайный набор снежинок
void make_random(double t0min, double t0max)
{
  for (int i = 0; i < SNOWFLAKES; ++i)
    snowflakes[i].make_random(t0min, t0max);
}


//
// Обработчики событий FreeGlut.
//

/// перерисовка сцены
void repaint()
{
  // очистить окно, закрасив его цветом очистки
  glClear(GL_COLOR_BUFFER_BIT);
  // подключить массивы:
  // количество компонент, тип компоненты, расстояние между соседними элементами, указатель на первый элемент
  glColorPointer  (4, GL_UNSIGNED_BYTE, sizeof(Snowflake), &snowflakes[0].color);
  glVertexPointer (2, GL_FLOAT, sizeof(Snowflake), &snowflakes[0].position);

  // передать примитив: тип, с какого начинать, количество
  glDrawArrays(GL_POINTS, 0, SNOWFLAKES);

  glFinish(); // дождаться выполнения всех OpenGL команд
  glutSwapBuffers(); // обменять буферы кадра, чтобы вывести в окно новый кадр (двойная буферизация)
}

/// расчёт нового положения снежинок
void frame_move()
{
  // "спать" 8 миллисекунд, ограничивая таким образом перерисовку 1000/8 = 125 кадрами в секунду
  avo::sleep(0.008);  // -- на деле это может быть далеко не 8мс, а как захочет ОС
  const auto t = avo::seconds();
  for (int i = 0; i < SNOWFLAKES; ++i)
    snowflakes[i].move(t); // рассчитать следующие позиции
  glutPostRedisplay(); // попросим FreeGlut выполнить перерисовку
}

/// обработать нажатие "обычной" клавиши (x, y -- координаты мыши в окне на момент нажатия)
void on_key(unsigned char key, int x, int y)
{
  switch(key)
  {
  case 27: case 'q': case 'Q':
    glutLeaveMainLoop(); // покинуть цикл обработки сообщений
    break;

  case ' ': case '\r': case '\n':
    make_random(avo::seconds(), avo::seconds() + 60.0);
    glutPostRedisplay();
    break;

  default:
    return; // остальные коды игнорируются
  }
}


//
// main
//

int main(int argc, char **argv)
{
  // базовая инициализация
  glutInit(&argc, argv); // обработка параметров командной строки
  std::srand((unsigned)(avo::seconds() * 1e+6) % 15485863);

  // создать окно с привязанным контекстом OpenGL
  // режим: 32-битный цвет RGBA, двойная буферизация DOUBLE
  glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE);
  glutInitWindowPosition(0, 0); // координаты верхнего левого угла окна на экране
  glutInitWindowSize(800, 600); // полные ширина и высота окна
  glutCreateWindow("Snow 2D");  // создать окно, в функцию передаётся заголовок создаваемого окна

  // параметры OpenGL для данного примера
  glEnable(GL_COLOR_MATERIAL); // выбирать цвет материала командой glColor (через Color::select)
  glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE); // этот цвет -- с обеих сторон полигона, ambient и diffuse компоненты
  glClearColor(0.0f, 0.0f, 0.0f, 0.0f); // установить цвет очистки (фон) -- чёрный
  glPointSize(5.f); // установить диаметр точки ("снежинки") -- 5 пикселей
  glEnable(GL_POINT_SMOOTH); // включить "сглаживание точек"
    // -- без этой команды на некоторых устройствах/версиях драйверов точки будут отображаться не кружками, а квадратами

  // включить рисование примитивов, заданных массивами
  glEnableClientState(GL_COLOR_ARRAY);  // включить массив цветов вершин
  glEnableClientState(GL_VERTEX_ARRAY); // включить массив координат вершин

  // 2D "камера"
  glMatrixMode(GL_PROJECTION); // режим матрицы проекции
  glLoadIdentity();            // сбросить в единичную
  glOrtho(0.0, 1.0, 0.0, 1.0, -1.0, 1.0); // домножить на матрицу изометрической проекции,
  // координаты в окне: (0, 0) -- нижний левый угол, (1, 1) -- верхний правый угол
  glMatrixMode(GL_MODELVIEW);  // режим матрицы вида
  
  // привязка функций-обработчиков событий
  glutIdleFunc(frame_move); // вызывать, когда нет событий
  glutDisplayFunc(repaint); // вызывать для отрисовки кадра
  glutKeyboardFunc(on_key); // вызывать для обработки нажатий "обычных" клавиш

  // старт
  std::cout << "Starting Snow2D\n";
  make_random(avo::seconds(), avo::seconds() + 60.0);
  glutMainLoop(); // вход в цикл обработки событий
  std::cout << "\nbye!";
  return 0;
}
