/*
 * This file is part of AV Orchid.
 * Copyright (c) 2013, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * : SortVis
 * :  .. https://evetro.wordpress.com/
 *    .
 *        
 * (  std::distance, std::advance  ..).
 * .  http://www.sorting-algorithms.com/ , http://home.westman.wave.ca/~rhenry/sort/
 */
#pragma once
#include <cassert>
#include <cstddef>
#include <climits>
#include <iterator>
#include <utility>
#include <memory>
#include <algorithm>


//      .
//  .

///  -   ""  .
///    bucket_sort.
template <std::size_t BucketBytes, class RanIt, class Val, class T>
void bucket_sort_inner(RanIt from, RanIt to, Val val, T * buf)
{
  using I = typename std::iterator_traits<RanIt>::difference_type;
  using T = typename std::iterator_traits<RanIt>::value_type;
  using F = decltype(val(std::declval<T>()));
  const I bucket_size = BucketBytes / sizeof(T);
  
  using std::distance;
  using std::next;

  const I sz = distance(from, to);
  if (sz <= 3 * bucket_size)
    return std::sort(from, to);

  // Minimal and maximal values in [from, to).
  F minv, maxv;
  {
    const auto pp = std::minmax_element(from, to);
    minv = val(*pp.first);
    maxv = val(*pp.second);
    if (minv == maxv)
      return;
  }

  // Prepare buckets, offsets and positions.
  const I buckets = 1 + sz / bucket_size;
  const F factor = F(buckets - 1) / (maxv - minv);

  // Is it possible to evade memory allocation here?
  const auto index_store = std::make_unique<I[]>(buckets);
  const auto offset = index_store.get();
  std::fill_n(offset, buckets, I(0));

  // Compute bucket sizes.
  for (RanIt p = from; p != to; ++p)
    offset[I(factor * (val(*p) - minv))]++;

  // Compute offsets and positions.
  for (I i = 0, s = 0; i < buckets; ++i)
  {
    const I t = offset[i] + s;
    offset[i] = s;
    s = t;
  }

  // Scatter values into buckets.
  for (RanIt p = from; p != to; ++p)
    buf[offset[I(factor * (val(*p) - minv))]++] = std::move(*p);

  // Copy buckets back into the source array.
  std::move(buf, buf + sz, from);

  // Sort buckets.
  for (I i = 0, prev = 0; i < buckets; ++i)
  {
    bucket_sort_inner<BucketBytes>(
      next(from, prev),
      next(from, offset[i]),
      val, buf);

    prev = offset[i];
  }
}


///  .
///     "" (  ),   val.
///       ,  ""  " ".
///      .
///     std::sort.
template <std::size_t BucketBytes, class RanIt, class Val>
void bucket_sort(RanIt from, RanIt to, Val val)
{   
  using I = typename std::iterator_traits<RanIt>::difference_type;
  using T = typename std::iterator_traits<RanIt>::value_type;
  using F = decltype(val(std::declval<T>()));
  const I bucket_size = BucketBytes / sizeof(T);

  using std::distance;
  using std::next;

  const I sz = distance(from, to);
  if (sz <= bucket_size)
    return std::sort(from, to);

  // Minimal and maximal values in [from, to).
  F minv, maxv;
  {
    const auto pp = std::minmax_element(from, to);
    minv = val(*pp.first);
    maxv = val(*pp.second);
    if (minv == maxv)
      return;
  }

  // Prepare buckets, offsets and positions.
  const I buckets = 1 + (sz + bucket_size - 1) / bucket_size;
  const F factor = F(buckets - 1) / (maxv - minv);

  const auto buf_store = std::make_unique<T[]>(sz);
  const auto buf = buf_store.get();

  const auto index_store = std::make_unique<I[]>(buckets);
  const auto offset = index_store.get();
  std::fill_n(offset, buckets, I(0));

  // Compute bucket sizes.
  for (RanIt p = from; p != to; ++p)
    offset[I(factor * (val(*p) - minv))]++;

  // Compute positions.
  for (I i = 0, s = 0; i < buckets; ++i)
  {
    const auto t = offset[i] + s;
    offset[i] = s;
    s = t;
  }

  // Scatter values into buckets.
  for (RanIt p = from; p != to; ++p)
    buf[offset[I(factor * (val(*p) - minv))]++] = std::move(*p);

  // Copy buckets back into the source array.
  std::move(buf, buf + sz, from);

  // Sort buckets.
  for (I i = 0, prev = 0; i < buckets; ++i)
  {
    bucket_sort_inner<BucketBytes>(
      next(from, prev),
      next(from, offset[i]),
      val, buf);

    prev = offset[i];
  }
}



//  .

///      .
///     "" ( ),   val.
template <unsigned DigitBits, class RanIt, class Val>
void lsd_radix_sort(RanIt from, RanIt to, Val val)
{
  using I = typename std::iterator_traits<RanIt>::difference_type;
  using T = typename std::iterator_traits<RanIt>::value_type;
  using B = decltype(val(std::declval<T>()));
  using std::distance;

  static_assert(DigitBits <= sizeof(B) * CHAR_BIT, "Too big DigitBits value.");

  static const B buckets = B(1) << DigitBits, bit_mask = buckets - 1;
  const I sz = distance(from, to);  

  const auto bufstore = std::make_unique<T[]>(sz);
  const auto buf = bufstore.get();
  I offset[buckets];

  for (unsigned shift = 0; shift < sizeof(B) * CHAR_BIT; shift += DigitBits)
  {
    std::fill_n(offset, buckets, I(0));

    // Compute bucket sizes.
    for (RanIt p = from; p != to; ++p)
      offset[(val(*p) >> shift) & bit_mask]++;

    if (std::find(std::begin(offset), std::end(offset), sz) == std::end(offset))
    {
      // Compute positions.
      for (I i = 0, s = 0; i < buckets; ++i)
      {
        const I t = offset[i] + s;
        offset[i] = s;
        s = t;
      }

      // Scatter values into buckets.
      for (RanIt p = from; p != to; ++p)
        buf[offset[(val(*p) >> shift) & bit_mask]++] = std::move(*p);

      // Copy buckets back into the source array.
      std::move(buf, buf + sz, from);
    }
  }
}



///      .
///     "" ( ),   val.
template <unsigned DigitBits, int MinRecSize, class RanIt, class Val, class T>
void msd_radix_sort_inner(RanIt from, RanIt to, Val val, unsigned shift, T * buf)
{
  using I = typename std::iterator_traits<RanIt>::difference_type;
  using B = decltype(val(std::declval<T>()));
  using std::distance;
  using std::next;

  static const B buckets = B(1) << DigitBits, bit_mask = buckets - 1;
  const I sz = distance(from, to);  

  if (sz < MinRecSize)
    return std::sort(from, to);

  I offset[buckets]{};

  // Compute bucket sizes.
  for (RanIt p = from; p != to; ++p)
    offset[(val(*p) >> shift) & bit_mask]++;

  if (std::find(std::begin(offset), std::end(offset), sz) == std::end(offset))
  {
    // Compute positions.
    for (I i = 0, s = 0; i < buckets; ++i)
    {
      const auto t = offset[i] + s;
      offset[i] = s;
      s = t;
    }

    // Scatter values into buckets.
    for (RanIt p = from; p != to; ++p)
      buf[offset[(val(*p) >> shift) & bit_mask]++] = std::move(*p);

    // Copy buckets back into the source array.
    std::move(buf, buf + sz, from);
  }

  if (shift == 0)
    return;
  if (shift <= DigitBits)
    shift = 0;
  else
    shift -= DigitBits;

  // Sort buckets.
  for (I i = 0, prev = 0; i < buckets; ++i)
  {
    msd_radix_sort_inner<DigitBits, MinRecSize>(
      next(from, prev),
      next(from, offset[i]),
      val, shift, buf);

    prev = offset[i];
  }
}


///      .
///     "" ( ),   val.
template <unsigned DigitBits, int MinRecSize, class RanIt, class Val>
void msd_radix_sort(RanIt from, RanIt to, Val val)
{
  using I = typename std::iterator_traits<RanIt>::difference_type;
  using T = typename std::iterator_traits<RanIt>::value_type;
  using B = decltype(val(std::declval<T>()));
  using std::distance;
  using std::next;

  static_assert(DigitBits <= sizeof(B) * CHAR_BIT, "Too big DigitBits value.");
  static_assert(MinRecSize > 0, "MinRecSize must be 1 or greater.");

  static const B buckets = B(1) << DigitBits, bit_mask = buckets - 1;
  const I sz = distance(from, to);  

  if (sz < MinRecSize)
    return std::sort(from, to);

  const auto bufstore = std::make_unique<T[]>(sz);
  const auto buf = bufstore.get();
  unsigned shift = sizeof(B) * CHAR_BIT - DigitBits;
  I offset[buckets]{};

  // Compute bucket sizes.
  for (RanIt p = from; p != to; ++p)
    offset[(val(*p) >> shift) & bit_mask]++;

  if (std::find(std::begin(offset), std::end(offset), sz) == std::end(offset))
  {
    // Compute positions.
    for (I i = 0, s = 0; i < buckets; ++i)
    {
      const auto t = offset[i] + s;
      offset[i] = s;
      s = t;
    }

    // Scatter values into buckets.
    for (RanIt p = from; p != to; ++p)
      buf[offset[(val(*p) >> shift) & bit_mask]++] = std::move(*p);

    // Copy buckets back into the source array.
    std::move(buf, buf + sz, from);
  }

  if (shift == 0)
    return;
  if (shift <= DigitBits)
    shift = 0;
  else
    shift -= DigitBits;

  // Sort buckets.
  for (I i = 0, prev = 0; i < buckets; ++i)
  {
    msd_radix_sort_inner<DigitBits, MinRecSize>(
      next(from, prev),
      next(from, offset[i]),
      val, shift, buf);

    prev = offset[i];
  }
}
