﻿/*
 * This file is part of AV Orchid.
 * Copyright (c) 2013, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * Пример: SortVis
 * Автор: Кувшинов Д.Р. https://evetro.wordpress.com/
 * Реализации демонстрируемых алгоритмов сортировки.
 * Некоторые алгоритмы явно требуют итераторы случайного доступа 
 * (не используют std::distance, std::advance и т.п.).
 * См. также http://www.sorting-algorithms.com/ , http://home.westman.wave.ca/~rhenry/sort/
 */
#pragma once
#include <cassert>
#include <vector>
#include <stack>       // в tree_sort для замены рекурсии / сокращения кода
#include <set>         // std-based O(NlogN) tree sort
#include <algorithm>   // std::sort + некоторые вспомогательные элементы
#include <iterator>    // std::iterator_traits, std::next, std::prev
#include <tuple>       // std::tie
#include <cstdlib>     // std::rand в random_try_sort
#include <climits>     // CHAR_BIT
#include <type_traits> // свойства типов
#include <memory>      // std::unique_ptr
#include <functional>
#include <random>
#include <cmath>


/// Вспомогательная процедура для улучшения внешнего вида кода, реализует компаратор сравнивающей сети.
template <class It> // если *q < *p, то обменять *p, *q
inline void ascend(const It p, const It q)
{
  using std::iter_swap;
  if (*q < *p)
    iter_swap(p, q);
}


//
// квадратичные по сложности (в среднем) алгоритмы

/// одна итерация алгоритма сортировки пузырьком, используется в bubble_sort и cocktail_sort
/** возвращает истину, если были перестановки элементов, ложь, если последовательность уже отсортирована */
template <class It>
bool bubble_round(It begin, It end, bool ascending = true)
{
  using std::iter_swap;
  bool result = false;
  // q всегда следует за p 
  for (It q = begin, p = q++; q != end; p = q++)
  {
    // форсируем выполнение неравенства *p <= *q, если ascending,
    // либо *q <= *p, если не ascending
    if ((ascending && *q < *p) || (!ascending && *p < *q))
    {
      result = true;          
      iter_swap(p, q);
    }
  }
  return result;
}

/// простейший и широко известный алгоритм сортировки пузырьком
template <class It>
void bubble_sort(It begin, It end)
{
  // повторять, пока происходят перестановки элементов
  while (bubble_round(begin, end))
    --end;
}

/// слегка улучшенный вариант пузырьковой сортировки: "коктейль" (также называют "челночной" или "шейкерной" сортировкой)
template <class It>
void cocktail_sort(It begin, It end)
{
  // воспользуемся стандартным reverse_iterator,
  // чтобы свести задачу к использованию bubble_round (то в одну сторону, то в другую)
  std::reverse_iterator<It> rbegin(end), rend(begin);
  while (bubble_round(begin, end, true) && bubble_round(++rbegin, rend, false))
    ++begin, --end, --rend;
}

///////////////////////////////////////////////////////////////////////////////

/// Сортировка вставками.
template <class It>
void insertion_sort(It begin, It end)
{
  if (begin == end) return;
  using std::prev;
  using std::next;
  for (It p = next(begin); p != end; ++p)
  {
    // Чтобы избежать лишнего копирования и сравнения
    // для уже упорядоченных подпоследовательностей
    // (на практике это может замедлить сортировку маленьких объектов (int и т.п.)).
    It q = prev(p);
    if (*p < *q)
    {
      const auto stone = *p; // "камень"
      *p = *q;
      while (begin != q)
      {
        It r = prev(q);
        if (!(stone < *r)) break;
        *q = *r;
        q = r;
      }
      *q = stone;
    }
  }
}

///////////////////////////////////////////////////////////////////////////////

/// Сортировка выбором минимума и максимума.
/**
 * На каждом шаге выбирается минимум и максимум в диапазоне.
 * Минимум обменивается с началом, максимум -- с концом, диапазон стягивается на две позиции.
 * Повторяется, пока в диапазоне есть, по крайней мере, два элемента.
 */
template <class It>
void minmax_sort(It begin, It end)
{
  using std::iter_swap;
  while (begin != end)
  {
    // Найти положение минимума и максимума.
    /*
    It minp = begin, maxp = begin;
    for (It p = begin; ++p != end; )
    {
      if (*p < *minp)
        minp = p;
      else if (*maxp < *p)
        maxp = p;
    }
    */
    // Примерно то же самое на C++11:
    It minp, maxp;
    std::tie(minp, maxp) = std::minmax_element(begin, end);

    if (minp == maxp)
      return;

    --end; // end далее указывает на последний элемент диапазона
    // Обменять элементы.
    if (maxp != begin)
    {
      if (minp != begin) iter_swap(minp, begin);
      if (maxp != end)   iter_swap(maxp, end);
    }
    else
    {
      if (maxp != end)   iter_swap(maxp, end);
      if (minp != end)   iter_swap(minp, begin); // если minp был end, то его уже обменяли с begin == maxp
    }

    if (begin != end)
      ++begin;
  }
}

///////////////////////////////////////////////////////////////////////////////

/// Проверка на эквивалентность.
template <class T>
inline bool equivalent(const T &a, const T &b)
{
  return !((a < b) || (b < a));
}

/// Найти первый, не эквивалентный заданному элементу.
template <class It, class T>
It first_nonequivalent(It begin, It end, const T &element)
{
  return std::find_if_not (begin, end, 
    std::bind(equivalent<T>, element, std::placeholders::_1));
  /*while (begin != end && equivalent(element, *begin))
    ++begin;
  return begin;*/
}

/// Передвинуть от начала на количество элементов, меньших заданного.
template <class It, class T>
It advance_by_lesser_amount(It begin, It end, const T &element)
{
  for (It q = begin; ++q != end; )
    if (*q < element) ++begin;
  return begin;
}

/// Сортировка разворачиванием циклов, минимизирует количество записей.
template <class It>
void cycle_sort(It begin, It end)
{
  using std::swap;
  for (; begin != end; ++begin)
  {
    auto current = *begin;

    It p = advance_by_lesser_amount(begin, end, current);
    if (p != begin 
    && (p = first_nonequivalent(p, end, current)) != end)
    {
      swap(*p, current);
      while (p != begin)
      {
        p = first_nonequivalent
            (advance_by_lesser_amount(begin, end, current), 
             end, current);
        if (p == end) break;
        swap(*p, current);
      }
    }
  }
  
  // TASK: можно ли как-то ускорить (в среднем) cycle_sort,
  // не увеличивая количество выполняемых записей. (Вообще, возможно ли это?)
}


// TASK: реализовать квадратичную чётно-нечётную сортировку 
// (название напоминает чётно-нечётную сортировку Бэтчера, но это другой алгоритм).
// См. http://en.wikipedia.org/wiki/Odd%E2%80%93even_sort

///////////////////////////////////////////////////////////////////////////////

//
// Линейно-логарифмические по сложности (в среднем) алгоритмы.

/// Значительно улучшенная пузырьковая сортировка с переменным зазором: "сортировка расчёской".
/// Формально квадратичная в худшем и даже среднем случае, однако ведёт себя подобно линейно-логарифмическим.
template <class It>
void comb_sort(It begin, It end)
{
  using std::distance;
  using std::next;
  using std::iter_swap;

  using Diff = typename std::iterator_traits<It>::difference_type;
  const Diff length = distance(begin, end);
  Diff gap = length;
  double dgap = gap;
  bool repeat = true;

  while (gap > 1 || repeat)
  {
    if (gap > 1)
    {
      // exp(phi)/(exp(phi) - 1), где phi -- "золотое сечение", даёт O(N log N) скорость алгоритма
      dgap /= 1.247330950103979;
      gap = static_cast<Diff>(dgap);
    }
    // Алгоритм можно ускорить, если переключаться на сортировку вставками
    // при достижении малых значений gap.

    repeat = false;
    It p = begin, q = next(begin, gap);
    for (Diff i = 0; i < length - gap; ++i, ++p, ++q)
    {
      if (*q < *p)
      {
        repeat = true;
        iter_swap(p, q);
      }
    }
  }
}


/// Гибрид comb_sort + insertion_sort.
template <class It>
void ci_sort(It begin, It end)
{
  using std::distance;
  using std::next;

  using Diff = typename std::iterator_traits<It>::difference_type;
  const Diff length = distance(begin, end);
  Diff gap = length;
  double dgap = gap;
  while (true)
  {
    // exp(phi)/(exp(phi) - 1), где phi -- "золотое сечение"
    dgap /= 1.247330950103979;
    if (dgap < 6.0) break; // выйти из цикла, если ожидается gap < 6
    gap = static_cast<Diff>(dgap);
    
    It p = begin, q = next(begin, gap);
    for (Diff i = 0; i < length - gap; ++i, ++p, ++q)
      ascend(p, q);
  }

  // Добить сортировкой вставками.
  insertion_sort(begin, end);
}


/// Челночная версия comb_sort + insertion_sort с настраиваемым делителем.
template <class It>
void cis_sort_impl(It begin, It end, double divisor, double min_gap)
{
  using std::distance;
  using std::next;
  using std::prev;

  using Diff = typename std::iterator_traits<It>::difference_type;
  const Diff length = distance(begin, end);
  Diff gap = length;
  double dgap = gap;
  while (true)
  {
    dgap /= divisor;
    if (dgap <= min_gap) break;
    gap = static_cast<Diff>(dgap);
    {
      It p = begin, q = next(begin, gap);
      for (Diff i = 0; i < length - gap; ++i, ++p, ++q)
        ascend(p, q);
    }

    dgap /= divisor;
    if (dgap <= min_gap) break;
    gap = static_cast<Diff>(dgap);
    {
      It q = prev(end), p = prev(q, gap);
      for (Diff i = length - gap; i > 0; --i, --p, --q)
        ascend(p, q);
    }
  }

  // Добить сортировкой вставками.
  insertion_sort(begin, end);
}

template <class It>
void cis_sort(It begin, It end)
{
  cis_sort_impl(begin, end, 1.247330950103979, 6.0);
}

template <class It>
void cis14_sort(It begin, It end)
{
  cis_sort_impl(begin, end, 1.4, 6.0);
}

template <class It>
void cis16_sort(It begin, It end)
{
  cis_sort_impl(begin, end, 1.6, 4.5);
}


///////////////////////////////////////////////////////////////////////////////

/// Классическая полностью рекурсивная реализация быстрой сортировки 
/// с копированием среднего элемента в качестве опорного.
template <class It>
void quick_sort(It begin, It end)
{
  using std::distance;
  using std::next;
  using std::iter_swap;
  if (begin == end) return;

  It i = begin, j = end;
  if (i == --j) return;

  // Выбирает опорный элемент из середины,
  // алгоритм будет работать и при других вариантах выбора опорного элемента,
  // но с различной скоростью на различных распределениях входных данных.
  const auto pivot = *next(begin, distance(begin, end) / 2);
  do // Разбиение (partition).
  {
    while (*i < pivot) ++i;
    while (pivot < *j) --j;

    if(i <= j) 
    {
      if (i != j) iter_swap(i, j);
      ++i;
      --j;
    }
  } while (i <= j);
  
  // Наивная рекурсия на две ветки.
  if (i < end)   quick_sort(i,     end);
  if (begin < j) quick_sort(begin, ++j);
}


/// Челночная версия comb_sort + quick_sort с настраиваемым делителем.
template <class It>
void cqs_sort_impl(It begin, It end, double divisor, std::size_t max_repeats)
{
  using std::distance;
  using std::next;
  using std::prev;

  using Diff = typename std::iterator_traits<It>::difference_type;
  const Diff length = distance(begin, end);
  Diff gap = length;
  double dgap = gap;
  while (max_repeats-- != 0)
  {
    dgap /= divisor;
    if (dgap <= 6) break; // выйти из цикла, если ожидается gap < 6
    gap = static_cast<Diff>(dgap);
    {
      It p = begin, q = next(begin, gap);
      for (Diff i = 0; i < length - gap; ++i, ++p, ++q)
        ascend(p, q);
    }

    dgap /= divisor;
    if (dgap <= 6) break; // выйти из цикла, если ожидается gap < 6
    gap = static_cast<Diff>(dgap);
    {
      It q = prev(end), p = prev(q, gap);
      for (Diff i = length - gap; i > 0; --i, --p, --q)
        ascend(p, q);
    }
  }

  // Добить быстрой сортировкой.
  quick_sort(begin, end);
}

template <class It>
void cqs16_sort(It begin, It end)
{
  cqs_sort_impl(begin, end, 1.6,
    static_cast<std::size_t>(0.5 * std::log2(2 + 0.17*std::distance(begin, end))));
}


/// Рекурсивная реализация быстрой сортировки с использованием std::partition
/// и защитой от переполнения стека. Выбирает средний элемент.
template <class It>
void std_quick_sort(It begin, It end)
{
  using std::next;
  using std::distance;
  using std::iter_swap;

  while (true) switch (const auto span = distance(begin, end))
  {
  case 0: case 1: return;
  case 2:
    ascend(begin, next(begin));
    return;

  default:
    {
      const auto pivot = *next(begin, span / 2);
      const auto mid = std::partition(begin, end,
        [&](typename std::iterator_traits<It>::reference item)
        { return item < pivot; });

      if (mid == begin)
      {
        // Разделить не удалось, т.к. средний элемент равен минимуму всех элементов.
        // Обменяем его с начальным элементом,
        iter_swap(begin, next(begin, span / 2));
        // а на следующей итерации цикла начнём сортировать оставшуюся часть.
        ++begin;
      }
      else if (distance(begin, mid) < distance(mid, end))
      {
        // Левая часть меньше правой, отсортируем её рекурсивно.
        std_quick_sort(begin, mid);
        // Правую часть отсортируем в цикле.
        begin = mid;
      }
      else
      {
        // Правая часть не больше левой, отсортируем её рекурсивно.
        std_quick_sort(mid, end);
        // Левую часть отсортируем в цикле.
        end = mid;
      }
    }
  }
}


/// Рекурсивная реализация устойчивой быстрой сортировки с использованием std::stable_partition
/// и защитой от переполнения стека. Выбирает средний элемент.
template <class It>
void stable_quick_sort(It begin, It end)
{
  using std::next;
  using std::distance;
  using std::iter_swap;
  using Ref = typename std::iterator_traits<It>::reference;

  while (true) switch (const auto span = distance(begin, end))
  {
  case 0: case 1: return;
  case 2:
    ascend(begin, next(begin));
    return;

  default:
    {
      const auto pivot = *next(begin, span / 2);
      const auto mid = std::stable_partition(begin, end,
        [&](Ref item) { return item < pivot; });

      if (mid == begin)
      {
        // Средней элемент = минимуму.
        // Отделим все минимумы устойчиво.
        begin = std::stable_partition(begin, end,
          [&](Ref item) { return !(pivot < item); });
      }
      else if (distance(begin, mid) < distance(mid, end))
      {
        // Левая часть меньше правой, отсортируем её рекурсивно.
        stable_quick_sort(begin, mid);
        // Правую часть отсортируем в цикле.
        begin = mid;
      }
      else
      {
        // Правая часть не больше левой, отсортируем её рекурсивно.
        stable_quick_sort(mid, end);
        // Левую часть отсортируем в цикле.
        end = mid;
      }
    }
  }
}


/// Рекурсивная реализация быстрой сортировки с использованием std::partition
/// и защитой от переполнения стека. Выбирает условно случайный элемент.
struct Std_pr_quick_sort
{
  std::size_t seed;

  Std_pr_quick_sort(std::size_t seed)
    : seed(seed) {}

  template <class It>
  void operator()(It begin, It end)
  {
    using std::next;
    using std::distance;
    using std::iter_swap;

    while (true) switch (const auto span = distance(begin, end))
    {
    case 0: case 1: return;
    case 2:
      ascend(begin, next(begin));
      return;

    default:
      {
        // Простейший линейный конгруэнтный генератор.
        seed = (4096 * seed + 150889) % 714025;

        const auto pivot_pos = next(begin, seed % span);
        const auto pivot = *pivot_pos;
        const auto mid = std::partition(begin, end,
          [=](typename std::iterator_traits<It>::reference item)
          { return item < pivot; });

        if (mid == begin)
        {
          // Разделить не удалось, т.к. средний элемент равен минимуму всех элементов.
          // Обменяем его с начальным элементом,
          iter_swap(begin, pivot_pos);
          // а на следующей итерации цикла начнём сортировать оставшуюся часть.
          ++begin;
        }
        else if (distance(begin, mid) < distance(mid, end))
        {
          // Левая часть меньше правой, отсортируем её рекурсивно.
          (*this)(begin, mid);
          // Правую часть отсортируем в цикле.
          begin = mid;
        }
        else
        {
          // Правая часть не больше левой, отсортируем её рекурсивно.
          (*this)(mid, end);
          // Левую часть отсортируем в цикле.
          end = mid;
        }
      }
    }
  }
};

template <class It>
void std_pr_quick_sort(It begin, It end)
{
  Std_pr_quick_sort(std::distance(begin, end))(begin, end);
}


// TASK: реализовать 3-way partitioning quicksort
// см. http://www.sorting-algorithms.com/static/QuicksortIsOptimal.pdf

// TASK*: реализовать quicksort с выборкой медианы в качестве опорного элемента
// см. http://en.wikipedia.org/wiki/Selection_algorithm#Linear_general_selection_algorithm_-_Median_of_Medians_algorithm

///////////////////////////////////////////////////////////////////////////////

/// сортировка слияниями, использующая заранее подготовленный внешний буфер полного размера
template <class It, class BufIt> // bb -- buffer begin, be -- buffer end
void merge_sort_impl(It begin, It end, BufIt bb, BufIt be)
{
  using std::distance;
  using std::next;

  const auto gap = distance(begin, end) / 2;
  if (gap < 1) return;
  // средний элемент
  It   mid = next(begin, gap);
  BufIt bm = next(bb, gap);

  merge_sort_impl(begin, mid, bb, bm);
  merge_sort_impl(mid,   end, bm, be);

  // ещё не отсортировано?
  using std::prev;
  if (*mid < *prev(mid))
  {
    // слияние двух упорядоченных последовательностей (половинок), запись в buf
    std::merge(begin, mid, mid, end, bb);
    // замена исходной последовательности содержимым буфера
    std::copy(bb, be, begin);
  }
}

/// Сортировка слияниями.
template <class It>
void merge_sort(It begin, It end)
{
  using std::distance;
  // промежуточный буфер
  using Value = typename std::iterator_traits <It>::value_type;
  std::vector<Value> buf(distance(begin, end));
  // собственно сортировка
  merge_sort_impl(begin, end, buf.begin(), buf.end());
}


/// Сортировка слияниями "на месте".
template <class It>
void inplace_merge_sort(It begin, It end)
{
  using std::distance;
  using std::next;
  using std::prev;
  using std::iter_swap;

  switch (const auto span = distance(begin, end))
  {
  case 0: case 1: return;
  case 2:
    ascend(begin, next(begin));
    return;

  default:
    {
      const auto mid = next(begin, span / 2);
      inplace_merge_sort(begin, mid);
      inplace_merge_sort(mid, end);

      // ещё не отсортировано?
      if (*mid < *prev(mid))
        std::inplace_merge(begin, mid, end);
    }
  }
}


/// Челночная версия comb_sort + imerge_sort с настраиваемым делителем.
template <class It>
void cims_sort_impl(It begin, It end, double divisor, std::size_t max_repeats)
{
  using std::distance;
  using std::next;
  using std::prev;

  using Diff = typename std::iterator_traits<It>::difference_type;
  const Diff length = distance(begin, end);
  Diff gap = length;
  double dgap = gap;
  while (max_repeats-- != 0)
  {
    dgap /= divisor;
    if (dgap <= 6) break; // выйти из цикла, если ожидается gap < 6
    gap = static_cast<Diff>(dgap);
    {
      It p = begin, q = next(begin, gap);
      for (Diff i = 0; i < length - gap; ++i, ++p, ++q)
        ascend(p, q);
    }

    dgap /= divisor;
    if (dgap <= 6) break; // выйти из цикла, если ожидается gap < 6
    gap = static_cast<Diff>(dgap);
    {
      It q = prev(end), p = prev(q, gap);
      for (Diff i = length - gap; i > 0; --i, --p, --q)
        ascend(p, q);
    }
  }

  // Добить сортировкой слиянием.
  inplace_merge_sort(begin, end);
}

template <class It>
void cims16_sort(It begin, It end)
{
  cims_sort_impl(begin, end, 1.6,
    static_cast<std::size_t>(0.5 * std::log2(2 + 0.5*std::distance(begin, end))));
}


/// Отсортировать 3 элемента (устойчиво).
template <class It>
inline void ascend3(const It a, const It b, const It c)
{
  using std::iter_swap;
  if (*b < *a) // .b.a.
  {
    if (*c < *b) // abc -> cba
      iter_swap(a, c);
    else // b[ac]
      iter_swap(a, b), // abc -> bac
      ascend(b, c);    // sort ac
  }
  else // .a.b.
  {
    if (*c < *a) // abc -> bac -> cab
      iter_swap(a, b),
      iter_swap(a, c);
    else if (*c < *b) // abc -> acb
      iter_swap(b, c);
    // abc already sorted otherwise
  }

  assert(!(*b < *a));
  assert(!(*c < *a));
  assert(!(*c < *b));
}

/// отсортировать 3 элемента (устойчиво)
template <class It>
inline void ascend3(const It a, const It c)
{
  using std::next;
  ascend3(a, next(a), c);
}




/// отсортировать 4 элемента (устойчиво?)
template <class It>
inline void ascend4(const It a, const It b, const It c, const It d)
{
  ascend(b, d);
  ascend(a, c);
  ascend(c, d);
  ascend(a, b);
  ascend(b, c);
  
  assert(!(*b < *a));
  assert(!(*c < *b));
  assert(!(*d < *c));
}

/// отсортировать 4 элемента (устойчиво?)
template <class It>
inline void ascend4(const It a, const It d)
{
  using std::next;
  const It b = next(a), c = next(b);
  ascend4(a, b, c, d);
}

/// Сортировка слияниями без рекурсии (восходящая) + микрооптимизация (начинает не с пар, а с четвёрок).
template <class RanIt> // поддерживает не только размеры степени двойки
void merge_sort_2(RanIt begin, RanIt end)
{
  using Diff = typename std::iterator_traits <RanIt>::difference_type;
  const Diff size = end - begin;
  if (size < 2)
    return;

  // оптимизация: отсортировать все четвёрки
  for (Diff i = 0, isize = size & -4; i < isize; i += 4)
    ascend4(begin + i, begin + (i + 3));
  // "хвост"
  switch (size & 3)
  {
  case 2: ascend  (end - 2, end - 1); break;
  case 3: ascend3 (end - 3, end - 1); break;
  }

  // все восьмёрки и т.д.
  if (size < 5)
    return;

  // промежуточный буфер
  using Value = typename std::iterator_traits <RanIt>::value_type;
  std::vector<Value> buf (size);

  // собственно сортировка, сливает подпоследовательности размера gap
  for (Diff gap = 4, gap2 = 8; gap < size; gap = gap2, gap2 *= 2) 
  {
    for (Diff i = 0; i + gap < size; i += gap2)
    {
      const RanIt b     = begin + i, 
                  bgap  = begin + (i + gap),
                  bgap2 = begin + std::min(size, i + gap2);

      // ещё не отсортировано?
      if (*bgap < *(bgap - 1))
      {
        const auto buf_end =
          std::merge(b, bgap, bgap, bgap2, buf.begin());
        std::copy(buf.begin(), buf_end, b);
      }
    }
  }
}


/// Сортировка слияниями без рекурсии (восходящая) + микрооптимизация (начинает не с пар, а с четвёрок).
template <class RanIt> // поддерживает не только размеры степени двойки
void inplace_merge_sort_2(RanIt begin, RanIt end)
{
  using Diff = typename std::iterator_traits <RanIt>::difference_type;
  const Diff size = end - begin;
  if (size < 2)
    return;

  // оптимизация: отсортировать все четвёрки
  for (Diff i = 0, isize = size & -4; i < isize; i += 4)
    ascend4(begin + i, begin + (i + 3));
  // "хвост"
  switch (size & 3)
  {
  case 2: ascend  (end - 2, end - 1); break;
  case 3: ascend3 (end - 3, end - 1); break;
  }

  // все восьмёрки и т.д.
  if (size < 5)
    return;

  // собственно сортировка, сливает подпоследовательности размера gap
  for (Diff gap = 4, gap2 = 8; gap < size; gap = gap2, gap2 *= 2) 
  {
    for (Diff i = 0; i + gap < size; i += gap2)
    {
      const RanIt b     = begin + i, 
                  bgap  = begin + (i + gap),
                  bgap2 = begin + std::min(size, i + gap2);

      // ещё не отсортировано?
      if (*bgap < *(bgap - 1))
        std::inplace_merge(b, bgap, bgap2);
    }
  }
}


/// маркировка упорядоченных подпоследовательностей (> 0 -- возрастающая, < 0 -- убывающая),
/// возвращает количество размеченных подпоследовательностей
template <class It, class Subs>
std::ptrdiff_t mark_sorted_subsequences(It begin, It end, Subs &subs)
{
  std::ptrdiff_t size = end - begin, subs_last = 0, subs_i = 0;
  // конечный автомат, 
  // последовательность эквивалентных считаем "возрастающей" (не требует сортировки)
  while (true)
  {
    // начинается новая подпоследовательность
    if (++subs_last == size)
      goto subs_finish; // -- состоит из единственного элемента в конце

    if (begin[subs_last] < begin[subs_last - 1])
    { // убывающая
      while (true)
      {
        if (++subs_last == size)
        {
          subs[subs_i++] = -size;
          goto subs_end;
        }
        // прервалась?
        if (begin[subs_last - 1] < begin[subs_last])
        {
          subs[subs_i++] = -subs_last;
          break; // "перезапуск"
        }
      }
    }
    else // возрастающая
    {
      while (true)
      {
        if (++subs_last == size)
          goto subs_finish;
        // прервалась?
        if (begin[subs_last] < begin[subs_last - 1])
        {
          subs[subs_i++] = subs_last;
          break; // "перезапуск"
        }
      }
    }
  }

subs_finish:
  subs[subs_i++] = size;
subs_end:
  return subs_i;
}


/// восходящая сортировка слияниями с размеченными возрастающими подпоследовательностями
template <class RanIt, class Subs, class Diff>
void merge_up_sort(RanIt begin, RanIt end, Subs &subs, Diff subs_i)
{
  Diff size = end - begin;
  // промежуточный буфер
  using Value = typename std::iterator_traits <RanIt>::value_type;
  std::vector<Value> buf (size);
  do
  {
    Diff a = 0, i = 0, j = 0;
    // цикл объединения пар последовательностей
    while (true)
    {
      // конец первой последовательности
      const Diff b = subs[i++];
      if (b == size) // нет второй
        break;

      // конец второй последовательности
      const Diff c = subs[i++];

      // объединить, если ещё не отсортировано
      if (begin[b] < begin[b - 1])
      {
        const auto buf_end =
          std::merge(begin + a, begin + b, begin + b, begin + c, buf.begin());
        std::copy(buf.begin(), buf_end, begin + a);
      }

      // сохранить конец объединённой последовательности
      subs[j] = c;
      // дошли до конца?
      if (c == size)
        break;

      ++j;
      a = c;

      if (i >= subs_i)
        break;
    }
    subs[subs_i = j] = size;
  } while (subs_i > 0);
}


/// Восходящая сортировка слияниями с предварительным выявляением монотонных последовательностей.
template <class RanIt>
void merge_sort_3(RanIt begin, RanIt end)
{
  using Diff = typename std::iterator_traits <RanIt>::difference_type;
  const Diff size = end - begin;
  if (size < 2)
    return;

  // 1) структура подпоследовательностей ("шаги")
  // > 0 -- возрастающая, < 0 -- убывающая
  std::vector<Diff> subs (size / 2 + 1);
  Diff subs_i = mark_sorted_subsequences(begin, end, subs);

  // 2) обратить все убывающие
  for (Diff sub_begin = 0, i = 0; i < subs_i; ++i)
  {
    Diff sub_end = subs[i];
    if (sub_end < 0)
    {
      subs[i] = sub_end = -sub_end;
      std::reverse(begin + sub_begin, begin + sub_end);
    }
    sub_begin = sub_end;
  }

  // 3) восходящее слияние
  merge_up_sort(begin, end, subs, subs_i);
}


// Двоичный целочисленный логарифм, округлённый вниз + 1, возвращает 0 для x == 0 (наивный алгоритм).
inline std::size_t ulog2r(std::size_t x)
{
  std::size_t result = 0;
  while (x != 0)
    ++result, x >>= 1u;
  return result;

}

/// восходящая сортировка с предварительной подготовкой упорядоченных последовательностей
/// сортировкой вставками в число степени двойки
template <class RanIt>
void merge_sort_4(RanIt begin, RanIt end)
{
  if (begin == end) return;
  using Diff = typename std::iterator_traits<RanIt>::difference_type;

  // целевой размер блока <= 16 элементов (для реальной сортировки -- до 30)
  const Diff
    size = end - begin,
    log_blocks = ulog2r((std::size_t)size >> 4),
    blocks = Diff(1) << log_blocks,
    block_size = size / blocks + 1;

  // 1) отсортировать блоки, помечая их как упорядоченные подпоследовательности
  std::vector<Diff> subs(blocks);
  RanIt block = begin;
  for (Diff i = 0; i < blocks; ++i)
  {
    const RanIt block_end = block + std::min(block_size, end - block);
    subs[i] = block_end - begin;
    insertion_sort(block, block_end);
    block = block_end;
  }

  // 2) выполнить восходящее слияние
  merge_up_sort(begin, end, subs, blocks);
}


///////////////////////////////////////////////////////////////////////////////

/// Пирамидальная сортировка: корректировка пирамиды.
template <class RanIt>
void adjust_heap(RanIt begin, std::size_t sz, std::size_t root)
{
  using std::swap;
  while (true)
  {
    std::size_t left    = 2 * root + 1,
                right   = left + 1, 
                largest = root;

    if (left < sz   && begin[largest] < begin[left])
      largest = left;
    if (right < sz  && begin[largest] < begin[right])
      largest = right;
    if (largest == root)
      return;

    swap(begin[root], begin[largest]);
    root = largest; // продолжить с поддеревом
  }
}  

/// Пирамидальная сортировка: создание пирамиды на месте.
template <class RanIt> // то же, что и std::make_heap
void my_make_heap(RanIt begin, RanIt end)
{
  const auto sz = end - begin; // в норме число со знаком
  assert(0 <= sz);
  for (auto i = sz / 2; 0 <= i; --i)
    adjust_heap(begin, sz, i);
}

/// Пирамидальная сортировка: сортировка пирамиды.
template <class RanIt>
void my_sort_heap(RanIt begin, RanIt end)
{
  using std::swap;
  for (auto i = end - begin - 1; 0 < i; --i)
  {
    // обмен наибольшего с последним в пирамиде
    swap(begin[i], begin[0]);
    // коррекция остатка пирамиды
    adjust_heap(begin, i, 0);
  }
}

/// Пирамидальная сортировка.
template <class RanIt>
void heap_sort(RanIt begin, RanIt end)
{
  my_make_heap(begin, end);
  my_sort_heap(begin, end);
}


/// Пирамидальная сортировка средствами STL.
template <class RanIt>
void std_heap_sort(RanIt begin, RanIt end)
{
  std::make_heap(begin, end);
  std::sort_heap(begin, end);
}

/// Гибрид стандартной и "собственной" пирамидальных сортировок.
template <class RanIt>
void hybrid_heap_sort(RanIt begin, RanIt end)
{
  my_make_heap(begin, end);
  std::sort_heap(begin, end);
}


/// Пирамидальная сортировка: корректировка пирамиды-3.
template <class RanIt>
void adjust_heap3(RanIt begin, std::size_t sz, std::size_t root)
{
  using std::swap;
  while (true)
  {
    std::size_t ch1 = 3 * root + 1,
                ch2 = ch1 + 1,
                ch3 = ch1 + 2,
                largest = root;

    if (ch1 < sz && begin[largest] < begin[ch1])
      largest = ch1;
    if (ch2 < sz && begin[largest] < begin[ch2])
      largest = ch2;
    if (ch3 < sz && begin[largest] < begin[ch3])
      largest = ch3;
    if (largest == root)
      return;

    swap(begin[root], begin[largest]);
    root = largest; // продолжить с поддеревом
  }
}

/// Пирамидальная сортировка: создание пирамиды-3 на месте.
template <class RanIt> 
void make_heap3(RanIt begin, RanIt end)
{
  const auto sz = end - begin; // в норме число со знаком
  assert(0 <= sz);
  for (auto i = sz / 3; 0 <= i; --i)
    adjust_heap3(begin, sz, i);
}

/// Пирамидальная сортировка: сортировка пирамиды-3.
template <class RanIt>
void sort_heap3(RanIt begin, RanIt end)
{
  using std::swap;
  for (auto i = end - begin - 1; 0 < i; --i)
  {
    // обмен наибольшего с последним в пирамиде
    swap(begin[i], begin[0]);
    // коррекция остатка пирамиды
    adjust_heap3(begin, i, 0);
  }
}

/// Пирамидальная сортировка-3.
template <class RanIt>
void heap3_sort(RanIt begin, RanIt end)
{
  make_heap3(begin, end);
  sort_heap3(begin, end);
}


/// Пирамидальная сортировка: корректировка пирамиды-4.
template <class RanIt>
void adjust_heap4(RanIt begin, std::size_t sz, std::size_t root)
{
  while (true)
  {
    std::size_t ch1 = 4 * root + 1,
                largest = root;

    if (ch1 + 3 < sz)
    {
      for (std::size_t i = 0; i < 4; ++i)
        if (begin[largest] < begin[ch1 + i])
          largest = ch1 + i;
    }
    else
    {
      for (std::size_t i = 0; ch1 + i < sz; ++i)
        if (begin[largest] < begin[ch1 + i])
          largest = ch1 + i;
    }

    if (largest == root)
      return;

    swap(begin[root], begin[largest]);
    root = largest; // продолжить с поддеревом
  }
}

/// Пирамидальная сортировка: создание пирамиды-4 на месте.
template <class RanIt>
void make_heap4(RanIt begin, RanIt end)
{
  const auto sz = end - begin; // в норме число со знаком
  assert(0 <= sz);
  for (auto i = sz / 4; 0 <= i; --i)
    adjust_heap4(begin, sz, i);
}

/// Пирамидальная сортировка: сортировка пирамиды-4.
template <class RanIt>
void sort_heap4(RanIt begin, RanIt end)
{
  for (auto i = end - begin - 1; 0 < i; --i)
  {
    swap(begin[i], begin[0]);
    adjust_heap4(begin, i, 0);
  }
}

/// Пирамидальная сортировка-4.
template <class RanIt>
void heap4_sort(RanIt begin, RanIt end)
{
  make_heap4(begin, end);
  sort_heap4(begin, end);
}



///////////////////////////////////////////////////////////////////////////////

//
// алгоритмы на сортирующих сетях
// данные реализации корректно работают только с последовательностями размера степени двойки

// получить степень двойки, не превосходящую заданное число, не определено для отрицательных
template <class Int>
inline Int lower_pow2(Int x)
{
  static_assert (std::is_integral<Int>::value, 
    "This algorithm uses sizeof(Int), which works only for builtin integral types");
  const std::size_t bits = sizeof(Int) * CHAR_BIT / 2;
  for (std::size_t bit = 1; bit <= bits; bit *= 2)
    x |= x >> bit;
  return x - (x >> 1);
}

// ограничить диапазон степенью двойки
template <class It>
inline void force_pow2_length(It begin, It &end)
{
  using std::distance;
  using std::next;
  end = next(begin, lower_pow2(distance(begin, end)));
}


/// рекурсивная битоническая сортировка: слияние
template <class It>
void bitonic_merge(It begin, It end, bool up)
{
  using std::distance;
  using std::next;
  using std::iter_swap;

  auto gap = distance(begin, end) / 2;
  if (gap > 0)
  {
    It mid = next(begin, gap);
    for (It p = begin, q = mid; p != mid; ++p, ++q)
    {
      if (up == (*q < *p))
        iter_swap(q, p);
    }

    bitonic_merge(begin, mid, up);
    bitonic_merge(mid,   end, up);
  }
}

/// рекурсивная битоническая сортировка: сортировка
template <class It>
void bitonic_sort_impl(It begin, It end, bool up)
{
  using std::distance;
  using std::next;
  const auto gap = distance(begin, end) / 2;
  if (gap > 0)
  {
    const It mid = next(begin, gap);
    bitonic_sort_impl(begin, mid, true);
    bitonic_sort_impl(mid,   end, false);
    bitonic_merge(begin, end, up);
  }
}

template <class It>
void bitonic_sort(It begin, It end)
{
  force_pow2_length(begin, end); // caveat user! "хвост" последовательности просто игнорируется
  bitonic_sort_impl(begin, end, true);
}


/// нерекурсивная битоническая сортировка
template <class RanIt>
void bitonic_sort_2(RanIt begin, RanIt end)
{
  force_pow2_length(begin, end); // caveat user!
  using std::iter_swap;
  const int len = end - begin;
  for (int i = 2; i <= len; i *= 2) 
  {
    for (int j = i; 1 < j; j /= 2)
    {
      const int half_j = j / 2;
      for (int k = 0; k < len; k += j) 
      {
        const bool up = (k & i) == 0;
        RanIt p = begin + k, 
              q = p + half_j;

        for (RanIt p_end = q; p != p_end; ++p, ++q)
          if (up == *q < *p)
            iter_swap(p, q);
      }
    }
  }
}

///////////////////////////////////////////////////////////////////////////////

/// чётно-нечётная сортировка: слияние
template <class RanIt, class Diff>
void oddeven_merge(RanIt begin, RanIt end, Diff r)
{
  const Diff step = r * Diff(2);
  if (step < end - begin)
  {
    oddeven_merge(begin, end - r, step);
    oddeven_merge(begin + r, end, step);
    RanIt p = begin + r,
          q = p + r;

    for (RanIt p_end = end - r; p < p_end; p += step, q += step)
      ascend(p, q);
  }
  else
    ascend(begin, begin + r);
}

/// чётно-нечётная сортировка: сортировка
template <class RanIt>
void oddeven_sort_impl(RanIt begin, RanIt end)
{
  using Diff = typename std::iterator_traits<RanIt>::difference_type;
  const Diff gap = (end - begin) / 2;
  if (gap > 0)
  {
    RanIt mid = begin + gap;
    oddeven_sort_impl(begin, mid);
    oddeven_sort_impl(mid,   end);
    oddeven_merge(begin, end, Diff(1));
  }
}

template <class RanIt>
void oddeven_sort(RanIt begin, RanIt end)
{
  force_pow2_length(begin, end); // caveat user!
  oddeven_sort_impl(begin, end);
}

///////////////////////////////////////////////////////////////////////////////

/// попарная сортировка: слияние
template <class RanIt, class Diff>
void pairwise_merge(RanIt begin, RanIt end, Diff r, Diff step)
{
  if (r < end - begin)
  {
    pairwise_merge( begin, end - step, r, Diff(2) * step );
    pairwise_merge( begin + step, end, r, Diff(2) * step );
    for (RanIt p = begin, p_end = end - step; p < p_end; p += step)
      ascend( p + r, p + step );
  }
}

/// попарная сортировка: сортировка
template <class RanIt, class Diff>
void pairwise_sort_impl(RanIt begin, RanIt end, Diff r)
{
  const Diff step = r * Diff(2);
  if (step < end - begin)
  {
    // разбить
    for (RanIt p = begin + r; p < end; p += step)
      ascend(p - r, p);

    // отсортировать
    pairwise_sort_impl(begin, end - r, step);
    pairwise_sort_impl(begin + r, end, step);

    // объединить
    pairwise_merge(begin, end, r, step);    
  }
  else if (r < end - begin)
    ascend(begin, begin + r);
}

template <class RanIt>
void pairwise_sort(RanIt begin, RanIt end)
{
  force_pow2_length(begin, end); // caveat user!
  using Diff = typename std::iterator_traits<RanIt>::difference_type;
  pairwise_sort_impl(begin, end, Diff(1));
}

///////////////////////////////////////////////////////////////////////////////

//
// другие алгоритмы

/// "глупая" сортировка рандомом, впрочем, лучше bogosort :)
template <class It>
void random_try_sort(It begin, It end)
{
  using std::next;
  using std::distance;
  using Diff = typename std::iterator_traits<It>::difference_type;
  const Diff sz = distance(begin, end);
  while (!std::is_sorted(begin, end))
  {
    for (Diff i = 0; i < sz - 1; ++i)
    {
      // не работает для последовательностей длины больше RAND_MAX!
      ascend( next(begin, i),
              next(begin, (i + 1) + std::rand() % (sz - (i + 1))) );
    }   
  }
}

///////////////////////////////////////////////////////////////////////////////

// операция применения перестановки к массиву (размеры массива и перестановки совпадают)
template <class RanIt, class PermIt> // perm -- итератор начала перестановки
void apply_permutation(RanIt begin, RanIt end, PermIt perm)
{
  using std::iter_swap;
  using Diff = typename std::iterator_traits<RanIt>::difference_type;
  const Diff size = end - begin;

  // чтобы реализовать правильный порядок, необходимо выполнить перестановку order,
  // будем выполнять её факторизацию на циклы с одновременной ротацией элементов

  // 1) затрём все одноэлементные циклы, не требующие реальных перемещений объектов
  for (Diff i = 0; i < size; ++i)
    if (perm[i] == i)
      perm[i] = size;

  // 2) выполним все двухэлементные циклы обменами
  for (Diff i = 0; i < size - 1; ++i)
  {
    if (perm[i] != size && i == perm[perm[i]])
    {
      iter_swap(begin + i, begin + perm[i]);
      perm[perm[i]] = size;
      perm[i] = size;
    }
  }

  // 3) теперь будем "разматывать" оставшиеся циклы, храня временную копию объекта
  for (Diff i = 0; i < size - 2; ++i)
  {
    if (perm[i] != size)
    {
      auto displaced_element = std::move(begin[i]);
      Diff j = i;
      for (Diff k; perm[j] != i; k = j, j = perm[j], perm[k] = size)
        begin[j] = std::move(begin[perm[j]]);
      begin[j] = std::move(displaced_element);
      perm[j] = size;
    }
  }
}


// сортировка созданием BST (несбалансированное для простоты)
// фактически, это косвенная сортировка
template <class RanIt>
void tree_sort(RanIt begin, RanIt end)
{
  using Diff = typename std::iterator_traits<RanIt>::difference_type;
  const Diff size = end - begin;
  // size используется также как NULL-маркер, поэтому заполняем метки
  // левых и правых потомков значением "size"
  std::vector<Diff> L (size, size), R (size, size);
  // поместить все элементы в дерево
  // корень -- элемент *begin
  for (RanIt p = begin + 1; p != end; ++p)
  {
    Diff parent = 0; // begin[0]
    while (true)
    {
      Diff *const next = *p < begin[parent]? &L[parent]: &R[parent];
      if (*next == size) // добавить, покинуть цикл
      {
        *next = p - begin;
        break;
      }
      parent = *next;
    }
  }

  // обойти дерево, выводя элементы в отсортированном порядке
  std::vector<Diff> order(size); // сортирующая перестановка
  std::stack<Diff>  parents; // стек предков текущего узла
  Diff current = 0; // текущая позиция в результирующем массиве (order)
  Diff node = 0; // текущий узел дерева (индекс элемента в L, R, исходном массиве)
  while (true) // одна ветка "рекурсии" раскрыта в цикле, поэтому "правые" в стек не идут
  {
    if (node != size)
    {
      for (; L[node] != size; node = L[node])
        parents.push(node);
    }
    else
    {
      if (parents.empty())
        break;
      node = parents.top();
      parents.pop();
    }

    order[current++] = node;
    node = R[node];
  }

  // применим перестановку
  apply_permutation(begin, end, order.begin());
}


// O(NlogN) tree sort поверх сбалансированного дерева, реализованного в STL
template <class It>
void tree_sort_balanced(It begin, It end)
{
  using Value = typename std::iterator_traits<It>::value_type;
  std::multiset<Value> tree(begin, end);
  std::copy( tree.cbegin(), tree.cend(), begin );
}

// TASK: реализовать косвенную сортировку поверх любого алгоритма сортировки и apply_permutation

// TASK: реализовать сортировку Шелла.


///////////////////////////////////////////////////////////////////////////////

// Сортировка с помощью nth_element.
// По идее, напоминает быструю с выбором медианы...
template <class RanIt>
void nel_sort(RanIt begin, RanIt end)
{
  using std::swap;
  switch (const auto size = end - begin)
  {
  case 0: case 1: return;
  case 2: return ascend(begin, begin + 1);
  case 3: return ascend3(begin, begin + 2);
  case 4: return ascend4(begin, begin + 3);
  default:
    {
      const auto mid = begin + size / 2;
      std::nth_element(begin, mid, end);
      nel_sort(begin, mid);
      nel_sort(mid, end);
    }
  }
}


// Эксперимент: make_heap + quick_sort.
template <class RanIt>
void heap_quick_sort(RanIt begin, RanIt end)
{
  using RevIt = std::reverse_iterator<RanIt>;
  my_make_heap(RevIt(end), RevIt(begin));
  quick_sort(begin, end);
}

// Эксперимент: shuffle + quick_sort.
template <class RanIt>
void shuffle_quick_sort(RanIt begin, RanIt end)
{
  std::mt19937 rng(end - begin);
  std::shuffle(begin, end, rng);
  quick_sort(begin, end);
}

// Эксперимент: shuffle + tree_sort.
template <class RanIt>
void shuffle_tree_sort(RanIt begin, RanIt end)
{
  std::mt19937 rng(end - begin);
  std::shuffle(begin, end, rng);
  tree_sort(begin, end);
}

// Эксперимент: ascend + quick_sort
template <class RanIt>
void a2_quick_sort(RanIt begin, RanIt end)
{
  const auto sz = end - begin;
  if (sz > 3)
  {
    const auto sz3 = sz / 3;
    for (auto p1 = begin, p2 = p1 + 1 + sz3;
      p2 != end; ++p1, ++p2)
      ascend(p1, p2);
  }
  
  quick_sort(begin, end);
}

// Эксперимент: ascend3 + quick_sort
template <class RanIt>
void a3_quick_sort(RanIt begin, RanIt end)
{
  switch (const size_t sz = end - begin)
  {
  case 0: case 1: return;
  case 2: ascend(begin, begin + 1); return;
  case 3: ascend3(begin, begin + 2); return;
  default:
    if (sz > 6)
    {
      const auto sz3 = sz / 3;
      for (auto
        p1 = begin,
        p2 = p1 + 1 + sz3 / 2,
        p3 = p1 + 2 + sz3;
        p3 != end; ++p1, ++p2, ++p3)
        ascend3(p1, p2, p3);
    }
  }
  quick_sort(begin, end);
}

// Эксперимент: ascend4 + quick_sort
template <class RanIt>
void a4_quick_sort(RanIt begin, RanIt end)
{
  switch (const size_t sz = end - begin)
  {
  case 0: case 1: return;
  case 2: ascend(begin, begin + 1); return;
  case 3: ascend3(begin, begin + 2); return;
  case 4: ascend4(begin, begin + 3); return;
  default:
    if (sz > 12)
    {
      const auto sz3 = sz / 3;
      for (auto
        p1 = begin,
        p2 = p1 + 1 + sz3 / 4,
        p3 = p1 + 2 + sz3 / 2,
        p4 = p1 + 3 + sz3;
        p4 != end; ++p1, ++p2, ++p3, ++p4)
        ascend4(p1, p2, p3, p4);
    }
  }
  quick_sort(begin, end);
}


/// Сортирующая корректировка пирамиды.
template <class RanIt>
void adjust_sort_heap(RanIt begin, std::size_t sz, std::size_t root)
{
  using std::swap;
  std::size_t left    = 2 * root + 1,
              right   = left + 1;

  if (right < sz && begin[left] < begin[right])
  {
    swap(begin[left], begin[right]);
    adjust_sort_heap(begin, sz, right);
  }

  if (left < sz && begin[root] < begin[left])
  {
    swap(begin[root], begin[left]);
    adjust_sort_heap(begin, sz, left);

    if (right < sz && begin[left] < begin[right])
    {
      swap(begin[left], begin[right]);
      adjust_sort_heap(begin, sz, right);
    }
  }
}  

/// Создание сортированной пирамиды на месте.
template <class RanIt>
void make_sort_heap(RanIt begin, RanIt end)
{
  switch (const auto sz = end - begin)
  {
  case 0: case 1: return;
  case 2: return ascend(begin, begin + 1);
  default:
    make_sort_heap(begin + sz/2 + 1, end);
    for (auto i = sz / 2; 0 <= i; --i)
      adjust_sort_heap(begin, sz, i);
  }
}

/// Сортировка с помощью make_sort_heap.
template <class RanIt>
void sort_heap_sort(RanIt begin, RanIt end)
{
  make_sort_heap(begin, end);
  std::sort_heap(begin, end);
}
