﻿/*
 * This file is part of AV Orchid.
 * Copyright (c) 2013, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * @file Screen.hpp
 *
 * Базовый класс экрана с привязкой обработчиков событий FreeGLUT (не все возможные)
 * к виртуальным функциям текущего экрана.
 * Вспомогательные функции drawString (рисование строки текста)
 * и fit_window для подгонки окна под заданные размеры.
 */
#pragma once
#include <string>


/// FreeGlut-реализация средств отображения текста, пытается "впихнуть" строку в указанный прямоугольник
/**
 * @param font может быть GLUT_STROKE_ROMAN или GLUT_STROKE_MONO_ROMAN
 * @param right может быть <= left, тогда выбирается автоматически
 * @param top может быть <= bottom, тогда выбирается автоматически
 * Для выбора толщины линий, которыми рисуется текст, можно использовать glLineWidth(толщина в пикселях).
 */
void draw_stroke_string(void *font, const std::string &text,
                      float left, float bottom, float right = 0.f, float top = 0.f);

/// установить размеры окна
/**
 * @param target_width желаемая ширина окна (пиксели)
 * @param target_height желаемая высота окна (пиксели)
 *
 * В случае выхода за пределы экрана, вписывает в экран в пропорции target_width : target_height.
 * Естественно, оба параметра должны быть больше нуля.
 * Позиционирует окно в центре экрана.
 */
void fit_window(float target_width, float target_height);


/// получение размеров области рисования окна, экрана (в пикселях)
/**
 * Получает информацию о текущем окне средствами glutGet.
 */
namespace Drawing_area_info
{
  int x(); ///< координата x верхнего левого угла
  int y(); ///< координата y верхнего левого угла
  int width();  ///< ширина области
  int height(); ///< высота области
  int border_width();  ///< ширина границы окна
  int border_height(); ///< высота границы окна
  int screen_width();  ///< ширина экрана (полная)
  int screen_height(); ///< высота экрана (полная)
};


/// абстракция текущего экрана + инкапсуляция привязки обработчиков событий FreeGlut
/**
 * Предполагается, что классы-наследники дадут свои определения нужным им обработчикам событий.
 * Готовые элементы реализации предполагают выключенный буфер глубины и включенную двойную буферизацию.
 */
class AScreen
{
public:
  /// базовый деструктор выполняет также "отвязку" данного экрана, если он был активным, но on_deselect уже не вызывает
  virtual ~AScreen();

  /// использовать данный экран для обработки событий (сделать его активным)
  /**
   * В каждый конкретный момент может быть выбран только один активный экран.
   */
  void select();

  /// узнать выбран ли какой-либо объект AScreen в качестве активного
  static bool has_active();
  /// получить ссылку на текущий активный экран
  static AScreen& get_active();

protected:
  /// вызывается при выборе данного экрана
  virtual void on_select() {}
  /// вызывается при выборе другого экрана, если этот экран был активным
  virtual void on_deselect() {}
  /// отсутствие событий, вызывается периодически
  virtual void on_idle() {}
  /// перерисовка кадра
  virtual void on_display() = 0;
  /// действие, выполняемое перед отрисовкой кадра
  virtual void on_display_before();
  /// действие, выполняемое после отрисовки кадра
  virtual void on_display_after();
  /// изменение размеров окна извне
  /**
   * @param width новая ширина области рисования (в пикселах)
   * @param height новая высота области рисования (в пикселах)
   */
  virtual void on_reshape(int width, int height);
  /// перемещение окна (не реализовано в FreeGLUT 2.8)
  /**
   * @param x смещение вправо от левого верхнего угла экрана
   * @param y смещение вниз от левого верхнего угла экрана
   */
  virtual void on_position(int x, int y) {}
  /// изменение видимости окна (присутствует на экране, не присутствует на экране)
  /**
   * @param visible признак видимости окна (true -- видно)
   */
  virtual void on_visibility(bool visible) {}
  /// нажатие клавиши (код клавиши преобразуется в соответствующий ASCII-код)
  /**
   * @see AScreen::on_mouse
   * @param code ASCII-код нажатой клавиши (нелатинские буквы могут не работать)
   * @param mouse_x x-координата мыши в окне в момент нажатия
   * @param mouse_y y-координата мыши в окне в момент нажатия
   */
  virtual void on_char(unsigned char code, int mouse_x, int mouse_y) {}
  /// отпускание зажатой клавиши (код клавиши преобразуется в соответствующий ASCII-код)
  /**
   * @see AScreen::on_mouse
   * @param code ASCII-код отпущенной клавиши (нелатинские буквы могут не работать)
   * @param mouse_x x-координата мыши в окне в момент отпускания
   * @param mouse_y y-координата мыши в окне в момент отпускания
   */
  virtual void on_char_up(unsigned char code, int mouse_x, int mouse_y) {}
  /// нажатие специальной клавиши
  /**
   * @see AScreen::on_mouse
   * @param code код (см. GLUT_KEY_...) нажатой клавиши
   * @param mouse_x x-координата мыши в окне в момент нажатия
   * @param mouse_y y-координата мыши в окне в момент нажатия
   */
  virtual void on_special_key(int code, int mouse_x, int mouse_y) {}
  /// отпускание зажатой специальной клавиши
  /**
   * @see AScreen::on_mouse
   * @param code код (см. GLUT_KEY_...) отпущенной клавиши
   * @param mouse_x x-координата мыши в окне в момент отпускания
   * @param mouse_y y-координата мыши в окне в момент отпускания
   */
  virtual void on_special_key_up(int code, int mouse_x, int mouse_y) {}
  /// вход/выход мыши в окно
  /**
   * @param entered признак входа указателя мыши в окно (true -- вошёл, false -- вышел)
   */
  virtual void on_entry(bool entered) {}
  /// движение мыши с отпущенными кнопками
  /**
   * Срабатывает периодически с некоторой частотой в процессе движения.
   *
   * @see AScreen::on_mouse
   * @param mouse_x x-координата мыши в окне
   * @param mouse_y y-координата мыши в окне
   */
  virtual void on_hover(int mouse_x, int mouse_y) {}
  /// движение мыши с зажатыми кнопками
  /**
   * Срабатывает периодически с некоторой частотой в процессе движения.
   *
   * @see AScreen::on_mouse
   * @param mouse_x x-координата мыши в окне
   * @param mouse_y y-координата мыши в окне
   */
  virtual void on_drag(int mouse_x, int mouse_y) {}
  /// нажатие кнопки мыши
  /**
   * Координаты мыши отсчитываются от верхнего левого угла окна (не области рисования, без учёта заголовка?)
   * вправо по горизонтали (x) и вниз по вертикали (y).
   *
   * @param code код (GLUT_LEFT_BUTTON, GLUT_RIGHT_BUTTON или GLUT_MIDDLE_BUTTON) нажатой кнопки
   * @param mouse_x x-координата мыши в окне в момент нажатия
   * @param mouse_y y-координата мыши в окне в момент нажатия
   */
  virtual void on_mouse(int button, int mouse_x, int mouse_y) {}
  /// отпускание кнопки мыши
  /**
   * @see AScreen::on_mouse
   * @param code код нажатой кнопки
   * @param mouse_x x-координата мыши в окне в момент отпускания
   * @param mouse_y y-координата мыши в окне в момент отпускания
   */
  virtual void on_mouse_up(int button, int mouse_x, int mouse_y) {}
  /// прокручивание колеса мыши
  /**
   * @param wheel_number -- "номер колеса" (можно игнорировать?)
   * @param direction -- направление прокрутки (теоретически должно быть равно либо +1, либо -1)
   * @param mouse_x x-координата мыши в окне в момент срабатывания
   * @param mouse_y y-координата мыши в окне в момент срабатывания
   */
  virtual void on_mouse_wheel(int wheel_number, int direction, int mouse_x, int mouse_y) {}

private:
  // FreeGlut вызывает нижеперечисленные функции в качестве обработчиков событий.
  // Static-функции по сути являются "свободными", и могут быть переданы в качестве колбеков FreeGlut'у.
  // То, что они являются членами AScreen, позволяет им переадресовать вызов
  // обработчику сообщения, являющимуся методом объекта AScreen (вызвать protected-функцию).
  static void current_screen_select(AScreen*);
  static void on_idle_wrapper();
  static void on_display_wrapper();
  static void on_reshape_wrapper(int width, int height);
  static void on_position_wrapper(int x, int y);
  static void on_visibility_wrapper(int status);
  static void on_char_wrapper(unsigned char code, int mouse_x, int mouse_y);
  static void on_char_up_wrapper(unsigned char code, int mouse_x, int mouse_y);
  static void on_special_key_wrapper(int code, int mouse_x, int mouse_y);
  static void on_special_key_up_wrapper(int code, int mouse_x, int mouse_y);
  static void on_entry_wrapper(int status);
  static void on_hover_wrapper(int mouse_x, int mouse_y);
  static void on_drag_wrapper(int mouse_x, int mouse_y);
  static void on_mouse_button_wrapper(int button, int state, int x, int y);
  static void on_mouse_wheel_wrapper(int wheel, int direction, int x, int y);
};
